/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2011. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL.
 * $Id: BackendMesh.cpp 347 2011-09-20 20:58:41Z ivoras $
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <pthread.h>
#include <string.h>
#include <poll.h>
#include <err.h>

#include "logger.h"
#include "BackendMesh.h"
#include "LockPool.h"
#include "mc_protocol.h"

extern char **backend_peers;
extern unsigned n_backend_peers;


/* Basically, this is the same protocol as defined in mc_protocol.h */
#define bc_header mc_header
#define bc_init_packet mc_init_packet
#define bc_resp_status mc_resp_status
#define BC_DEFAULT_RECV_SIZE	64

enum BCMD {
	BCMD_RDLOCK	= 200,
	BCMD_WRLOCK	= 201,
	BCMD_UNLOCK	= 202,
	BCMD_TRY_RDLOCK	= 203,
	BCMD_TRY_WRLOCK = 204,

	BCRESP_STATUS	= 210,
};

enum BCRESP {
	BCRESP_INVALID	= 0,
	BCRESP_LOCK_OK	= 1,
	BCRESP_LOCK_FAIL = 2,
};

struct __packed bc_lock_op {
	struct bc_header	header;
	unsigned char 		lock_type;
	unsigned char 		flags;
	unsigned short int 	n_locks;
	unsigned 		nlist[];
};


void
BackendConnection::_init() {
	pthread_mutex_init(&lock_op_wait_list_lock, NULL);
}


BackendConnection::BackendConnection(int socket, struct sockaddr_in *sin, BackendMesh *bmesh) {
	_init();
	this->sock = socket;
	this->bmesh = bmesh;
	memcpy(&this->sin, sin, sizeof(this->sin));
	state = BC_REGULAR;
}


BackendConnection::BackendConnection(struct sockaddr_in *sin, enum BC_STATE st, BackendMesh *bmesh) {
	_init();
	this->sock = -1;
	this->bmesh = bmesh;
	memcpy(&this->sin, sin, sizeof(this->sin));
	state = st;
}


BackendConnection::~BackendConnection() {
	if (sock >= 0)
		close(sock);
}


/**
 * Sends a formatted command to the other side. Makes sure only one thread does
 * it at a time.
 */
void
BackendConnection::sendCommand(struct mc_header *hcmd) throw (ConnectionException) {
	sock_write_mtx.Lock();
	ssize_t res = send(sock, hcmd, hcmd->total_size, 0);
	if (res <= 0 || res != (ssize_t)(hcmd->total_size)) {
		sock_write_mtx.Unlock();
		throw new ConnectionException((char*)"send() failure", errno);
	}
	sock_write_mtx.Unlock();
}


int
BackendConnection::postLockOp(struct bc_lock_op *pkt) {
	try {
		sendCommand((struct mc_header*)pkt);
	} catch (ConnectionException *e) {
		return 1;
	}
	return 0;
}


/* Waits (blocks) for a response on the given packet (assumed sent) */
int
BackendConnection::waitLockOpResponse(struct bc_lock_op *pkt) {
	struct lock_op_wait *low = (struct lock_op_wait*) malloc(sizeof(*low));
	int result;

	pthread_cond_init(&low->cv, NULL);
	low->pkt = pkt;
	low->cn = this;
	low->result = 0;

	CRIT_ASSERT(pthread_mutex_lock(&lock_op_wait_list_lock) == 0);
	lock_op_wait_list[pkt->header.seq] = low;
	CRIT_ASSERT(pthread_cond_wait(&low->cv, &lock_op_wait_list_lock) == 0);
	CRIT_ASSERT(pthread_mutex_unlock(&lock_op_wait_list_lock) == 0);

	result = low->result;
	free(low);
	return result;
}


/**
 * Runs the BackendConnection thread (for a single connection to another node in the
 * mesh of peers). It needs to be a thread because then it can directly hold locks
 * in the LockPool.
 */
void
BackendConnection::run() {
	bool cleanup_exit = false;
	ssize_t max_size = BC_DEFAULT_RECV_SIZE;
	struct bc_header *rhead = (struct bc_header*) malloc(max_size);
	int res;

	mc_log(LOG_DEBUG, "BackendConnection thread %p starting", this);
	while (!stop_requested) {
		/* Ok, this the only place which reads other-side inquiries and replies,
		 * and dispatches the replies where needed. Because of this, it doesn't
		 * need to protect the socket with a Mutex. */
		DEBUG_STATUS_POINT(bc_st, 'r');
		ssize_t rsize = recv(sock, rhead, max_size, 0);
		DEBUG_STATUS_POINT(bc_st, 'R');

		if (rsize <= 0) {
			mc_log(LOG_ERR, "recv() in BackendConnection::run() failed with errno %d (%s)",
			    errno, strerror(errno));
			break;
		}
		if (rsize < (ssize_t)sizeof(struct bc_header)) {
			char *rdata = (char*) rhead;
			rdata += rsize;
			ssize_t mrsize = recv(sock, rdata, sizeof(struct bc_header) - rsize, MSG_WAITALL);
			if (mrsize <= 0) {
				mc_log(LOG_ERR, "recv() in BackendConnection::run() [header] failed with errno %d (%s)",
				    errno, strerror(errno));
				break;
			}
			rsize += mrsize;
		}
		assert(rsize >= (ssize_t)sizeof(struct bc_header));
		/* We have at least the header here */
		if (rsize < (ssize_t)(rhead->total_size)) {
			/* Read the rest of it */
			if ((ssize_t)(rhead->total_size) > max_size) {
				struct bc_header *rrhead = (struct bc_header*) realloc(rhead, rhead->total_size);
				if (rrhead == NULL) {
					mc_log(LOG_ERR, "realloc(%u) failed in BackendConnection::run()", rhead->total_size);
					break;
				}
				rhead = rrhead;
				max_size = rrhead->total_size;
			}
			while (rsize < (ssize_t)(rhead->total_size)) {
				char *rdata = (char*) rhead;
				rdata += rsize;
				ssize_t mrsize = recv(sock, rdata, rhead->total_size - rsize, MSG_WAITALL);
				if (mrsize <= 0) {
					mc_log(LOG_ERR, "recv() in BackendConnection::run() [body] failed with errno %d (%s)",
					    errno, strerror(errno));
					goto endrun;
				}
				rsize += mrsize;
			}
		}
		/* Ok so now we have the whole packet here. */
		DEBUG_STATUS_POINT(bc_st, 'p');
		assert(rsize == (ssize_t)(rhead->total_size));

		if (rhead->mcmd == BCRESP_STATUS) {
			/* We've received a status response from one of our commands */
			if (lock_op_wait_list.count(rhead->seq) > 0) {
				/* Someone is blocking on this status? */
				struct lock_op_wait *low = lock_op_wait_list[rhead->seq];
				CRIT_ASSERT(pthread_mutex_lock(&lock_op_wait_list_lock) == 0);
				lock_op_wait_list.erase(rhead->seq);
				CRIT_ASSERT(pthread_mutex_unlock(&lock_op_wait_list_lock) == 0);
				struct bc_resp_status *resp = (struct bc_resp_status*)rhead;
				low->result = resp->status;
				CRIT_ASSERT(pthread_cond_signal(&low->cv) == 0);
			}
		} else {
			res = processBackendCommand(rhead);
			if (res == -1) {
				cleanup_exit = true;
				break;
			} else if (res > 0) {
				/* The command returned a status which must be reported back */
				struct bc_resp_status resp_status;
				bc_init_packet(&resp_status, sizeof(resp_status), rhead->seq, BCRESP_STATUS);
				resp_status.status = res;
				resp_status.data = 0;
				res = send(sock, &resp_status, sizeof(resp_status), 0);
				if (res <= 0) {
					mc_log(LOG_ERR, "send() failed in BackendMesh while sending status response.");
					cleanup_exit = true;
					break;
				}
			}
		}
	}
endrun:
	mc_log(LOG_DEBUG, "BackendConnection thread %p exiting", this);
	if (cleanup_exit) {
		close(sock);
		emergencyUnlock();
		sock = -1;
	}
}


/* Records locks taken by this BackendConnection */
void
BackendConnection::recordLocks(struct bc_lock_op *lop) {
	switch (lop->lock_type) {
	case LT_ITEMS:
		for (int i = 0; i < lop->n_locks; i++)
			items_locks.insert(lop->nlist[i]);
		break;
	case LT_TAGS:
		for (int i = 0; i < lop->n_locks; i++)
			tags_locks.insert(lop->nlist[i]);
		break;
	}

}


/* Removes locks from the record */
void
BackendConnection::recordUnLocks(struct bc_lock_op *lop) {
	switch (lop->lock_type) {
	case LT_ITEMS:
		for (int i = 0; i < lop->n_locks; i++)
			items_locks.erase(lop->nlist[i]);
		break;
	case LT_TAGS:
		for (int i = 0; i < lop->n_locks; i++)
			tags_locks.erase(lop->nlist[i]);
		break;
	}

}


/* Process a command from a connected backend */
int
BackendConnection::processBackendCommand(struct bc_header *rhead) {
	struct bc_lock_op * const lop = (struct bc_lock_op*) rhead;

	switch (rhead->mcmd) {
	case BCMD_RDLOCK:
		bmesh->dataLockOp(LOP_RDLOCK, lop->nlist, lop->n_locks, (enum LOCK_TYPE)lop->lock_type);
		recordLocks(lop);
		return BCRESP_LOCK_OK;
	case BCMD_WRLOCK:
		bmesh->dataLockOp(LOP_WRLOCK, lop->nlist, lop->n_locks, (enum LOCK_TYPE)lop->lock_type);
		recordLocks(lop);
		return BCRESP_LOCK_OK;
	case BCMD_UNLOCK:
		bmesh->dataLockOp(LOP_UNLOCK, lop->nlist, lop->n_locks, (enum LOCK_TYPE)lop->lock_type);
		recordUnLocks(lop);
		return BCRESP_LOCK_OK;
	case BCMD_TRY_RDLOCK:
		if (bmesh->dataLockOp(LOP_TRY_RDLOCK, lop->nlist, lop->n_locks, (enum LOCK_TYPE)lop->lock_type)) {
			recordLocks(lop);
			return BCRESP_LOCK_OK;
		}
		return BCRESP_LOCK_FAIL;
	case BCMD_TRY_WRLOCK:
		if (bmesh->dataLockOp(LOP_TRY_WRLOCK, lop->nlist, lop->n_locks, (enum LOCK_TYPE)lop->lock_type)) {
			recordLocks(lop);
			return BCRESP_LOCK_OK;
		}
		return BCRESP_LOCK_FAIL;
	}
	mc_log(LOG_DEBUG, "Unsupported BackendConnection command: %d", rhead->mcmd);
	return -1;
}


/* TODO: review */
void
BackendConnection::emergencyUnlock() {
	unsigned *nlocks;
	int n;

	n = 0;
	nlocks = new unsigned[items_locks.size()];
	for (std::set<int>::iterator i = items_locks.begin(); i != items_locks.end(); i++)
		nlocks[n++] = *i;
	bmesh->dataLockOp(LOP_UNLOCK, nlocks, n, LT_ITEMS);
	delete[] nlocks;

	n = 0;
	nlocks = new unsigned[tags_locks.size()];
	for (std::set<int>::iterator i = tags_locks.begin(); i != tags_locks.end(); i++)
		nlocks[n++] = *i;
	bmesh->dataLockOp(LOP_UNLOCK, nlocks, n, LT_TAGS);
	delete[] nlocks;
}


/* Can only be called outside the thread */
void
BackendConnection::abort() {
	request_stop();
	/* rollback recorded locks */
	emergencyUnlock();
}


BackendMesh::BackendMesh(DataStore *ds) : Thread() {
	is_notifier_setup = false;
	_inotif = NULL;
	_tnotif = NULL;
	this->ds = ds;

	CRIT_ASSERT((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) >= 0);

	int flags = 1;
	CRIT_ASSERT(setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char*)&flags, sizeof(flags)) == 0);
	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl(INADDR_ANY);
	sin.sin_port = htons(MC_BACKEND_MESH_PORT);
	CRIT_ASSERT(bind(sock, (struct sockaddr*)&sin, sizeof(sin)) == 0);

	seqcounter = 0;
	pthread_mutex_init(&seqcounter_lock, NULL);
	peers = NULL;
	peers_size = 0;
	pthread_rwlock_init(&peers_lock, NULL);
}


BackendMesh::~BackendMesh() {
	close(sock);
}


bool
BackendMesh::peers_add(BackendConnection *cn)  {
	int place = -1;

	pthread_rwlock_wrlock(&peers_lock);
	for (unsigned i = 0; i < peers_size; i++)
		if (peers[i] == 0) {
			place = i;
			break;
		}
	if (place == -1) {
		BackendConnection **newpeers;
		newpeers = (BackendConnection**) realloc(peers, peers_size + 1);
		if (newpeers != NULL) {
			peers = newpeers;
			place = peers_size;
		} else {
			pthread_rwlock_unlock(&peers_lock);
			return false;
		}
	}
	peers[place] = cn;
	pthread_rwlock_unlock(&peers_lock);
	return true;
}


unsigned
BackendMesh::peers_len() {
	unsigned count = 0;

	pthread_rwlock_rdlock(&peers_lock);
	for (unsigned i = 0; i < peers_size; i++)
		if (peers[i] != NULL)
			count++;
	pthread_rwlock_unlock(&peers_lock);
	return count;
}


void
BackendMesh::run() {
	int flags = 1, pr;
	struct pollfd pfd;

	mc_log(LOG_INFO, "Backend mesh thread starting");

	for (unsigned i = 0; i < n_backend_peers; i++) {
		mc_log(LOG_INFO, "Adding peer: %s", backend_peers[i]);
	}

	CRIT_ASSERT(listen(sock, 5) == 0);

	pfd.fd = sock;
	pfd.events = POLLIN;

	while (!stop_requested) {	/* The accept() loop */
		int csock = 0;
		struct sockaddr_in csin;
		socklen_t csin_len = sizeof(csin);

		DEBUG_STATUS_POINT(mesh_st, 'a');
		pr = poll(&pfd, 1, 500);
		if (pr == 0) {
			DEBUG_STATUS_POINT(mesh_st, 'A');
			continue;
		} else if (pr < 0) {
			if (stop_requested)
				break;
			continue;
		} else if (pr > 0)
			csock = accept(sock, (struct sockaddr*)&csin, &csin_len);

		if (csock >= 0) {
			/* TODO: prevent duplicate connections */
			CRIT_ASSERT(setsockopt(csock, IPPROTO_TCP, TCP_NODELAY, (char*)&flags, sizeof(flags)) == 0);
			BackendConnection *cconn = new BackendConnection(csock, &sin, this);
			if (peers_add(cconn))
				cconn->start();
			else {
				mc_log(LOG_ERR, "accepted a new peer but cannot add it to the list?");
			}
		} else
			mc_log(LOG_ERR, "accept() returned errno %d (%s)", errno, strerror(errno));

		/* loop through mesh entries, remove dead ones (!running && state ).
		 * Lock opportunistically for the no-op case. */
		pthread_rwlock_rdlock(&peers_lock);
		for (unsigned i = 0; i < peers_size; i++)
			if (peers[i] != NULL && !peers[i]->is_running() && peers[i]->getState() != BC_PERSISTENT_PEER) {
				pthread_rwlock_unlock(&peers_lock);
				pthread_rwlock_wrlock(&peers_lock);
				for (unsigned j = 0; j < peers_size; i++) {
					if (peers[j] != NULL && !peers[j]->is_running()) {
						peers[j]->join();
						if (peers[j]->getState() != BC_PERSISTENT_PEER) {
							delete peers[j];
							peers[j] = NULL;
						}
					}
				}
				break;
			}
		pthread_rwlock_unlock(&peers_lock);
	}
	mc_log(LOG_INFO, "Backend mesh thread exiting; aborting %u peers", peers_len());

	pthread_rwlock_wrlock(&peers_lock);
	for (unsigned i = 0; i < peers_size; i++)
		if (peers[i]->is_running()) {
			peers[i]->abort();
			peers[i]->join();
			delete peers[i];
			peers[i] = NULL;
		}
	mc_log(LOG_DEBUG, "Backend mesh thread exited");
}


/* Pre-shutdown teardown of notifiers */
void
BackendMesh::teardown_locks_notifier() {
	assert(is_notifier_setup);
	assert(_inotif != NULL);
	assert(_tnotif != NULL);
	ds->unset_lock_notifiers();
	delete _inotif;
	delete _tnotif;
	_inotif = NULL;
	_tnotif = NULL;
}


/* ============= calls from DataStore ============= */
/* These calls are invoked in the context of whatever thread the DataStore
 * method is called in, via LockManager, via ItemLocksNotifier, after the local
 * locks are successfully taken. */
enum LOCK_RESULT
BackendMesh::dispatchLockOp(enum LOCK_OP lop, unsigned *nlist, unsigned list_size, enum LOCK_TYPE lt) {
	struct bc_lock_op *loppacket;
	const unsigned psize = sizeof(*loppacket) + sizeof(*nlist) * list_size;
	unsigned bcmd;

	switch (lop) {
	case LOP_RDLOCK:
		bcmd = BCMD_RDLOCK;
		break;
	case LOP_WRLOCK:
		bcmd = BCMD_WRLOCK;
		break;
	case LOP_UNLOCK:
		bcmd = BCMD_UNLOCK;
		break;
	case LOP_TRY_RDLOCK:
		bcmd = BCMD_TRY_RDLOCK;
		break;
	case LOP_TRY_WRLOCK:
		bcmd = BCMD_TRY_WRLOCK;
		break;
	default:
		mc_log(LOG_ERR, "Invalid LOCK_TYPE: %d", lop);
		return LR_DISASTER;
	}

	loppacket = (struct bc_lock_op*) malloc(psize);
	if (loppacket == NULL) {
		mc_log(LOG_ERR, "malloc(%u) failure", psize);
		return LR_DISASTER;
	}
	mc_init_packet(loppacket, psize, get_next_seq(), bcmd);
	loppacket->lock_type = lt;
	loppacket->flags = 0;
	loppacket->n_locks = list_size;
	memcpy(loppacket->nlist, nlist, sizeof(*nlist) * list_size);

	pthread_rwlock_rdlock(&peers_lock);
	int post_status[peers_size], wait_status[peers_size];
	for (unsigned i = 0; i < peers_size; i++)
		post_status[i] = wait_status[i] = -1;

	bool send_rollback = false, wait_rollback = false;
	for (unsigned i = 0; i < peers_size; i++) {
		if (peers[i] == NULL && !peers[i]->is_running())
			continue;
		post_status[i] = peers[i]->postLockOp(loppacket);
		if (post_status[i] != 0) {
			send_rollback = true;
			break;
		}
	}
	pthread_rwlock_unlock(&peers_lock);

	for (unsigned i = 0; i < peers_size; i++)
		if (post_status[i] == 0) {
			if (peers[i] == NULL) /* Unsafe, but avoids holding peers_lock */
				continue;
			wait_status[i] = peers[i]->waitLockOpResponse(loppacket);
			if (wait_status[i] != BCRESP_LOCK_OK)
				wait_rollback = true;
		}


	/* TODO: decide what to rollback */
	if (send_rollback) {
	}
	if (wait_rollback) {
	}
	free(loppacket);

	return LR_DISASTER;
}


/* ============= Lock notifier dispatches (from LockPool from DataStore) ============= */
enum LOCK_RESULT
ItemLocksNotifier::rdlock(unsigned *nlist, unsigned list_size) {
	return bm->dispatchLockOp(LOP_RDLOCK, nlist, list_size, LT_ITEMS);
}


enum LOCK_RESULT
ItemLocksNotifier::wrlock(unsigned *nlist, unsigned list_size) {
	return bm->dispatchLockOp(LOP_WRLOCK, nlist, list_size, LT_ITEMS);
}


enum LOCK_RESULT
ItemLocksNotifier::unlock(unsigned *nlist, unsigned list_size) {
	return bm->dispatchLockOp(LOP_UNLOCK, nlist, list_size, LT_ITEMS);
}


enum LOCK_RESULT
ItemLocksNotifier::try_rdlock(unsigned *nlist, unsigned list_size) {
	return bm->dispatchLockOp(LOP_TRY_RDLOCK, nlist, list_size, LT_ITEMS);
}


enum LOCK_RESULT
ItemLocksNotifier::try_wrlock(unsigned *nlist, unsigned list_size) {
	return bm->dispatchLockOp(LOP_TRY_WRLOCK, nlist, list_size, LT_ITEMS);
}


enum LOCK_RESULT
TagLocksNotifier::rdlock(unsigned *nlist, unsigned list_size) {
	return bm->dispatchLockOp(LOP_RDLOCK, nlist, list_size, LT_TAGS);
}


enum LOCK_RESULT
TagLocksNotifier::wrlock(unsigned *nlist, unsigned list_size) {
	return bm->dispatchLockOp(LOP_WRLOCK, nlist, list_size, LT_TAGS);
}


enum LOCK_RESULT
TagLocksNotifier::unlock(unsigned *nlist, unsigned list_size) {
	return bm->dispatchLockOp(LOP_UNLOCK, nlist, list_size, LT_TAGS);
}


enum LOCK_RESULT
TagLocksNotifier::try_rdlock(unsigned *nlist, unsigned list_size) {
	return bm->dispatchLockOp(LOP_TRY_RDLOCK, nlist, list_size, LT_TAGS);
}


enum LOCK_RESULT
TagLocksNotifier::try_wrlock(unsigned *nlist, unsigned list_size) {
	return bm->dispatchLockOp(LOP_TRY_WRLOCK, nlist, list_size, LT_TAGS);
}
