/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2010. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: BackendMesh.h 347 2011-09-20 20:58:41Z ivoras $
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include "tree.h"
#include "queue.h"
#include "debug.h"
#include "Thread.h"
#include "LockPool.h"
#include "DataStore.h"
#include "mc_protocol.h"

#ifndef _BACKENDMESH_H_
#define _BACKENDMESH_H_

#include <list>
#include <set>
#include <tr1/unordered_map>
#include <exception>
#include "Mutex.h"

#define	MC_BACKEND_MESH_PORT	1333

class BackendMesh;
class BackendConnection;

class ItemLocksNotifier : public LockNotifier {
private:
	BackendMesh *bm;
public:
	ItemLocksNotifier(BackendMesh *bm) { this->bm = bm; }
	virtual enum LOCK_RESULT rdlock(unsigned *nlist, unsigned list_size);
	virtual enum LOCK_RESULT wrlock(unsigned *nlist, unsigned list_size);
	virtual enum LOCK_RESULT unlock(unsigned *nlist, unsigned list_size);
	virtual enum LOCK_RESULT try_rdlock(unsigned *nlist, unsigned list_size);
	virtual enum LOCK_RESULT try_wrlock(unsigned *nlist, unsigned list_size);
};

class TagLocksNotifier : public LockNotifier {
private:
	BackendMesh *bm;
public:
	TagLocksNotifier(BackendMesh *bm) { this->bm = bm; }
	virtual enum LOCK_RESULT rdlock(unsigned *nlist, unsigned list_size);
	virtual enum LOCK_RESULT wrlock(unsigned *nlist, unsigned list_size);
	virtual enum LOCK_RESULT unlock(unsigned *nlist, unsigned list_size);
	virtual enum LOCK_RESULT try_rdlock(unsigned *nlist, unsigned list_size);
	virtual enum LOCK_RESULT try_wrlock(unsigned *nlist, unsigned list_size);
};


class ConnectionException : public std::exception {
private:
	char *msg;
	int _errno;
public:
	ConnectionException(char *msg, int _errno) : exception() {
		this->msg = msg;
		this->_errno = _errno;
	}
};


struct lock_op_wait {
	pthread_cond_t		cv;
	BackendConnection 	*cn;
	int			result;
	void 			*pkt;
};

/* BackendConnection objects can be "alive" but disconnected - in case the remote
 * peer disconnect, we'll keep a place for him. (only if listed under backend_peers) */
enum BC_STATE { BC_REGULAR, BC_PERSISTENT_PEER };

class BackendConnection : public Thread {
public:
#ifndef NDEBUG
	struct debug_status_record bc_st;
#endif
	BackendConnection(int socket, struct sockaddr_in *sin, BackendMesh *bmesh);
	BackendConnection(struct sockaddr_in *sin, enum BC_STATE st, BackendMesh *bmesh);
	~BackendConnection();
	virtual void run();
	void abort();
	int postLockOp(struct bc_lock_op *pkt);
	int waitLockOpResponse(struct bc_lock_op *pkt);
	enum BC_STATE getState() { return state; }
	void setState(enum BC_STATE st) { state = st; }
private:
	enum BC_STATE		state;
	int 			sock;
	Mutex			sock_write_mtx;
	struct sockaddr_in	sin;
	BackendMesh		*bmesh;
	std::set<int>		items_locks;
	std::set<int>		tags_locks;
	std::tr1::unordered_map<unsigned, struct lock_op_wait*> lock_op_wait_list;
	pthread_mutex_t		lock_op_wait_list_lock;
	void _init();
	int processBackendCommand(struct mc_header *rhead);
	void sendCommand(struct mc_header *hcmd) throw (ConnectionException);
	void recordLocks(struct bc_lock_op *lop);
	void recordUnLocks(struct bc_lock_op *lop);
	void emergencyUnlock();
};


/**
 * The locking dance, as seen from the BackendMesh:
 *
 *	The BackendMesh will register its lock notifiers to the DataStore,
 *	which the DataStore will pass to its lock pools.
 *
 *	The DataStore will issue locking operations to the pools, which will
 *	in turn issue callbacks to the lock notifiers *when the local lock
 *	operation completes*. In case of complex trylock operations (i.e. multi-lock
 *	operations), it will do the trylock-and-rollback dance with the locks if
 *	the notifiers indicate failure.
 *
 *	The notifiers will in turn use the BackendMesh to dispatch
 *	the equivalent locks to lock pools on the connected nodes. Success or
 *	failure will follow from these responses.
 */
class BackendMesh : public Thread {
public:
#ifndef NDEBUG
	struct debug_status_record mesh_st;
#endif
	BackendMesh(DataStore *ds);
	~BackendMesh();
	virtual void run();
	
	/* Callback from DataStore */
	enum LOCK_RESULT dispatchLockOp(enum LOCK_OP lop, unsigned *nlist, unsigned list_size, enum LOCK_TYPE lt);
	
	/* Call to DataStore executed from the context of per-backend-connection threads,
	 * called from BackendConnection::run(). The ops behave as "regular" locks -
	 * they block the (remote) callee until the lock can be taken or failure.*/
	enum LOCK_RESULT dataLockOp(enum LOCK_OP lop, unsigned *nlist, unsigned list_size, enum LOCK_TYPE lt) {
		return ds->_internalLockOp(lop, nlist, list_size, lt);
	}
private:
	int 			sock;
	struct sockaddr_in	sin;
	BackendConnection	**peers;
	unsigned int		peers_size;
	pthread_rwlock_t	peers_lock;
	DataStore 		*ds;
	bool			is_notifier_setup;
	ItemLocksNotifier	*_inotif;
	TagLocksNotifier	*_tnotif;
	pthread_mutex_t		seqcounter_lock;
	unsigned short int	seqcounter;
	
	void setup_locks_notifier() {
		if (!is_notifier_setup) {
			_inotif = new ItemLocksNotifier(this);
			_tnotif = new TagLocksNotifier(this);
			ds->set_lock_notifiers(_inotif, _tnotif);
			is_notifier_setup = true;
		}
	}
	
	unsigned short int get_next_seq() {
		unsigned short int seq;
		pthread_mutex_lock(&seqcounter_lock);
		seq = seqcounter++;
		pthread_mutex_unlock(&seqcounter_lock);
		return seq;
	}	
	
	void teardown_locks_notifier();
	
	bool peers_add(BackendConnection *cn);
	unsigned peers_len();
};

#endif
