/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2012. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: DataStore.cpp 482 2012-04-11 14:56:03Z ivoras $
 */

/* Data Storage Internals */

#include <stdlib.h>
#include <sys/fcntl.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <syslog.h>
#include <assert.h>

#ifdef DMALLOC
#include <dmalloc.h>
#endif

#include "tree.h"
#include "ISet.h"
#include "miniatomic.h"

#include "misc.h"
#include "logger.h"
#include "mc_protocol.h"
#include "common_utils.h"
#include "DataStore.h"

/* Generate helper functions for data storage trees */
RB_GENERATE(dh_item_tree, dh_data, linkage, sh_item_cmp);
RB_GENERATE(tag_value_tree, tag_value, linkage, sh_tag_value_cmp);

extern time_t current_time;


/**
 * Fills in the dh_data entry (internal storage structure) from the given mc_data_entry
 * (protocol structure).
 */
void
DataStore::fill_dh_data(struct dh_data *entry, struct mc_data_entry *inpkt)
 {
	/* Cache the inpkt data into the entry */
	entry->tags = (struct mc_tag*)(inpkt->data);
	entry->n_tags = inpkt->n_tags;
	entry->name = inpkt->data + inpkt->n_tags * sizeof(struct mc_tag);
	entry->name_size = inpkt->name_size;
	entry->data = entry->name + entry->name_size;
	entry->data_size = inpkt->data_size;
	entry->mtime = current_time;
#ifdef DEEP_STATS
	entry->stat_n_read_ops = entry->stat_n_write_ops = 0;
#endif
	if (inpkt->exptime == 0)
		entry->exptime = 0x7fffffff;	/* XXX: Pretend this is infinity */
	else if (inpkt->exptime > REL_TIME_CUTOFF)
		entry->exptime = inpkt->exptime;
	else
		entry->exptime = current_time + inpkt->exptime;
	entry->bucket = NULL; /* Don't know it yet */
	entry->inpkt = inpkt;
}


/**
 * This is the principal data structure, the container of all hash tables
 * and the roots of all the trees. It's two main structures.
 */
DataStore::DataStore() throw (MDCacheException)
{
	for (unsigned i = 0; i < N_ITEM_BUCKETS; i++) {
		items[i].index = i;
		RB_INIT(&items[i].tree_head);
		items[i].memory_used = 0;
		items[i].stat_n_write_ops = 0;
#ifdef DEEP_STATS
		items[i].stat_n_read_ops = 0;
#endif
	}
	for (unsigned i = 0; i < N_TAG_BUCKETS; i++) {
		tags[i].index = i;
		tags[i].keys_size = DEFAULT_N_DATA_NODES;
		tags[i].keys = (tag_key*)calloc(DEFAULT_N_DATA_NODES, sizeof(*(tags[i].keys)));
	}
	n_item_buckets = N_ITEM_BUCKETS;
	n_tag_buckets = N_TAG_BUCKETS;
	stat_reap_dead_count = stat_reap_memory_count = 0;
}


/**
 * Destruct.
 */
DataStore::~DataStore()
{
	struct dh_data *var, *nxt;
	struct tag_value *vvar, *vnxt;

	for (unsigned i = 0; i < N_ITEM_BUCKETS; i++) {
		items_locks.wrlock(i);
		for (var = RB_MIN(dh_item_tree, &items[i].tree_head); var != NULL; var = nxt) {
			nxt = RB_NEXT(dh_item_tree, &items[i].tree_head, var);
			RB_REMOVE(dh_item_tree, &items[i].tree_head, var);
			if (var->inpkt != NULL)
				free(var->inpkt);
			free(var);
		}
		items_locks.unlock(i);
	}
	for (unsigned i = 0; i < N_TAG_BUCKETS; i++) {
		tags_locks.wrlock(i);
		for (unsigned j = 0; j < tags[i].keys_size; j++) {
			for (vvar = RB_MIN(tag_value_tree, &tags[i].keys[j].tree_head); vvar != NULL; vvar = vnxt) {
				vnxt = RB_NEXT(tag_value_tree, &tags[i].keys[j].tree_head, vvar);
				RB_REMOVE(tag_value_tree, &tags[i].keys[j].tree_head, vvar);
				if (vvar->items_size != 0)
					free(vvar->vitems);
				free(vvar);
			}
		}
		free(tags[i].keys);
		tags_locks.unlock(i);
	}
	/* Autodestruct */
	memset(items, 0, sizeof(items));
	memset(tags, 0, sizeof(tags));
	n_item_buckets = n_tag_buckets = 0;
}


void
DataStore::set_lock_notifiers(LockNotifier *ilock, LockNotifier *tlock) {
	items_locks.set_lock_notifier(ilock);
	tags_locks.set_lock_notifier(tlock);
}


void
DataStore::unset_lock_notifiers() {
	items_locks.unset_lock_notifier();
	tags_locks.unset_lock_notifier();
}


/* An internal lock request, not to be dispatched to BackendMesh */
enum LOCK_RESULT
DataStore::_internalLockOp(enum LOCK_OP lop, unsigned *nlist, unsigned list_size, enum LOCK_TYPE lt) {
	if (lt == LT_ITEMS)
		return items_locks._internalLockOp(lop, nlist, list_size);
	else if (lt == LT_TAGS)
		return tags_locks._internalLockOp(lop, nlist, list_size);
	else {
		mc_log(LOG_ERR, "Unknown LOCK_TYPE: %d", lt);
		return LR_DISASTER;
	}
}

/**
 * A debugging function, returns true if all the locks in the lock pools
 * are free. (Does not hold locks when returned).
 */
bool
DataStore::allLocksClear() {
	return items_locks.allClear() && tags_locks.allClear();
}


/**
 * Dump the cache database contents into the "prewarm" file.
 * If so desired, the dump will be database-atomic (all item buckets read-locked).
 */
bool
DataStore::prewarm_cache_dump(char *filename, bool atomic)
{
	FILE *f;
	int count = 0;

	f = fopen(filename, "w+b");
	if (f == NULL) {
		mc_log(LOG_DEBUG, "Error creating cache dump file '%s': %s", filename, strerror(errno));
		return false;
	}
	setvbuf(f, NULL, _IOFBF, 65536);

	struct mdcached_prewarm_file_header hdr = {
		PREWARM_SIGNATURE,
		PREWARM_FORMAT,
		0,
		N_ITEM_BUCKETS,
		N_TAG_BUCKETS
	};

	if (fwrite(&hdr, sizeof(hdr), 1, f) != 1) {
		mc_log(LOG_ERR, "Error writing header to cache dump file %s", filename);
		fclose(f);
	}

	bool ok = true;
	if (atomic) {
		unsigned buckets[N_ITEM_BUCKETS];
		for (unsigned i = 0; i < N_ITEM_BUCKETS; i++)
			buckets[i] = i;
		items_buckets_rdlock(buckets, N_ITEM_BUCKETS);
		for (unsigned i = 0; i < N_ITEM_BUCKETS; i++) {
			int bcount = prewarm_cache_dump_bucket(f, i);
			if (bcount < 0)
				ok = false;
			else
				count += bcount;
		}
		items_buckets_unlock(buckets, N_ITEM_BUCKETS);
	} else {
		for (unsigned i = 0; i < N_ITEM_BUCKETS; i++) {
			items_buckets_rdlock(&i, 1);
			int bcount = prewarm_cache_dump_bucket(f, i);
			if (bcount < 0)
				ok = false;
			else
				count += bcount;
			items_buckets_unlock(&i, 1);
		}
	}
	if (!ok) {
		mc_log(LOG_ERR, "Aborting cache dump to %s", filename);
		fclose(f);
		return false;
	}

	if (fseek(f, 0, SEEK_SET) != 0) {
		mc_log(LOG_ERR, "lseek() failure - aborting cache dump to %s", filename);
		fclose(f);
		return false;
	}
	/* Mark the file as sane, rewrite header */
	hdr.flags |= PREWARM_FLAG_SANE;
	hdr.n_records = count;
	ok = fwrite(&hdr, sizeof(hdr), 1, f) == 1;
	fclose(f);
	return ok;
}


/**
 * Internal function - dump a single cache bucket.
 * Assumes the bucket is at least read-locked.
 */
int
DataStore::prewarm_cache_dump_bucket(FILE *f, unsigned bi)
{
	struct dh_items_bucket * const ib = &items[bi];
	struct dh_data *d;
	int count = 0;

	RB_FOREACH(d, dh_item_tree, &ib->tree_head) {
		int writ = fwrite(d->inpkt, d->inpkt->header.total_size, 1, f);
		if (writ != 1) {
			mc_log(LOG_ERR, "fwrite() error %s (%d)", strerror(errno), errno);
			return -1;
		}
		count++;
	}
	return count;
}


/**
 * Reads the prewarm cache dump.
 */
bool
DataStore::prewarm_cache_read(char *filename, bool atomic)
{
	FILE *f = fopen(filename, "r+b");
	if (f == NULL) {
		mc_log(LOG_DEBUG, "Error opening cache dump file '%s': %s", filename, strerror(errno));
		return false;
	}

	setvbuf(f, NULL, _IOFBF, 65536);
	struct mdcached_prewarm_file_header hdr;
	if (fread(&hdr, sizeof(hdr), 1, f) != 1) {
		mc_log(LOG_ERR, "Cannot read header from the cache dump file %s", filename);
		fclose(f);
		return false;
	}
	if (hdr.signature != PREWARM_SIGNATURE) {
		mc_log(LOG_ERR, "Header signature mismatch for cache dump file %s", filename);
		fclose(f);
		return false;
	}
	if (hdr.format_version > PREWARM_FORMAT) {
		mc_log(LOG_ERR, "Header format version mismatch for cache dump file %s", filename);
		fclose(f);
		return false;
	}
	if ((hdr.flags & PREWARM_FLAG_SANE) == 0) {
		mc_log(LOG_ERR, "Header says its not sane for cache dump file %s (file corrupted or incomplete)", filename);
		fclose(f);
		return false;
	}
	if (hdr.n_item_buckets != N_ITEM_BUCKETS || hdr.n_tag_buckets != N_TAG_BUCKETS) {
		mc_log(LOG_ERR, "Cache dump file header geometry doesn't match server: %d,%d should be %d,%d",
			hdr.n_item_buckets, hdr.n_tag_buckets, N_ITEM_BUCKETS, N_TAG_BUCKETS);
		fclose(f);
		return false;
	}

	if (atomic) {
		unsigned buckets[N_ITEM_BUCKETS];
		for (unsigned i = 0; i < N_ITEM_BUCKETS; i++)
			buckets[i] = i;
		items_buckets_wrlock(buckets, N_ITEM_BUCKETS);
		/* TODO: tag_buckets_wrlock() */
		for (unsigned i = 0; i < hdr.n_records; i++)
			prewarm_cache_read_put_record(f, false);
		items_buckets_unlock(buckets, N_ITEM_BUCKETS);
	} else {
		for (unsigned i = 0; i < hdr.n_records; i++)
			prewarm_cache_read_put_record(f, true);
	}

	fclose(f);
	return true;
}


/**
 * Internal function - read and put a single dumped record.
 */
bool
DataStore::prewarm_cache_read_put_record(FILE *f, bool do_lock)
{
	struct mc_header rhdr;

	if (fread(&rhdr, sizeof(rhdr), 1, f) != 1) {
		mc_log(LOG_ERR, "Error reading record header from cache dump file");
		return false;
	}

	struct mc_data_entry *mce = (struct mc_data_entry*)malloc(rhdr.total_size);
	memcpy(mce, &rhdr, sizeof(rhdr));
	char *data = (char*)mce + sizeof(rhdr);
	if (fread(data, rhdr.total_size - sizeof(rhdr), 1, f) != 1) {
		mc_log(LOG_ERR, "Error reading record from cache dump file");
		return false;
	}

	struct dh_data *d = (struct dh_data*)malloc(sizeof *d);
	DataStore::fill_dh_data(d, mce);

	if (do_lock)
		put(d);
	else
		put_nolock(NULL, &d);
	return true;
}


/**
 * Add a record from the simple information provided. Expects appropriate locks
 * to be held.
 */
struct dh_data *
DataStore::internal_add_record_nolock(const char *name, unsigned name_size, const char *data, unsigned data_size, const struct mc_tag *tags, unsigned n_tags)
{
	struct dh_data *d;
	struct mc_data_entry *md = NULL;
	int res;

	if ((res = mc_build_data_entry(&md, name, name_size, data, data_size, tags, n_tags, 0, NULL)) != MCREASON_OK) {
		mc_log(LOG_ERR, "mc_build_data_entry failed: %d", res);
		return NULL;
	}
	d = (struct dh_data*)malloc(sizeof(*d));
	DataStore::fill_dh_data(d, md);
	if (put_nolock(NULL, &d) != 0) {
		free(d);
		return NULL;
	}

	return d;
}


/**
 * Test if the data item is still "alive" - accessible to the clients. This
 * factors in data expiry.
 */
bool
DataStore::item_is_alive(struct dh_data *d)
{
	return (d->exptime > current_time);
}


/**
 * Perform a "complex" lock op, with the rollback-and-retry protocol.
 * Can return false if a critical error happens; if returns true it means
 * all the locks have been acquired successfuly.
 */
bool
DataStore::data_item_safe_lock_op(struct dh_data *d, enum LOCK_OP lop, bool doitem, bool dotags)
{
	unsigned counter = 0;
	unsigned ntb = 0;
	int lr = 0;
	unsigned ti;

	assert(lop == LOP_RDLOCK || lop == LOP_WRLOCK || lop == LOP_UNLOCK);
	assert(doitem || dotags);
	assert(d != NULL);
	assert(d->name != NULL);

#ifdef LOCK_TRACE_ENABLE
	mc_log(LOG_DEBUG, "%%data_item_safe_lock_op(%p, %d, %d, %d)", d, lop, doitem, dotags);
#endif

	unsigned tbuckets[d->n_tags];
	const unsigned ib_index = d->bucket != NULL ? d->bucket->index : ITEM_BUCKET_INDEX(d->name, d->name_size);

	/* Make tbuckets contain unique tag bucket indexes */
	if (dotags) {
		ISet<N_TAG_BUCKETS> tbset;
		for (unsigned i = 0; i < d->n_tags; i++)
			tbset.add(TAG_BUCKET_INDEX(d->tags[i].key));
		tbset.get_elements(tbuckets);
	}

retry:
	if (++counter % 1000 == 0)
		mc_log(LOG_WARNING, "data_item_safe_lock_op retry counter is large: %u", counter);

	if (likely(doitem)) {
		switch (lop) {
		case LOP_RDLOCK:
			lr = items_locks.rdlock(ib_index);
			break;
		case LOP_WRLOCK:
			lr = items_locks.wrlock(ib_index);
			break;
		case LOP_UNLOCK:
			lr = items_locks.unlock(ib_index);
			break;
		default:
			mc_log(LOG_CRIT, "Unknown lop: %d", lop);
		}
		if (lr != LR_OK) {
			mc_log(LOG_ERR, "items_locks op (%d) failed in data_item_safe_lock_op", lop);
			return false; /* This is a blocking op. If it fails, it's serious. */
		}
	}

	if (!dotags)
		goto end;

	/* Perform tag lock operations */
	for (unsigned i = 0; i < ntb; i++) {
		ti = tbuckets[i];

		switch (lop) {
		case LOP_RDLOCK:
			lr = tags_locks.try_rdlock(ti);
			break;
		case LOP_WRLOCK:
			lr = tags_locks.try_wrlock(ti);
			break;
		case LOP_UNLOCK:
			lr = tags_locks.unlock(ti);
			break;
		default:
			mc_log(LOG_CRIT, "Unknown lop: %d", lop);
		}

		if (lr != LR_OK && lop != LOP_UNLOCK) {
			mc_log(LOG_DEBUG, "data_item_safe_lock_op: rollback&retry: %d on %d", lr, ti);
			/* Rollback on failure */
			for (unsigned j = 0; j < i; j++)
				tags_locks.unlock(TAG_BUCKET_INDEX(d->tags[j].key));
			if (doitem)
				items_locks.unlock(ib_index);
			pthread_yield();
			/* TODO: sleep? */
			goto retry;
		} else if (lr != LR_OK && lop == LOP_UNLOCK) {
			mc_log(LOG_CRIT, "tag lock op failed (%d) on LOP_UNLOCK!", lr);
		}
	}
end:
#ifdef LOCK_TRACE_ENABLE
	mc_log(LOG_DEBUG, "%%data_item_safe_lock_op(%p, %d, %d, %d) done", d, lop, doitem, dotags);
#endif
	return true;
}


/**
 * Put an item into the cache database. This also takes care of the (any eventual)
 * tags and does full locking.
 */
int
DataStore::put(struct dh_data *d)
{
	const unsigned ib_index = ITEM_BUCKET_INDEX(d->name, d->name_size);
	struct dh_items_bucket * const ib = &items[ib_index];
	int res;

retry:
	data_item_safe_lock_op(d, LOP_WRLOCK, true, false);
	res = put_nolock(ib, &d);
	data_item_safe_lock_op(d, LOP_UNLOCK, true, false);
	if (res == LR_TRYAGAIN) {
		pthread_yield();
		goto retry;
	}

	return TRUE;
}


/**
 * Puts an item into the cache database like put() but does no locking.
 * Assumes the correct item bucket is already locked for writing, but tries to
 * lock tag buckets itself.
 *
 * Will return 0 or fail with LR_TRYAGAIN if tag locks cannot be acquired (in
 * which case the caller MUST release the item lock, yield, re-acquire it and
 * try again. This is because put_nolock can replace an item but the details
 * of tag locking are hairy. 
 */
int inline
DataStore::put_nolock(struct dh_items_bucket *ib, struct dh_data **d)
{
	struct dh_data *d0;
	ISet<N_TAG_BUCKETS> tbset_new;
	unsigned tb_new[(*d)->n_tags];

	if (ib == NULL)
		ib = &items[ITEM_BUCKET_INDEX((*d)->name, (*d)->name_size)];

	/* In b4, see if we can acquire tag bucket locks for the new record */
	if ((*d)->n_tags > 0) {
		for (unsigned i = 0; i < (*d)->n_tags; i++)
			tbset_new.add(TAG_BUCKET_INDEX((*d)->tags[i].key));
		tbset_new.get_elements(tb_new);
		if (tags_locks.try_wrlock(tb_new, tbset_new.count()) != LR_OK)
			return LR_TRYAGAIN;
	}
	
	ib->stat_n_write_ops++;
	d0 = RB_INSERT(dh_item_tree, &ib->tree_head, *d);
	if (likely(d0 == NULL)) {
		/* INSERT succeded, a unique item was put. */
		d0 = *d;
	} else {
		/*
		 * Insert failed because of uniqueness conflict, so overwrite
		 * the existing node with the new one. This gets very much
		 * complicated with the need to maintain tags.
		 */
		 
		/* Step 1: Acquire tag locks for tags present in the old record
		 * but not in the new record. */
		ISet<N_TAG_BUCKETS> tbset_add_old;
		for (unsigned i = 0; i < d0->n_tags; i++)
			if (!tbset_new.test(TAG_BUCKET_INDEX(d0->tags[i].key)))
				tbset_add_old.add(TAG_BUCKET_INDEX(d0->tags[i].key));
		
		unsigned tbuckets[d0->n_tags];
		tbset_add_old.get_elements(tbuckets);
		if (tags_locks.try_wrlock(tbuckets, tbset_add_old.count()) != LR_OK) {
			/* Rollback */
			tags_locks.unlock(tb_new, tbset_new.count());
			return LR_TRYAGAIN;
		}
		
		/* Step 2: Remove tag links for the old record */
		for (unsigned i = 0; i < d0->n_tags; i++)
			tag_unlink(d0, &d0->tags[i]);
		
		/* Step 3: Unlock tags which were present in the old record but
		 * not in the new record, leaving only the new record's tag
		 * locks acquired. */
		tags_locks.unlock(tbuckets, tbset_add_old.count());
		
		/* Step 4: Replace the old record with the new one. */
		data_cleanup(d0);	/* Remove existing data */
		d0->inpkt = (*d)->inpkt;	/* Linkup new data */
		d0->name = (*d)->name;
		d0->name_size = (*d)->name_size;
		d0->data_size = (*d)->data_size;
		d0->n_tags = (*d)->n_tags;
		d0->data = (*d)->data;
		d0->tags = (*d)->tags;
		d0->exptime = (*d)->exptime;
		free(*d);		/* Free the failed insertion */
		*d = d0;		/* Keep track of it */
	}
	d0->mtime = current_time;
	d0->bucket = ib;
	ib->memory_used += d0->inpkt->header.total_size;
#ifdef DEEP_STATS
	d0->stat_n_write_ops++;
#endif

	if (d0->n_tags > 0) {
		for (unsigned i = 0; i < d0->n_tags; i++)
			tag_link(d0, &d0->tags[i]);
		tags_locks.unlock(tb_new, tbset_new.count());
	}

	return 0;
}


/**
 * Retrieve a raw copy of the item identified by its name, to be freed by the
 * caller.
 */
struct mc_data_entry*
DataStore::get_data(char *name, unsigned int name_size)
{
	const unsigned ib_index = ITEM_BUCKET_INDEX(name, name_size);
	struct dh_items_bucket * const ib = &items[ib_index];
	struct dh_data search;
	struct dh_data *d;
	struct mc_data_entry *pkt;

	search.name = name;
	search.name_size = name_size;

	if (unlikely(items_locks.rdlock(ib_index) != LR_OK))
		return NULL;

	d = RB_FIND(dh_item_tree, &ib->tree_head, &search);
	if (unlikely(d == NULL || d->inpkt == NULL || !item_is_alive(d))) {
		items_locks.unlock(ib_index);
		return NULL;
	}
	pkt = (struct mc_data_entry*) malloc(d->inpkt->header.total_size);
	if (unlikely(pkt == NULL)) {
		mc_log(LOG_DEBUG, "malloc failure while creating a copy of the packet");
		items_locks.unlock(ib_index);
		return NULL;
	}
	/* Since d could be changed or deleted by other threads, return a copy. */
#ifdef DEEP_STATS
	atomic_add_32(&d->stat_n_read_ops, 1);
#endif
	memcpy(pkt, d->inpkt, d->inpkt->header.total_size);
	if (unlikely(items_locks.unlock(ib_index) != LR_OK)) {
		free(pkt);
		return NULL;
	}
	return pkt;
}


/**
 * Retrieve a direct pointer to the item identified by its name. The items bucket
 * is left in read-locked state until got_data_unlock() is called. This pair of
 * functions supports very efficient, malloc()- and memcpy()- free get operations on a
 * single key/value pair.
 */
struct mc_data_entry*
DataStore::get_data_and_lock(char *name, unsigned int name_size, unsigned int *lock_cookie)
{
	const unsigned ib_index = ITEM_BUCKET_INDEX(name, name_size);
	struct dh_items_bucket * const ib = &items[ib_index];
	struct dh_data search;
	struct dh_data *d;

	search.name = name;
	search.name_size = name_size;

	if (unlikely(items_locks.rdlock(ib_index) != LR_OK)) {
		mc_log(LOG_ERR, "items_locks.rdlock(%d) failure", ib_index);
		return NULL;
	}

	d = RB_FIND(dh_item_tree, &ib->tree_head, &search);
	if (unlikely(d == NULL || d->inpkt == NULL || !item_is_alive(d))) {
		items_locks.unlock(ib_index);
		return NULL;
	}
	/* Leave the bucket locked */
	*lock_cookie = ib_index;
#ifdef LOCK_TRACE_ENABLED
	mc_log(LOG_DEBUG, "%%Returning from get_data_and_lock with bucket %d locked", ib_index);
#endif
#ifdef DEEP_STATS
	atomic_add_32(&d->stat_n_read_ops, 1);
#endif
	return d->inpkt;
}


/**
 * Find a return an item without any locking operations. Assumes the correct
 * item bucket is already locked for reading.
 */
inline struct mc_data_entry*
DataStore::get_data_nolock(char *name, unsigned int name_size)
{
	const unsigned ib_index = ITEM_BUCKET_INDEX(name, name_size);
	struct dh_items_bucket * const ib = &items[ib_index];
	struct dh_data search;
	struct dh_data *d;

	search.name = name;
	search.name_size = name_size;

	d = RB_FIND(dh_item_tree, &ib->tree_head, &search);
	if (unlikely(d == NULL || d->inpkt == NULL || !item_is_alive(d)))
		return NULL;
#ifdef DEEP_STATS
	atomic_add_32(&d->stat_n_read_ops, 1);
#endif
	return d->inpkt;
}


/**
 * The second part of the get_data_and_lock / send / got_data_unlock efficient
 * data sending idiom.
 */
void
DataStore::got_data_unlock(unsigned int lock_cookie)
{
#ifdef LOCK_TRACE_ENABLED
	mc_log(LOG_DEBUG, "%%Unlocking bucket %d", lock_cookie);
#endif
	items_locks.unlock(lock_cookie);
}


/**
 * Adds a tag to the correct tables (if applicable) and link it to the specified
 * data item. Fails silently on error. Expects the struct dh_data item to be at least
 * read-locked and the tag bucket write-locked.
 *
 * See the description in DataStore.h for on how the tags are stored.
 */
bool
DataStore::tag_link(struct dh_data *d, struct mc_tag *t)
{
	const unsigned tb_index = TAG_BUCKET_INDEX(t->key);
	struct dh_tags_bucket * const tb = &tags[tb_index];
	int fk = -1, fv = -1;
	struct tag_key *tk;
	struct tag_value *tv, tv_find;

	assert(d != NULL);
	assert(d->name != NULL);

	/* Examine the tag bucket in search for the actual key. */
	/* TODO: replace with a sorted array and binary search if the keyspace becomes large */
	for (unsigned i = 0; i < tb->keys_size; i++) {
		if (tb->keys[i].key == t->key) {
			fk = i;
			break;
		}
		if (tb->keys[i].tree_head.rbh_root == NULL) {
			fk = i;
			tb->keys[i].key = t->key;
			break;
		}
	}
	/*
	printf("After searching for key=%d, fk=%d\n", t->key, fk);
	log_hexdump(tb->keys, sizeof(*(tb->keys))*tb->keys_size, 8);
	*/
	if (unlikely(fk == -1)) {
		/* Append a new struct tag_key and initalize its value tree */
		struct tag_key *new_keys;

		fk = tb->keys_size++;
		new_keys = (struct tag_key*) realloc(tb->keys, tb->keys_size * sizeof(*(tb->keys)));
		if (unlikely(new_keys == NULL)) {
			tb->keys_size--;
			mc_log(LOG_ERR, "realloc() failed for %d bytes", tb->keys_size * sizeof(*(tb->keys)));
			return false;
		}
		/* fk now points to the newly created space at the end of new_keys */
		RB_INIT(&new_keys[fk].tree_head);
		new_keys[fk].key = t->key;
		if (tb->keys != new_keys) {
			/*
			 * Since the RB tree stores nodes' parent pointers, we need to make this
			 * fixup after realloc() because realloc() can move the allocated block.
			 */
			mc_log(LOG_DEBUG, "tag_link() realloc fixup");
			for (int i = 0; i < fk; i++) {
				/* Simply make parent pointers in the first left and right children nodes
				 * point to the correct (new) root. */
				assert(RB_ROOT(&new_keys[i].tree_head) != NULL);
				if (RB_LEFT(RB_ROOT(&new_keys[i].tree_head), linkage) != NULL)
					RB_PARENT(RB_LEFT(RB_ROOT(&new_keys[i].tree_head), linkage), linkage) = RB_ROOT(&new_keys[i].tree_head);
				if (RB_RIGHT(RB_ROOT(&new_keys[i].tree_head), linkage) != NULL)
					RB_PARENT(RB_RIGHT(RB_ROOT(&new_keys[i].tree_head), linkage), linkage) = RB_ROOT(&new_keys[i].tree_head);
			}
			tb->keys = new_keys;
		}
	}
	/* Add it to the found or created value tree indexed by fk */
	tv_find.value = t->value;
	tk = &tb->keys[fk];
	tv = RB_FIND(tag_value_tree, &tk->tree_head, &tv_find);
	if (unlikely(tv == NULL)) {
		/* A new struct tag_value node needs to be added to the tree */
		tv = (struct tag_value*) calloc(1, sizeof(*tv));
		if (unlikely(tv == NULL))
			return false;
		tv->value = t->value;
		tv->items_size = DEFAULT_N_DATA_NODES;
		tv->vitems = (struct dh_item_value_entry*) calloc(DEFAULT_N_DATA_NODES, sizeof(*tv->vitems));
		assert(RB_INSERT(tag_value_tree, &tk->tree_head, tv) == NULL);
	}
	for (unsigned i = 0; i < tv->items_size; i++) {
		if (likely(tv->vitems[i].item == NULL)) {
			fv = i;
			break;
		} else if (unlikely(tv->vitems[i].item == d)) {
			/* Already there? */
			return true;
		}
	}
	if (unlikely(fv == -1)) {
		/* No free spot in the items pointers array - resize it */
		struct dh_item_value_entry *new_vitems;

		fv = tv->items_size++;
		new_vitems = (struct dh_item_value_entry*) realloc(tv->vitems, tv->items_size * sizeof(*tv->vitems));
		if (unlikely(new_vitems == NULL)) {
			tv->items_size--;
			return false;
		}
		tv->vitems = new_vitems;
	}
	tv->vitems[fv].item = d;
	tv->vitems[fv].bucket = d->bucket;
	return true;
}


/**
 * Remove the links between the given struct dh_data item and the given tag
 * (if present). Assumes the struct dh_data item is at least rdlocked and
 * that the tag bucket is rwlocked.
 */
bool
DataStore::tag_unlink(struct dh_data *d, struct mc_tag *t)
{
	const unsigned tb_index = TAG_BUCKET_INDEX(t->key);
	struct dh_tags_bucket * const tb = &tags[tb_index];
	struct tag_key *tk;
	struct tag_value *tv, tv_find;
	int fk = -1;

	tv_find.value = t->value;

	for (unsigned i = 0; i < tb->keys_size; i++) {
		if (tb->keys[i].key == t->key) {
			fk = i;
			break;
		}
	}
	if (unlikely(fk == -1)) {
		/* Ignore missing data */
		return true;
	}
	tk = &tb->keys[fk];
	tv = RB_FIND(tag_value_tree, &tk->tree_head, &tv_find);
	if (unlikely(tv == NULL)) {
		/* Ignore missing data */
		return true;
	}
	for (unsigned i = 0; i < tv->items_size; i++)
		if (tv->vitems[i].item == d) {
			tv->vitems[i].item = NULL;
			tv->vitems[i].bucket = NULL;
			break;
		}
	return true;
}

/**
 * Retrieves items containing the specified key and matching any of the given
 * values. Returns struct mc_resp_multidata* to be freed by the caller.
 */
struct mc_resp_multidata*
DataStore::get_multidata_by_tag_values(const tag_key_t tkey, tag_value_t *tvalues, const unsigned int values_count)
{
	const unsigned tb_index = TAG_BUCKET_INDEX(tkey);
	struct dh_tags_bucket * const tb = &tags[tb_index];
	struct tag_key *tk = NULL;
	struct tag_value *tv, tv_search;
	struct dh_data *d;
	struct mc_resp_multidata *result;
	unsigned *buckets;
	unsigned n_buckets;
	std::set<struct dh_data*> iset;

	tags_locks.rdlock(tb_index);
	/* Scan through the keys of this tag bucket */
	for (unsigned i = 0; i < tb->keys_size; i++) {
		if (tb->keys[i].key == tkey) {
			tk = &tb->keys[i];
			break;
		}
	}
	if (tk == NULL) {
		/* Key not found */
		tags_locks.unlock(tb_index);
		return NULL;
	}

	/* Find data item buckets containing the given key and values */
	if (!get_items_buckets_by_tag_key_values(tk, tvalues, values_count, &buckets, &n_buckets)) {
		tags_locks.unlock(tb_index);
		return NULL;
	}

	if (!items_buckets_rdlock(buckets, n_buckets)) {
		mc_log(LOG_ERR, "Cannot rdlock items buckets!");
		tags_locks.unlock(tb_index);
		free(buckets);
		return NULL;
	}

	result = (struct mc_resp_multidata*) calloc(1, sizeof(*result));
	result->header.total_size = sizeof(*result);

	for (unsigned i = 0; i < values_count; i++) {
		tv_search.value = tvalues[i];
		tv = RB_FIND(tag_value_tree, &tk->tree_head, &tv_search);
		if (tv == NULL)
			continue;
		for (unsigned j = 0; j < tv->items_size; j++) {
			if (tv->vitems[j].item == NULL)
				continue;
			assert(tv->vitems[j].item != NULL && tv->vitems[j].bucket != NULL);
			if (!item_is_alive(tv->vitems[j].item))
				continue;
			std::pair<std::set<struct dh_data*>::iterator, bool> inserted = iset.insert(tv->vitems[j].item);
			if (!inserted.second)
				continue; /* Do not return duplicates */
			d = tv->vitems[j].item;
#ifdef DEEP_STATS
			atomic_add_32(&d->stat_n_read_ops, 1);
#endif
			if (!multidata_append(&result, d->inpkt)) {
				mc_log(LOG_ERR, "multidata_append() failed");
				free(result);
				free(buckets);
				tags_locks.unlock(tb_index);
				return NULL;
			}
		}
	}
	items_buckets_unlock(buckets, n_buckets);
	tags_locks.unlock(tb_index);
	free(buckets);
	if (result->n_records > 0)
		return result;
	else {
		free(result);
		return NULL;
	}
}


/**
 * Append a struct mc_data_entry item to the multidata response.
 * Note: this will mess up structure alignments of the resulting packet!
 */
bool
DataStore::multidata_append(struct mc_resp_multidata **md, struct mc_data_entry *item)
{
	struct mc_resp_multidata *newmd;

	assert(*md != NULL);
	assert(item != NULL);
	newmd = (struct mc_resp_multidata*) realloc(*md, (*md)->header.total_size + item->header.total_size);
	if (newmd != NULL) {
		memcpy(&(newmd->data[newmd->header.total_size - sizeof(*newmd)]), item, item->header.total_size);
		newmd->n_records++;
		newmd->header.total_size += item->header.total_size;
		*md = newmd;
		return true;
	} else {
		mc_log(LOG_ERR, "multidata_append(): cannot realloc buffer to %d", (int)((*md)->header.total_size + item->header.total_size));
		return false;
	}
}


/**
 * Retrieve all the record keys (key names) stored in the cache. Depending
 * on the flags this may happen with all buckets locked or one by one.
 */
struct mc_resp_multistring *
DataStore::get_all_keys(const unsigned flags)
{
	struct mc_resp_multistring *result;
	struct dh_data *d;

	result = (struct mc_resp_multistring*) calloc(1, sizeof(*result));
	result->header.total_size = sizeof(*result);

	if (flags & MCMD_FLAG_ATOMICALL) {
		unsigned buckets[n_item_buckets];

		for (unsigned i = 0; i < n_item_buckets; i++)
			buckets[i] = i;
		items_buckets_rdlock(buckets, n_item_buckets);

		for (unsigned i = 0; i < n_item_buckets; i++) {
			RB_FOREACH(d, dh_item_tree, &items[i].tree_head) {
				if (!item_is_alive(d))
					continue;
				if (!multistring_append(&result, d->name, d->name_size)) {
					items_buckets_unlock(buckets, n_item_buckets);
					free(result);
					mc_log(LOG_ERR, "multistring_append(%p,%u) failed", d->name, d->name_size);
					return NULL;
				}
			}
		}
		items_buckets_unlock(buckets, n_item_buckets);
	} else {
		for (unsigned i = 0; i < n_item_buckets; i++) {
			items_locks.rdlock(i);
			RB_FOREACH(d, dh_item_tree, &items[i].tree_head) {
				if (!item_is_alive(d))
					continue;
				if (!multistring_append(&result, d->name, d->name_size)) {
					items_locks.unlock(i);
					free(result);
					mc_log(LOG_ERR, "multistring_append(%p,%u) failed", d->name, d->name_size);
					return NULL;
				}
			}
			items_locks.unlock(i);
		}
	}
	if (result->n_records > 0)
		return result;
	else {
		free(result);
		return NULL;
	}
}


/**
 * Append a string to the multistring response packet. The string length is
 * limited to unsigned short int as the function is intended to be used with
 * record keys (key names).
 */
bool
DataStore::multistring_append(struct mc_resp_multistring **ms, const char *str, const unsigned short int str_size)
{
	struct mc_resp_multistring *newms;

	newms = (struct mc_resp_multistring*) realloc(*ms, (*ms)->header.total_size + sizeof(str_size) + str_size);
	if (newms != NULL) {
		char *data = newms->data + newms->data_size;
		
		memcpy(data, &str_size, sizeof(str_size));
		memcpy(data + sizeof(str_size), str, str_size);
		
		newms->data_size += sizeof(str_size) + str_size;
		newms->header.total_size += sizeof(str_size) + str_size;
		newms->n_records++;
		*ms = newms;
		return true;
	} else {
		mc_log(LOG_ERR, "multistring_append(): cannot realloc buffer to %d", (int)((*ms)->header.total_size + str_size + sizeof(str_size)));
		return false;
	}
}


/**
 * Delete a data item identified by name, while locking item and tags.
 */
int
DataStore::del_data_name(const char *name, const unsigned int name_size)
{
	const unsigned ib_index = ITEM_BUCKET_INDEX(name, name_size);
	struct dh_items_bucket * const ib = &items[ib_index];
	struct dh_data search;
	struct dh_data *d;

	search.name = (char*)name;
	search.name_size = name_size;

retry:
	items_locks.wrlock(ib_index);
	d = RB_FIND(dh_item_tree, &ib->tree_head, &search);
	if (unlikely(d == NULL || d->inpkt == NULL)) {
		items_locks.unlock(ib_index);
		return FALSE;
	}
	unsigned tbuckets[d->n_tags];
	ISet<N_TAG_BUCKETS> tbset;
	for (unsigned i = 0; i < d->n_tags; i++)
		tbset.add(TAG_BUCKET_INDEX(d->tags[i].key));
	tbset.get_elements(tbuckets);
	if (tags_locks.try_wrlock(tbuckets, tbset.count()) != LR_OK) {
		items_locks.unlock(ib_index);
		pthread_yield();
		goto retry;
	}
	
	remove_item_and_free_nolock(d);
	items_locks.unlock(ib_index);
	tags_locks.unlock(tbuckets, tbset.count());
	return TRUE;
}


/**
 * Delete a data item identified by name, without using locking operations. Expects
 * that the appropriate item lock and tag locks is already held.
 */
int
DataStore::del_data_name_nolock(const char *name, const unsigned int name_size)
{
	const unsigned ib_index = ITEM_BUCKET_INDEX(name, name_size);
	struct dh_items_bucket * const ib = &items[ib_index];
	struct dh_data search;
	struct dh_data *d;

	search.name = (char*)name;
	search.name_size = name_size;

	d = RB_FIND(dh_item_tree, &ib->tree_head, &search);
	if (unlikely(d == NULL || d->inpkt == NULL))
		return FALSE;
	remove_item_and_free_nolock(d);
	return TRUE;
}


/**
 * Remove the given dh_data item, expecting the item lock and tag locks to be held.
 */
int
DataStore::del_nolock(struct dh_items_bucket *ib, struct dh_data *d)
{
	if (unlikely(ib == NULL)) {
		const unsigned ib_index = ITEM_BUCKET_INDEX(d->name, d->name_size);
		ib = &items[ib_index];
	}
	remove_item_and_free_nolock(d);
	return TRUE;
}


/**
 * Delete data items identified by key and values array.
 */
int
DataStore::del_data_tag_values(const tag_key_t tkey, const unsigned int values_count, const tag_value_t *tvalues)
{
	const unsigned tb_index = TAG_BUCKET_INDEX(tkey);
	struct dh_tags_bucket * const tb = &tags[tb_index];
	struct tag_key *tk = NULL;
	struct tag_value *tv, tv_search;
	unsigned *buckets;
	unsigned n_buckets;

	/*
	 * 1. Acquire items bucket locks for all items matching the tag values
	 * 2. Construct a set of all items to be deleted.
	 * 4. Remove all the items
	 * 5. Release all tags and items bucket locks
	 */

	if (tkey == 0)
		return FALSE;

	tags_locks.rdlock(tb_index);
	for (unsigned i = 0; i < tb->keys_size; i++) {
		if (tb->keys[i].key == tkey) {
			tk = &tb->keys[i];
			break;
		}
	}
	if (tk == NULL) {
		tags_locks.unlock(tb_index);
		return FALSE;
	}

	/* 1a) Find data item buckets containing the given key and values */
	if (!get_items_buckets_by_tag_key_values(tk, tvalues, values_count, &buckets, &n_buckets)) {
		tags_locks.unlock(tb_index);
		return FALSE;
	}

	/* 1b) Lock them */
	if (!items_buckets_wrlock(buckets, n_buckets)) {
		mc_log(LOG_ERR, "Cannot wrlock items buckets!");
		tags_locks.unlock(tb_index);
		free(buckets);
		return FALSE;
	}

	/* 2) Find all items */
	std::set<struct dh_data*> iset;
	for (unsigned i = 0; i < values_count; i++) {
		tv_search.value = tvalues[i];
		tv = RB_FIND(tag_value_tree, &tk->tree_head, &tv_search);
		if (tv == NULL)
			continue;
		for (unsigned j = 0; j < tv->items_size; j++) {
			if (tv->vitems[j].item == NULL)
				continue;
			assert(tv->vitems[j].bucket != NULL);
			assert(tv->vitems[j].item->name != NULL);
			iset.insert(tv->vitems[j].item);
			assert(tv->vitems[j].item != NULL);
		}
	}

	/* 3) Remove items from the DataStore */
	for (std::set<struct dh_data*>::iterator it = iset.begin(); it != iset.end(); it++)
		remove_item_and_free_nolock(*it);

	/* 4) Release tag and items bucket lock  */
	items_buckets_unlock(buckets, n_buckets);
	tags_locks.unlock(tb_index);
	free(buckets);
	return TRUE;
}


/**
 * Returns an array of item bucket pointers with items identified by the given tag
 * key and any of the given tag values. The item bucket array is to be freed by
 * the caller (if non-NULL).
 * The function expects that the tag bucket is locked, so it can freely traverse
 * the bucket's trees.
 */
int
DataStore::get_items_buckets_by_tag_key_values(struct tag_key *tk,
    const tag_value_t *tvalues, const unsigned values_count,
    unsigned **buckets, unsigned *n_buckets)
{
	struct tag_value *tv, tv_search;
	ISet<N_ITEM_BUCKETS> bset;

	for (unsigned i = 0; i < values_count; i++) {
		tv_search.value = tvalues[i];
		tv = RB_FIND(tag_value_tree, &tk->tree_head, &tv_search);
		if (tv == NULL)
			continue;
		for (unsigned j = 0; j < tv->items_size; j++)
			if (tv->vitems[j].bucket != NULL) {
				assert(tv->vitems[j].item);
				bset.add(tv->vitems[j].bucket->index);
			}
	}

	*n_buckets = bset.count();
	*buckets = (unsigned*) malloc(*n_buckets * sizeof(unsigned));
	if (*buckets == NULL) {
		mc_log(LOG_ERR, "malloc failed for %u bytes", *n_buckets * sizeof(*buckets));
		*n_buckets = 0;
		return FALSE;
	}
	bset.get_elements(*buckets);
	return TRUE;
}


struct mc_resp_atomic *
DataStore::atomic_op(const int op, const char *name, const unsigned int name_size, const int64_t arg1, const int64_t arg2)
{
	const unsigned ib_index = ITEM_BUCKET_INDEX(name, name_size);
	struct dh_items_bucket * const ib = &items[ib_index];
	struct dh_data search;
	struct dh_data *d;
	struct mc_resp_atomic *mr;
	int64_t *dval;

	search.name = (char*)name;
	search.name_size = name_size;

	items_locks.wrlock(ib_index);
	d = RB_FIND(dh_item_tree, &ib->tree_head, &search);
	if (unlikely(d == NULL || d->inpkt == NULL)) {
		if (op == MCMD_ATOMIC_ADD) {
			/* ATOMIC_ADD is allowed to create a key */
			int64_t zero = 0;
			d = internal_add_record_nolock(name, name_size, (const char*)&zero, sizeof(zero), NULL, 0);
		} else if (op == MCMD_ATOMIC_CMPSET) {
			/* ATOMIC_CMPSET is allowed to create a key */
			d = internal_add_record_nolock(name, name_size, (const char*)&arg1, sizeof(arg1), NULL, 0);
		} else {
			items_locks.unlock(ib_index);
			return NULL;
		}
	}
	if (d->data_size != sizeof(int64_t)) {
		mc_log(LOG_ERR, "Atomic operation on a record of unsupported size (%u should be %u)",
			(unsigned)d->data_size, (unsigned)(sizeof(int64_t)));
		items_locks.unlock(ib_index);
		return NULL;
	}
	mr = (struct mc_resp_atomic*)malloc(sizeof(*mr));
	if (mr == NULL) {
		items_locks.unlock(ib_index);
		return NULL;
	}
	mr->header.total_size = sizeof(*mr);
	dval = (int64_t*)(d->data);
	switch (op) {
	case MCMD_ATOMIC_CMPSET:
		if (*dval == arg1) {
			*dval = arg2;
			mr->result = 1;
		} else
			mr->result = 0;
		break;
	case MCMD_ATOMIC_ADD:
		*dval += arg1;
		mr->result = *dval;
		break;
	case MCMD_ATOMIC_FETCHADD:
		mr->result = *dval;
		*dval += arg1;
		break;
	case MCMD_ATOMIC_READANDCLEAR:
		mr->result = *dval;
		*dval = 0;
		break;
	default:
		mc_log(LOG_ERR, "Invalid atomic op: %d", op);
		free(mr);
		return NULL;
	}
	items_locks.unlock(ib_index);
	return mr;
}


/** Multiple-put: puts multiple records into the cache */
int
DataStore::mput(struct dh_data **entries, unsigned n_records, unsigned flags)
{
	if (likely((flags & MCMD_FLAG_ATOMICALL) == 0)) {
		for (unsigned i = 0; i < n_records; i++)
			put(entries[i]);
	} else {
		ISet<N_ITEM_BUCKETS> bset;
		for (unsigned i = 0; i < n_records; i++) {
			unsigned ib_index = ITEM_BUCKET_INDEX(entries[i]->name, entries[i]->name_size);
			entries[i]->bucket = &items[ib_index];
			bset.add(ib_index);
		}
		unsigned bcount = bset.count();
		unsigned buckets[bcount];
		bset.get_elements(buckets);
		items_buckets_wrlock(buckets, bcount);

		/* XXX: tag locks??? */
		for (unsigned i = 0; i < n_records; i++)
			put_nolock(entries[i]->bucket, &entries[i]);

		items_buckets_unlock(buckets, bcount);
	}

	return TRUE;
}


/** Multiple-get: gets multiple records from the cache by record keys */
struct mc_resp_multidata*
DataStore::mget(struct mc_key_name **keys, unsigned n_names, unsigned flags)
{
	struct mc_resp_multidata *result;

	result = (struct mc_resp_multidata*) calloc(1, sizeof(*result));
	if (result == NULL) {
		mc_log(LOG_ERR, "mget: calloc() failed");
		return NULL;
	}
	result->header.total_size = sizeof(*result);

	if ((flags & MCMD_FLAG_ATOMICALL) == 0) {
		unsigned lock_cookie;
		struct mc_data_entry *mce;

		for (unsigned i = 0; i < n_names; i++) {
			mce = get_data_and_lock(keys[i]->name, keys[i]->name_size, &lock_cookie);
			if (mce == NULL) /* Tests aliveness */
				continue;
			multidata_append(&result, mce);
			got_data_unlock(lock_cookie);
		}
	} else {
		struct mc_data_entry *mce;
		ISet<N_ITEM_BUCKETS> bset;

		for (unsigned i = 0; i < n_names; i++) {
			unsigned ib_index = ITEM_BUCKET_INDEX(keys[i]->name, keys[i]->name_size);
			bset.add(ib_index);
		}

		unsigned bcount = bset.count();
		unsigned buckets[bcount];
		bset.get_elements(buckets);
		items_buckets_rdlock(buckets, bcount);

		for (unsigned i = 0; i < n_names; i++) {
			mce = get_data_nolock(keys[i]->name, keys[i]->name_size);
			if (mce == NULL) /* Tests liveness */
				continue;
			multidata_append(&result, mce);
		}

		items_buckets_unlock(buckets, bcount);
	}
	return result;
}


/** Multiple-del: deletes multiple records from the cache by record keys */
int
DataStore::mdel(struct mc_key_name **keys, unsigned n_names, unsigned flags)
{
	unsigned n_deleted = 0;

	if ((flags & MCMD_FLAG_ATOMICALL) == 0) {
		for (unsigned i = 0; i < n_names; i++)
			if (del_data_name(keys[i]->name, keys[i]->name_size))
				n_deleted++;
	} else {
		ISet<N_ITEM_BUCKETS> bset;

		for (unsigned i = 0; i < n_names; i++) {
			unsigned ib_index = ITEM_BUCKET_INDEX(keys[i]->name, keys[i]->name_size);
			bset.add(ib_index);
		}

		unsigned bcount = bset.count();
		unsigned buckets[bcount];
		bset.get_elements(buckets);
		items_buckets_wrlock(buckets, bcount);

		for (unsigned i = 0; i < n_names; i++)
			if (del_data_name_nolock(keys[i]->name, keys[i]->name_size))
				n_deleted++;

		items_buckets_unlock(buckets, bcount);
	}
	return n_deleted;
}


/** Tstack pop: returnes & removes a record from a virtual stack specified by the tag key & value */
struct mc_data_entry*
DataStore::tstack_pop(tag_key_t tkey, tag_value_t tval)
{
	unsigned tb_index = TAG_BUCKET_INDEX(tkey);
	struct dh_tags_bucket * const tb = &tags[tb_index];
	struct mc_data_entry *mce = NULL;
	struct tag_key *tk;
	struct tag_value *tv, tv_find;
	struct dh_item_value_entry *fitem;
	struct dh_data *found_item;
	struct dh_items_bucket *found_ib;
	int fk, fv, ii;

	tv_find.value = tval;
	
	/* This requires a careful locking dance. Lots of backtracking. */
	while (true) {
		fk = fv = ii = -1;
		tags_buckets_wrlock(&tb_index, 1);
		for (unsigned i = 0; i < tb->keys_size; i++)
			if (tb->keys[i].key == tkey) {
				fk = i;
				break;
			}
		if (fk == -1) {
			/* Key not found */
			tags_buckets_unlock(&tb_index, 1);
			return NULL;
		}
		tk = &tb->keys[fk];
		tv = RB_FIND(tag_value_tree, &tk->tree_head, &tv_find);
		if (tv == NULL || tv->items_size == 0) {
			/* Tag value not found */
			tags_buckets_unlock(&tb_index, 1);
			return NULL;
		}
		for (int i = tv->items_size - 1; i >= 0; i--)
			if (tv->vitems[i].item != NULL) {
				ii = i;
				break;
			}
		if (ii == -1) {
			tags_buckets_unlock(&tb_index, 1);
			return NULL;
		}
		/* Now, we found a record in tv->vitems[ii], but we need to lock it
		 * before we can do something with it, and items locking is parallel
		 * to tags locking. */
		fitem = &tv->vitems[ii];
		if (items_locks.try_wrlock(fitem->bucket->index) != LR_OK) {
			/* Dang, someone is working with the bucket. Go back to square 1. */
			tags_buckets_unlock(&tb_index, 1);
			continue;
		}
		/* Finally, we hold both the items bucket and the tag bucket! But alas,
		 * it's much easier to give up on the tag bucket and keep the items one. */
		found_item = fitem->item;
		found_ib = fitem->bucket;
		tags_buckets_unlock(&tb_index, 1);
		/* Test if the item is alive */
		if (!item_is_alive(found_item)) {
			items_locks.unlock(found_ib->index);
			return NULL;
		}
		/* Unfortunately, we need to make a copy of the record for layaring
		 * reasons. */
		mce = (struct mc_data_entry*)malloc(found_item->inpkt->header.total_size);
		if (mce == NULL) {
			mc_log(LOG_ERR, "malloc(%d) failed", found_item->inpkt->header.total_size);
			items_locks.unlock(found_ib->index);
			return NULL;
		}
		memcpy(mce, found_item->inpkt, found_item->inpkt->header.total_size);
		del_nolock(found_ib, found_item); /* It also frees it. */
		/* And finally, we're out. */
		items_locks.unlock(found_ib->index);
		break;
	}
	
	return mce;
}

/**
 * (Re)set the tags of a record 
 */
int
DataStore::set_tags(char *name, unsigned name_size, struct mc_tag *tags, unsigned n_tags)
{
	const unsigned ib_index = ITEM_BUCKET_INDEX(name, name_size);
	struct dh_items_bucket * const ib = &items[ib_index];
	unsigned tbuckets[N_TAG_BUCKETS];
	struct dh_data search, *d;
	struct mc_data_entry *newmde;
	int i;
	
	for (i = 0; i < N_TAG_BUCKETS; i++)
		tbuckets[i] = i;
	search.name = (char*)name;
	search.name_size = name_size;

	/* This could lead to horribly difficult locking, so for now we'll just
	 * lock all of the tag locks in addition to the item bucket lock. */
locking:
	while (true) {
		if (tags_locks.try_wrlock(tbuckets, N_TAG_BUCKETS) == LR_OK)
			break;
		/* Sleep? */
		pthread_yield();
	}
	if (items_locks.try_wrlock(ib_index) == LR_TRYAGAIN) {
		tags_locks.unlock(tbuckets, N_TAG_BUCKETS);
		pthread_yield();
		goto locking;
	}
	d = RB_FIND(dh_item_tree, &ib->tree_head, &search);
	if (d == NULL) {
		mc_log(LOG_ERR, "Not found: %s", name);
		items_locks.unlock(ib_index);
		tags_locks.unlock(tbuckets, N_TAG_BUCKETS);
		return FALSE;
	}
	
	for (i = 0; i < d->n_tags; i++)
		tag_unlink(d, &(d->tags[i]));
	
	if (mc_build_data_entry(&newmde, name, name_size, d->data, d->data_size, tags, n_tags, d->exptime, NULL) != MCREASON_OK) {
		items_locks.unlock(ib_index);
		tags_locks.unlock(tbuckets, N_TAG_BUCKETS);
		return FALSE;
	}
	free(d->inpkt);
	fill_dh_data(d, newmde);
	d->bucket = ib;
	
	for (i = 0; i < d->n_tags; i++)
		tag_link(d, &tags[i]);
	
	items_locks.unlock(ib_index);
	tags_locks.unlock(tbuckets, N_TAG_BUCKETS);
	return TRUE;
}


/**
 * Acquires read locks on all buckets in the given list, with the rollback-and-retry
 * protocol if all locks cannot be acquired in one step.
 */
bool
DataStore::items_buckets_rdlock(unsigned *buckets, unsigned n_buckets)
{
#ifdef LOCK_TRACE_ENABLE
	char buf[n_buckets * 8];
	char tmp[10];
	buf[0] = '\0';
	for (unsigned i = 0; i < n_buckets; i++) {
		sprintf(tmp, "%d ", buckets[i]);
		strlcat(buf, tmp, sizeof(buf));
	}
	mc_log(LOG_DEBUG, "%%items_buckets_rdlock of %d buckets: %s", n_buckets, buf);
#endif
	while (true) {
		if (items_locks.try_rdlock(buckets, n_buckets) == LR_OK)
			break;
		/* Sleep? */
		pthread_yield();
	}
	return true;
}


/**
 * Acquires read locks on all buckets in the given list, with the rollback-and-retry
 * protocol if all locks cannot be acquired in one step.
 */
bool
DataStore::items_buckets_wrlock(unsigned *buckets, unsigned n_buckets)
{
#ifdef LOCK_TRACE_ENABLE
	char buf[n_buckets * 8];
	char tmp[10];
	buf[0] = '\0';
	for (unsigned i = 0; i < n_buckets; i++) {
		sprintf(tmp, "%d ", buckets[i]);
		strlcat(buf, tmp, sizeof(buf));
	}
	mc_log(LOG_DEBUG, "%%items_buckets_wrlock of %d buckets: %s", n_buckets, buf);
#endif
	while (true) {
		if (items_locks.try_wrlock(buckets, n_buckets) == LR_OK)
			break;
		/* Sleep? */
		pthread_yield();
	}
	return true;
}


/**
 * Unlock the buckets in the array.
 */
void
DataStore::items_buckets_unlock(unsigned *buckets, unsigned n_buckets)
{
#ifdef LOCK_TRACE_ENABLE
	char buf[n_buckets * 8];
	char tmp[10];
	buf[0] = '\0';
	for (unsigned i = 0; i < n_buckets; i++) {
		sprintf(tmp, "%d ", buckets[i]);
		strncat(buf, tmp, sizeof(buf));
	}
	mc_log(LOG_DEBUG, "%%items_buckets_unlock of %d buckets: %s", n_buckets, buf);
#endif
	items_locks.unlock(buckets, n_buckets);
}


/**
 * Write-lock the tag buckets in the array.
 */
bool
DataStore::tags_buckets_wrlock(unsigned *buckets, unsigned n_buckets)
{
	while (true) {
		if (tags_locks.try_wrlock(buckets, n_buckets) == LR_OK)
			break;
		/* Sleep? */
		pthread_yield();
	}
	return true;
}


/**
 * Unlock the tag buckets in the array.
 */
void
DataStore::tags_buckets_unlock(unsigned *buckets, unsigned n_buckets)
{
	tags_locks.unlock(buckets, n_buckets);
}


/**
 * Remove the given item from all DataStore structures.
 * Expects that the item bucket lock for the item is locked, but it will
 * acquire and release tag bucket locks as needed.
 */
void
DataStore::remove_item_and_free(struct dh_data *d)
{
	assert(d != NULL);
	assert(d->bucket != NULL);
	assert(d->name != NULL);

	RB_REMOVE(dh_item_tree, &d->bucket->tree_head, d);
	data_item_safe_lock_op(d, LOP_WRLOCK, false, true);
	tags_cleanup(d);
	data_item_safe_lock_op(d, LOP_UNLOCK, false, true);
	data_cleanup(d);
	free(d);
}


/**
 * Remove the given item from all DataStore structures.
 * Expects that the tags and item is locked.
 */
void
DataStore::remove_item_and_free_nolock(struct dh_data *d)
{
	assert(d != NULL);
	assert(d->bucket != NULL);
	assert(d->name != NULL);

	RB_REMOVE(dh_item_tree, &d->bucket->tree_head, d);
	tags_cleanup(d);
	data_cleanup(d);
	free(d);
}

/**
 * Free the node's data, but not the node itself. Requires the item bucket be rwlocked.
 * Doesn't do anything with tags.
 */
void
DataStore::data_cleanup(struct dh_data *d)
{
	assert(d->inpkt != NULL);
	d->bucket->memory_used -= d->inpkt->header.total_size; 
	d->data = d->name = NULL;
	free(d->inpkt);
	d->inpkt = NULL;
	d->n_tags = 0;
}


/**
 * Unlinks and clears the node's tags. Requires that the tag buckets be rwlocked.
 */
void
DataStore::tags_cleanup(struct dh_data *d)
{
	for (unsigned i = 0; i < d->n_tags; i++)
		tag_unlink(d, &d->tags[i]);
}


/**
 * Like strdup() but binary safe.
 */
static char*
memdup(void *p, unsigned size)
{
	char *r = (char*)malloc(size);
	memcpy(r, p, size);
	return r;
}


/**
 * Cleans up dead items in two passes for each bucket:
 * 1. Iterates through the items while holding a read lock, it decides which
 *    items to reap, recording their names and name sizes in a list.
 * 2. Acquires the write lock, iterates through the list and deletes the
 *    found records.
 */
unsigned
DataStore::reap_dead_items()
{
	unsigned reap_count = 0;
	
	for (unsigned i = 0; i < n_item_buckets; i++) {
		struct dh_data *d;
		std::vector<std::pair<char*,unsigned>*> dead_list;
		unsigned dead_list_count = 0;
		
		/* First pass: avoid wrlocking and just look for dead items */
		items_locks.rdlock(i);
		RB_FOREACH(d, dh_item_tree, &items[i].tree_head) {
			if (!item_is_alive(d)) {
				dead_list.push_back(new std::pair<char*,unsigned>(memdup(d->name, d->name_size), d->name_size));
				dead_list_count++;
			}
		}
		items_locks.unlock(i);
		/* Second pass, remove found dead items (if any) */
		if (dead_list_count == 0)
			continue;
		while (!dead_list.empty()) {
			std::pair<char*,unsigned> *dead = dead_list.back();
			dead_list.pop_back();
			del_data_name(dead->first, dead->second);
			free(dead->first);
			delete dead;
		}
		reap_count += dead_list_count;
	}
	stat_reap_dead_count += reap_count;
	return reap_count;
}

/**
 * Tries to keep cache memory usage in check by removing items. It does so
 * per bucket:
 * 1. Iterates through the items while holding a read lock, it decides which
 *    items to reap, recording their names in a list.
 * 2. Acquires the write lock, iterates through the list and deletes the
 *    found items.
 * The decision on what items to remove is done in one of two ways: if DEEP_STATS
 * is set, the least used records are deleted, else, the oldest records are
 * deleted.
 */
static bool
reap_free_memory_sort(std::pair<struct dh_data *,unsigned> *d1, std::pair<struct dh_data *,unsigned> *d2) {
	return d1->second < d2->second;
}


unsigned
DataStore::reap_free_memory(uint64_t memory_limit)
{
	uint32_t bucket_memory_limit = memory_limit / N_ITEM_BUCKETS;
	if (bucket_memory_limit == 0)
		return 0;
	
	unsigned reap_count = 0;
	
	for (unsigned i = 0; i < n_item_buckets; i++) {
		if (items[i].memory_used < bucket_memory_limit)
			continue;
		
		uint32_t bucket_memory_tofree = items[i].memory_used - bucket_memory_limit;
		struct dh_data *d;
		std::vector<std::pair<struct dh_data *,unsigned>*> stat_list;
		std::vector<std::pair<char*,unsigned>*> dead_list;
		unsigned dead_list_count = 0;
		
		/* First pass: create a list of items to delete  */
		items_locks.rdlock(i);
		RB_FOREACH(d, dh_item_tree, &items[i].tree_head) {
#ifdef DEEP_STATS
			/* We have item read access statistics, so reap by usage (delete least used) */
			stat_list.push_back(new std::pair<struct dh_data *,unsigned>(d, d->stat_n_read_ops));
#else
			/* We can only reap by last modified time (delete oldest) */
			stat_list.push_back(new std::pair<struct dh_data *,unsigned>(d, d->mtime));
#endif
		}
		sort(stat_list.begin(), stat_list.end(), reap_free_memory_sort);
		uint32_t memory_freed = 0;
		std::vector<std::pair<struct dh_data *,unsigned>*>::iterator stat_it;
		for (stat_it = stat_list.begin(); stat_it != stat_list.end(); ++stat_it) {
			d = (*stat_it)->first;
			memory_freed += d->inpkt->header.total_size;
			dead_list.push_back(new std::pair<char*,unsigned>(memdup(d->name, d->name_size), d->name_size));
			dead_list_count++;
			if (memory_freed >= bucket_memory_tofree)
				break;
		}
		items_locks.unlock(i);
		/* Get rid of stat_list */
		while (!stat_list.empty()) {
			std::pair<struct dh_data *,unsigned>* p = stat_list.back();
			stat_list.pop_back();
			delete p;
		}
		
		/* Second pass, remove found dead items (if any) */
		if (dead_list_count == 0)
			continue;
		while (!dead_list.empty()) {
			std::pair<char*,unsigned> *dead = dead_list.back();
			dead_list.pop_back();
			del_data_name(dead->first, dead->second);
			free(dead->first);
			delete dead;
		}
		reap_count += dead_list_count;
	}
	stat_reap_memory_count += reap_count;
	return reap_count;
}



