/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2011. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL.
 * $Id: ISet.h 480 2012-04-10 22:37:02Z ivoras $
 */

#ifndef _ISET_H_
#define _ISET_H_

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <assert.h>

#include <err.h>

#define _ISET_TYPE uint64_t
#define _ISET_ONE ((_ISET_TYPE)1)
#define _ISET_BLEN (8 * sizeof(_ISET_TYPE))
#define _ISET_ALEN (N % _ISET_BLEN == 0 ? N / _ISET_BLEN : N / _ISET_BLEN + 1)


/**
 * The ISet class implements a compact set of unsigned integers. It can hold
 * integers from 0 to the given (N-1), implemented as a bitmap.
 */
template <unsigned N>
class ISet {
private:
	_ISET_TYPE bitmap[_ISET_ALEN];
public:
	/** Constructs an empty integer set */
	ISet();

	/** Adds an element to the set */
	void add(unsigned i);

	/** Tests if the element is in the set */
	bool test(unsigned i);

	/** Removes an element from the set */
	void remove(unsigned i);

	/** Clears the set */
	void clear();

	/** Returns the number of elements in the set */
	unsigned count();

	/**
	 * Returns the (integer) elements of the set into the given array, which
	 * must be allocated to contain at least count() elements.
	 */
	void get_elements(unsigned *arr);
	void get_elements_safe(unsigned *arr, unsigned arr_size);

	/* Debugging functions */
	_ISET_TYPE* _get_bitmap();
	unsigned _get_bitmap_len();
	unsigned _get_bitmap_element_bytes();
};


template <unsigned N>
ISet<N>::ISet()
{
	clear();
}


template <unsigned N> void
ISet<N>::add(unsigned i)
{
	assert(i < N);
	unsigned d = i / _ISET_BLEN;
	unsigned m = i % _ISET_BLEN;
	bitmap[d] |= (_ISET_ONE << m);
}


template <unsigned N> bool
ISet<N>::test(unsigned i)
{
	assert(i < N);
	unsigned d = i / _ISET_BLEN;
	unsigned m = i % _ISET_BLEN;
	return (bitmap[d] & (_ISET_ONE << m)) != 0;
}


template <unsigned N> void
ISet<N>::remove(unsigned i)
{
	assert(i < N);
	unsigned d = i / _ISET_BLEN;
	unsigned m = i % _ISET_BLEN;
	bitmap[d] &= ~(_ISET_ONE << m);
}


template <unsigned N> void
ISet<N>::clear()
{
	memset(bitmap, 0, sizeof(bitmap));
}


template <unsigned N> unsigned
ISet<N>::count()
{
	unsigned c = 0, i;
	_ISET_TYPE v;

	for (i = 0; i < _ISET_ALEN; i++) {
		v = bitmap[i];
		for (; v; c++)
			v &= v - 1;
	}
	return c;
}


template <unsigned N> void
ISet<N>::get_elements(unsigned *arr)
{
	unsigned c = 0, i;

	for (i = 0; i < N; i++)
		if (test(i))
			arr[c++] = i;
}


template <unsigned N> void
ISet<N>::get_elements_safe(unsigned *arr, unsigned arr_size)
{
	unsigned c = 0, i;

	for (i = 0; i < N; i++)
		if (test(i)) {
			if (c >= arr_size) {
				warnx("c=%u, arr_size=%u, count()=%u", c, arr_size, count());
				for (unsigned j = 0; j < _ISET_ALEN; j++)
					warnx("bitmap[%u] = %lx", j, bitmap[j]);
			}
			assert(c < arr_size);
			arr[c++] = i;
		}
}


template <unsigned N> _ISET_TYPE*
ISet<N>::_get_bitmap()
{
	return bitmap;
}


template <unsigned N> unsigned
ISet<N>::_get_bitmap_len()
{
	return _ISET_ALEN;
}


template <unsigned N> unsigned
ISet<N>::_get_bitmap_element_bytes()
{
	return sizeof(_ISET_TYPE);
}

#undef _ISET_TYPE
#undef _ISET_BLEN
#undef _ISET_ALEN
#undef _ISET_ONE

#endif
