/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2012. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: LockPool.h 475 2012-04-08 10:27:00Z ivoras $
 */

#ifndef _LOCKPOOL_H_
#define _LOCKPOOL_H_

#include <exception>
#include <pthread.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "misc.h"
#include "logger.h"

enum LOCK_RESULT { LR_OK, LR_TRYAGAIN, LR_DISASTER };

enum LOCK_OP { LOP_RDLOCK, LOP_WRLOCK, LOP_UNLOCK, LOP_TRY_RDLOCK, LOP_TRY_WRLOCK };

//#define LOCK_TRACE_ENABLE

#ifdef LOCK_TRACE_ENABLE
#define LOCK_TRACE(op,n) mc_log(LOG_DEBUG, "%%%slock on %d of %p[%d]", op, n, this, N);
#else
#define LOCK_TRACE(op,n)
#endif

class LockException : public std::exception {
private:
	char *msg;
public:
	LockException(char *msg) : exception() {
		this->msg = msg;
	}
};


class LockNotifier {
public:
	virtual enum LOCK_RESULT rdlock(unsigned *nlist, unsigned list_size) = 0;
	virtual enum LOCK_RESULT wrlock(unsigned *nlist, unsigned list_size) = 0;
	virtual enum LOCK_RESULT unlock(unsigned *nlist, unsigned list_size) = 0;
	virtual enum LOCK_RESULT try_rdlock(unsigned *nlist, unsigned list_size) = 0;
	virtual enum LOCK_RESULT try_wrlock(unsigned *nlist, unsigned list_size) = 0;
	virtual ~LockNotifier() {}
};


/**
 * LockPool is a giant pool of locks which are used practically in all locking
 * operations, with the specific purpose of making them available over the
 * network. It basically implements an array of rwlocks.
 */
template <unsigned N>
class LockPool {
private:
	pthread_rwlock_t 	locks[N];
	/** The generation is incremented every time a lock is wrlocked */
	unsigned 		generations[N];
	LockNotifier 		*lnf;
public:
	LockPool() throw(LockException);
	~LockPool();
	unsigned getSize() { return N; }
	unsigned getGeneration(unsigned n) { return generations[n]; }

	void set_lock_notifier(LockNotifier *lnf) { this->lnf = lnf; }
	void unset_lock_notifier() { this->lnf = NULL; }

	/* Success-or-death, atomic */
	enum LOCK_RESULT rdlock(unsigned n);
	enum LOCK_RESULT rdlock(unsigned *nlist, unsigned list_size);

	enum LOCK_RESULT wrlock(unsigned n);
	enum LOCK_RESULT wrlock(unsigned *nlist, unsigned list_size);

	enum LOCK_RESULT unlock(unsigned n);
	enum LOCK_RESULT unlock(unsigned *nlist, unsigned list_size);

	/* EWOULDBLOCK style; will return true only if both local and notifier
	 * return true. */
	enum LOCK_RESULT try_rdlock(unsigned n);
	enum LOCK_RESULT try_rdlock(unsigned *nlist, unsigned list_size);

	enum LOCK_RESULT try_wrlock(unsigned n);
	enum LOCK_RESULT try_wrlock(unsigned *nlist, unsigned list_size);

	/* Backdoor, not going through the notifier */
	enum LOCK_RESULT _internalLockOp(enum LOCK_OP lop, unsigned *nlist, unsigned list_size);

	/* Check if all the locks are free to be locked. Expensive and for debugging only */
	bool allClear();
};


template <unsigned N>
LockPool<N>::LockPool() throw (LockException) {
	for (unsigned i = 0; i < N; i++) {
		if (pthread_rwlock_init(&locks[i], NULL) != 0)
			throw new LockException((char*)"pthread_rwlock_init failed at " __FILE__);
	}
	lnf = NULL;
	memset(&generations, 0, sizeof(generations));
}


template <unsigned N>
LockPool<N>::~LockPool() {
	for (unsigned i = 0; i < N; i++)
		pthread_rwlock_destroy(&locks[i]);
}

static inline int
pthread_rwlock_rdlock_l(pthread_rwlock_t *lock) {
	int r;

	for (;;) {
		r = pthread_rwlock_rdlock(lock);
		if (r == 0 || r != EAGAIN)
			break;
	}
	return r;
}

template <unsigned N>
enum LOCK_RESULT LockPool<N>::rdlock(unsigned n) {
	assert(n < N);
	LOCK_TRACE("rd", n);
	CRIT_ASSERT(pthread_rwlock_rdlock_l(&locks[n]) == 0);
	if (unlikely(lnf != NULL)) {
		enum LOCK_RESULT lres = lnf->rdlock(&n, 1);

		if (unlikely(lres != LR_OK))
			CRIT_ASSERT(pthread_rwlock_unlock(&locks[n])); // Rollback
		return lres;
	}
	return LR_OK;
}


template <unsigned N>
enum LOCK_RESULT LockPool<N>::rdlock(unsigned *nlist, unsigned list_size) {
	for (unsigned i = 0; i < list_size; i++) {
		assert(nlist[i] < N);
		LOCK_TRACE("*rd", nlist[i]);
		CRIT_ASSERT(pthread_rwlock_rdlock_l(&locks[nlist[i]]) == 0);
	}
	if (unlikely(lnf != NULL)) {
		enum LOCK_RESULT lres = lnf->rdlock(nlist, list_size);

		if (unlikely(lres != LR_OK))
			for (unsigned i = 0; i < list_size; i++) // Rollback
				CRIT_ASSERT(pthread_rwlock_unlock(&locks[nlist[i]]) == 0);
		return lres;
	}
	return LR_OK;
}


template <unsigned N>
enum LOCK_RESULT LockPool<N>::wrlock(unsigned n) {
	assert(n < N);
	LOCK_TRACE("wr", n);
	CRIT_ASSERT(pthread_rwlock_wrlock(&locks[n]) == 0);
	if (unlikely(lnf != NULL)) {
		enum LOCK_RESULT lres = lnf->wrlock(&n, 1);

		if (unlikely(lres != LR_OK))
			CRIT_ASSERT(pthread_rwlock_unlock(&locks[n]) == 0); // Rollback
		return lres;
	}
	generations[n]++;
	return LR_OK;
}


template <unsigned N>
enum LOCK_RESULT LockPool<N>::wrlock(unsigned *nlist, unsigned list_size) {
	for (unsigned i = 0; i < list_size; i++) {
		assert(nlist[i] < N);
		LOCK_TRACE("*wr", nlist[i]);
		CRIT_ASSERT(pthread_rwlock_wrlock(&locks[nlist[i]]) == 0);
	}
	if (unlikely(lnf != NULL)) {
		enum LOCK_RESULT lres = lnf->wrlock(nlist, list_size);

		if (unlikely(lres != LR_OK))
			for (unsigned i = 0; i < list_size; i++)  // Rollback
				CRIT_ASSERT(pthread_rwlock_unlock(&locks[nlist[i]]) == 0);
		return lres;
	}
	for (unsigned i = 0; i < list_size; i++)
		generations[nlist[i]]++;
	return LR_OK;
}


/* TODO: What does a remote failure at UNLOCK mean? Will it leave inconsistent
 * lock states? Will it matter?*/
template <unsigned N>
enum LOCK_RESULT LockPool<N>::unlock(unsigned n) {
	LOCK_TRACE("un", n);
	int result;
	if ((result = pthread_rwlock_unlock(&locks[n])) != 0)
		mc_log(LOG_CRIT, "pthread_rwlock_unlock() failed on lock %u: errno %d (%s)", n, result, strerror(result));
	if (unlikely(lnf != NULL))
		lnf->unlock(&n, 1);
	return LR_OK;
}


template <unsigned N>
enum LOCK_RESULT LockPool<N>::unlock(unsigned *nlist, unsigned list_size) {
	for (unsigned i = 0; i < list_size; i++) {
		LOCK_TRACE("*un", nlist[i]);
		CRIT_ASSERT(pthread_rwlock_unlock(&locks[nlist[i]]) == 0);
	}
	if (unlikely(lnf != NULL))
		lnf->unlock(nlist, list_size);
	return LR_OK;
}


template <unsigned N>
enum LOCK_RESULT LockPool<N>::try_rdlock(unsigned n) {
	assert(n < N);
	LOCK_TRACE("tryrd", n);
	int result = pthread_rwlock_tryrdlock(&locks[n]);
	if (likely(result == 0)) {
		if (unlikely(lnf != NULL)) {
			if (lnf->try_rdlock(&n, 1) == LR_OK)
				return LR_OK;
			else {
				/* Does it matter if the remote locker failed with
				 * LR_TRYAGAIN or LR_DISASTER? */
				pthread_rwlock_unlock(&locks[n]); // Rollback
				return LR_TRYAGAIN;
			}
		} else
			return LR_OK;
	} else {
		if (result == EBUSY)
			return LR_TRYAGAIN;
		mc_log(LOG_CRIT, "Error in pthread_rwlock_tryrdlock: errno %d", result);
		return LR_DISASTER;
	}
}


template <unsigned N>
enum LOCK_RESULT LockPool<N>::try_rdlock(unsigned *nlist, unsigned list_size) {
	for (unsigned i = 0; i < list_size; i++) {
		LOCK_TRACE("*tryrd", nlist[i]);
		int result = pthread_rwlock_tryrdlock(&locks[nlist[i]]);
		if (unlikely(result != 0)) {
			for (unsigned j = 0; j < i; j++)
				pthread_rwlock_unlock(&locks[nlist[j]]);
			if (result != EBUSY)
				mc_log(LOG_CRIT, "Error in pthread_rwlock_tryrdlock: errno %d", result);
			return LR_TRYAGAIN;
		}
	}
	if (unlikely(lnf != NULL)) {
		if (lnf->try_rdlock(nlist, list_size) == LR_OK)
			return LR_OK;
		else {
			for (unsigned i = 0; i < list_size; i++) // Rollback
				pthread_rwlock_unlock(&locks[nlist[i]]);
			return LR_TRYAGAIN;
		}
	} else
		return LR_OK;
}


template <unsigned N>
enum LOCK_RESULT LockPool<N>::try_wrlock(unsigned n) {
	assert(n < N);
	LOCK_TRACE("trywr", n);
	int result = pthread_rwlock_trywrlock(&locks[n]);
	if (likely(result == 0)) {
		if (unlikely(lnf != NULL)) {
			if (lnf->try_wrlock(&n, 1) == LR_OK)
				return LR_OK;
			else {
				pthread_rwlock_unlock(&locks[n]); // Rollback
				return LR_TRYAGAIN;
			}
		}
		return LR_OK;
	} else {
		if (result == EBUSY)
			return LR_TRYAGAIN;
		mc_log(LOG_CRIT, "Error in pthread_rwlock_trywrlock: errno %d", result);
		return LR_DISASTER;
	}
}


template <unsigned N>
enum LOCK_RESULT LockPool<N>::try_wrlock(unsigned *nlist, unsigned list_size) {
	for (unsigned i = 0; i < list_size; i++) {
		LOCK_TRACE("*trywr", nlist[i]);
		int result = pthread_rwlock_trywrlock(&locks[nlist[i]]);
		if (unlikely(result != 0)) {
			for (unsigned j = 0; j < i; j++)
				pthread_rwlock_unlock(&locks[nlist[j]]);
			if (result != EBUSY)
				mc_log(LOG_CRIT, "Error in pthread_rwlock_trywrlock: errno %d", result);
			return LR_TRYAGAIN;
		}
	}
	if (unlikely(lnf != NULL)) {
		if (lnf->try_wrlock(nlist, list_size) == LR_OK)
			return LR_OK;
		else {
			for (unsigned i = 0; i < list_size; i++) // Rollback
				pthread_rwlock_unlock(&locks[nlist[i]]);
			return LR_TRYAGAIN;
		}
	} else
		return LR_OK;
}


template <unsigned N>
enum LOCK_RESULT LockPool<N>::_internalLockOp(enum LOCK_OP lop, unsigned *nlist, unsigned list_size) {
	unsigned i;
	int result;

	switch (lop) {
	case LOP_RDLOCK:
		for (i = 0; i < list_size; i++)
			CRIT_ASSERT(pthread_rwlock_rdlock_l(&locks[nlist[i]]) == 0);
		return LR_OK;
	case LOP_WRLOCK:
		for (i = 0; i < list_size; i++)
			CRIT_ASSERT(pthread_rwlock_wrlock(&locks[nlist[i]]) == 0);
		return LR_OK;
	case LOP_UNLOCK:
		for (i = 0; i < list_size; i++)
			CRIT_ASSERT(pthread_rwlock_unlock(&locks[nlist[i]]) == 0);
		return LR_OK;
	case LOP_TRY_RDLOCK:
		for (i = 0; i < list_size; i++)
			if ((result = pthread_rwlock_tryrdlock(&locks[nlist[i]])) != 0) {
				for (unsigned j = 0; j < i; j++) // Rollback
					CRIT_ASSERT(pthread_rwlock_unlock(&locks[nlist[j]]) == 0);
				return result == EBUSY ? LR_TRYAGAIN : LR_DISASTER;
			}
		return LR_OK;
	case LOP_TRY_WRLOCK:
		for (i = 0; i < list_size; i++)
			if ((result = pthread_rwlock_trywrlock(&locks[nlist[i]])) != 0) {
				for (unsigned j = 0; j < i; j++) // Rollback
					CRIT_ASSERT(pthread_rwlock_unlock(&locks[nlist[j]]) == 0);
				return result == EBUSY ? LR_TRYAGAIN : LR_DISASTER;
			}
		return LR_OK;
	default:
		mc_log(LOG_ERR, "Unknown lock op: %d", lop);
	}
	return LR_DISASTER;
}


/**
 * Test if all locks in the pool can be acquired. This is pointless to do except
 * for extreme debugging :)
 */
template <unsigned N>
bool LockPool<N>::allClear() {
	int r;
	for (unsigned i = 0; i < N; i++) {
		if ((r = pthread_rwlock_trywrlock(&locks[i])) != 0) {
			for (unsigned j = 0; j < i; j++)
				pthread_rwlock_unlock(&locks[j]);
			printf("allClear<%d>: taken: %d? err %d: \n", N, i, r);
			return false;
		}
	}
	for (unsigned i = 0; i < N; i++)
		pthread_rwlock_unlock(&locks[i]);
	return true;
}

#undef LOCK_TRACE

#endif
