/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2012. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: ServerClientConnection.cpp 480 2012-04-10 22:37:02Z ivoras $
 */

#include <sys/param.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <stdint.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <syslog.h>
#include <assert.h>
#include <err.h>

#ifdef DMALLOC
#include <dmalloc.h>
#endif

#include "queue.h"
#include "tree.h"
#include <list>
#include <iterator>

#include "debug.h"
#include "misc.h"
#include "logger.h"
#include "minilibevent.h"
#include "server.h"
#include "mc_protocol.h"
#include "DataStore.h"
#include "common_utils.h"

#include "Thread.h"
#include "ServerWorkerThread.h"
#include "ServerNetworkThread.h"
#include "ServerModelMetadata.h"
#include "ServerClientConnection.h"


/* A lot of dirty globals from server.cpp since all of this was refactored
 * from it. */
extern DataStore *cache;
extern unsigned int n_worker_threads;
extern unsigned int n_network_threads;
extern ServerWorkerThread *worker_threads;
extern unsigned int inbuf_alloc_size;
extern unsigned int outbuf_alloc_size;

#ifndef NDEBUG
extern bool internal_debug;
#endif

/**
 * Returns a human-friendly string describing the connection type.
 */
const char *
conn_type_name(enum E_CONN_TYPE type)
{
	switch (type) {
	case CONN_TYPE_CLIENT_UNIX:
		return "Unix/Local client";
	case CONN_TYPE_CLIENT_TCP:
		return "TCP client";
	case CONN_TYPE_SERVER_UNIX:
		return "Unix/local server";
	case CONN_TYPE_SERVER_TCP:
		return "TCP server";
	case CONN_TYPE_NO:
		return "*EMPTY SLOT*";
	default:
		return "!!UNKNOWN!!";
	}
}


/**
 * Assures buffers are allocated for the connection.
 * The idea here is to severly reduce memory allocation overhead by
 * allocating memory in fixed-sized chunks, and in advance.
 */
void
ServerClientConnection::assure_bufs(bool init_outbuf)
{
	if (inbuf == NULL)
		new_inbuf();
	inbuf_pos = 0;
	if (init_outbuf && outbuf != NULL) {
		free(outbuf);
		outbuf = NULL;
	}
	outbuf_len = 0;
	outbuf_pos = 0;
}


/**
 * Allocates a new inbuf on the connection.
 */
void
ServerClientConnection::new_inbuf()
{
	inbuf = (struct mc_header*) malloc(inbuf_alloc_size);
	inbuf_len = inbuf ? inbuf_alloc_size : 0;
	inbuf_pos = 0;
}


/**
 * Set up a newly allocated connection
 */
void
ServerClientConnection::setup_blank()
{
	this->type = CONN_TYPE_NO;
	this->inbuf = NULL;
	this->outbuf = NULL;
	assure_bufs(false);
}


/**
 * A client has disconnected
 */
void
ServerClientConnection::disconnect_client()
{
	assert(socket >= 0);
	disable_read_event();
	disable_write_event();
	close(socket);
	mc_log(LOG_DEBUG, "Closed %s (%p, fd %d), %u ops", conn_type_name(type), this, socket, ops_count);
	socket = -1;
	inbuf_pos = outbuf_pos = 0;
	type = CONN_TYPE_NO;
	/* We don't deallocate buffers; reuse them. */
}


/**
 * Outputs the specified buffer on the specified connection.
 * Tries never to touch this->outbuf, send()ing the buffer directly
 * if it's possible.
 *
 * To do buffer reuse, in normal command-response interaction with the
 * client, the inpkt must be considered invalid when output() returns,
 * as it may be overwritten by the *next* incoming packet. Therefore,
 * output() should be the last "inteligent" operation when responding to
 * a command.
 */
int
ServerClientConnection::output(void *obuf, unsigned int buf_size, bool clear_inbuf)
{
	char *cbuf = (char*)obuf;

	if (clear_inbuf) /* need to prepare inbuf before reponse to the current packet is to be received */
		inbuf_pos = 0;

#ifndef NDEBUG
	if (internal_debug)
		_mc_debug_printf("output(%u) [%d], outbuf_pos=%u, outbuf_len=%u\n", buf_size,
			((struct mc_header*)obuf)->total_size, outbuf_pos, outbuf_len);
#endif

	if (outbuf_pos == outbuf_len && outbuf_pos == 0) {
		/* No queued data to output; the buffer will most likely fit in a single send() call */
		size_t sent = send(socket, cbuf, buf_size, 0);
		_mc_debug_printf("<< output on %d: mcmd %d, size %u\n", socket, ((struct mc_header*)obuf)->mcmd, buf_size);
#ifndef NDEBUG
		if (internal_debug)
			_mc_debug_printf("output: send(%u) ret %d\n", buf_size, sent);
#endif
		if (sent == buf_size) {
			/* Entire buf is sent */
			return (sent);
		} else if (sent > 0) {
			/* Partial buf is sent */
			cbuf += sent;
			buf_size -= sent;
		} else if (sent < 0) {
			if (errno != EAGAIN && errno != EWOULDBLOCK) {
				mc_log(LOG_ERR, "Error in send() to client fd %d: %d (%s) -- disconnecting the client", this->socket, errno, strerror(errno));
				disconnect_client();
				return (-1);
			}
		}
	}
	/* Queue the (remainder of) the buffer for sending by the network
	 * thread. This is an expensive operation. */
	assure_bufs(true);
	if (outbuf_pos + buf_size > outbuf_len) {
		/* The worst case - outbuf needs to be realloc()ed */
		char *new_outbuf;
		new_outbuf = (char*)realloc(outbuf, outbuf_pos + buf_size);
#ifndef NDEBUG
		if (internal_debug)
			_mc_debug_printf("Rellocing outbuf to %u (outbuf_pos=%d)\n", outbuf_pos + buf_size, outbuf_pos);
#endif
		if (new_outbuf == NULL)
			mc_log(LOG_CRIT, "Not enough memory to queue %u bytes to outbuf", buf_size);
		outbuf = new_outbuf;
	}
	outbuf_len += buf_size;
	memcpy(outbuf + outbuf_pos, cbuf, buf_size);
	enable_write_event();
	return (0);
}


/**
 * Sends a simple status packet.
 */
void
ServerClientConnection::send_status(int seq, int status, int data)
{
	struct mc_resp_status outpkt;

	mc_init_packet(&outpkt, sizeof outpkt, seq, MCRESP_STATUS);
	outpkt.status = status;
	outpkt.data = data;
	output(&outpkt, sizeof outpkt, true);
}


/**
 * Processes a verified READ event on the connection. Calls into
 * offload_command_packet if a whole packet is received. The nthr argument
 * can be NULL if there are no network threads.
 *
 * This method is either called from the main network thread in server.cpp
 * (in the SPED case), a ServerNetworkThread (SYMPED) or a ServerWorkerThread
 * (all other cases).
 */
enum EVENT_PROC_RESULT
ServerClientConnection::process_read_event(ServerNetworkThread *nthr, ServerModelMetadata *mm)
{
	char *buf = (char*)this->inbuf; /* Buffer to fill with data from this connection */
	ssize_t recv_size;

	assert(buf != NULL);	/* inbuf is preallocated */
	DEBUG_STATUS_POINT(st, 'r');

	if (this->type == CONN_TYPE_CLIENT_UNIX || this->type == CONN_TYPE_CLIENT_TCP) {
		recv_size = recv(this->socket, &buf[this->inbuf_pos], this->inbuf_len - this->inbuf_pos, 0);
		if (recv_size <= 0) {	/* Connection dropped */
			DEBUG_STATUS_POINT(st, 'D');
#if 0
			mc_log(LOG_ERR, "recv() returned %d - dropped connection?", recv_size);
#endif
			return EVPROC_ERROR;
		}
	} else {
		mc_log(LOG_CRIT, "Unimplemented process_read_event handler for connection type %d", this->type);
		return EVPROC_ERROR;
	}

	this->inbuf_pos += recv_size;
	if (this->inbuf_pos >= sizeof(struct mc_header)) {
		DEBUG_STATUS_POINT(st, 'H');
		/* We've received at least the command packet header, so inspect it. */
		if (this->inbuf->total_size > this->inbuf_len) {
			void *newbuf;
			/* At this point the header tells us this is a really long command packet.
			 * realloc() is slow but this will be the only one. */
			DEBUG_STATUS_POINT(st, 'A');
			newbuf = realloc(this->inbuf, this->inbuf->total_size);
			if (newbuf == NULL) {
				mc_log(LOG_ERR, "Out of memory receiving %u byte packet", this->inbuf->total_size);
				disconnect_client();
				return EVPROC_OK; /* Try to continue by ignoring the packet (???) */
			}
			this->inbuf = (struct mc_header*)newbuf;
			this->inbuf_len = this->inbuf->total_size;
			buf = (char*)this->inbuf;
		}
		if (this->inbuf_pos == this->inbuf->total_size) {
			/* We have received exactly one packet */
			DEBUG_STATUS_POINT(st, 'O');
			assert(this->inbuf_pos == this->inbuf->total_size);
			offload_command_packet(nthr, mm);
		} else if (this->inbuf_pos > this->inbuf->total_size) {
			/* The problem with chaining here is that the data is processed
			 * in separate threads operating directly on the this->inbuf
			 * buffers (which therefore cannot be changed in flight).
			 * This could possibly be made to work if worker threads do
			 * a dance after each packet is processed to memcpy the
			 * next packet to the front and reset the pointers, which
			 * would then need to be mutex-protected... */
			mc_log(LOG_ERR, "Packet chaining detected on %p - disconnecting", this);
			disconnect_client();
		}
	}
	DEBUG_STATUS_POINT(st, 'X');
	return EVPROC_OK;
}


/**
 * Queues a command packed received on specified connection for execution in
 * a worker thread (high level interface).
 */
void inline
ServerClientConnection::offload_command_packet(ServerNetworkThread *nthr, ServerModelMetadata *mm)
{
	/* Called when we've certainly got one whole command packet, maybe more,
	 * on the wire. This buffer (this->inbuf) will be directly inspected
	 * and acted upon. */
	DEBUG_STATUS_POINT(st, 'o');
	assert(this->inbuf != NULL);
	if (this->inbuf_pos == this->inbuf->total_size) {
		/* the best case - we've got exactly one command packet */
		queue_work(nthr, mm);
	} else {
		mc_log(LOG_CRIT, "Packet chaining on %p. Expected %d bytes, got %d", this, this->inbuf->total_size, this->inbuf_pos);
		disconnect_client();
	}
}


/**
 * Does the low-level queuing of a job - either queues it on a
 * worker thread or executes it in the current thread.
 */
void
ServerClientConnection::queue_work(ServerNetworkThread *nthr, ServerModelMetadata *mm)
{
	ServerWorkerThread *wthr;

	DEBUG_STATUS_POINT(st, 'q');
	if (n_worker_threads > 0) { /* pass the job to worker threads */
		wthr = nthr != NULL ? nthr->getWorkerThread() : &worker_threads[0];

		/* Network thread iterations (not network threads themselves)
		 * are bound to worker threads via wthr locks to minimize lock
		 * acquisitions. There's a dance we do here between this part
		 * of the code and the code in ServerNetworkThread WRT locking -
		 * only in case n_worker_threads > 0. Note that there's a special
		 * case here when n_network_threads == 0 (and nthr == NULL)
		 * and n_worker_threads > 0, in which case the main loop needs
		 * to signal the worker threads there's work to do. */
		if (wthr->getQueueLocker() != nthr) {
			DEBUG_STATUS_POINT(st, 'l');
			/* Spin until an available worker is found */
			assert(nthr != NULL);
			while (1) {
				wthr = nthr->getWorkerThread();
				if (wthr->queueTryLock() == 0) {
					break;
				}
				nthr->rollWorkerThread();
			}
			wthr->setQueueLocker(nthr);
		}
		if (n_network_threads == 0 && n_worker_threads > 0)
			wthr->queueLock();
		/* This queueing will eventually result in a call to process_command() */
		wthr->queueAdd(this);
	} else { 
		/* Either single-threaded operation (called from the main
		   network loop in server.cpp) or SYMPED (called from a
		   ServerNetworkThread). */
		process_command(mm);
	}
}


/**
 * Processes a verified WRITE event on a connection (i.e. write whatever is queued
 * for the connection).
 */
enum EVENT_PROC_RESULT
ServerClientConnection::process_write_event()
{
	char *buf = this->outbuf + this->outbuf_pos;
	ssize_t send_size = this->outbuf_len - this->outbuf_pos;
	ssize_t sent;

	DEBUG_STATUS_POINT(global_st, 'w');
	assert(this->outbuf != NULL);
	sent = send(this->socket, buf, send_size, MSG_NOSIGNAL);
#ifndef NDEBUG
	if (internal_debug)
		_mc_debug_printf("process_write_event: send(%zd) ret %zd, errno %d\n", send_size, sent, errno);
#endif
	if (sent == 0) {
		_mc_debug_printf("process_write_event: Sent 0 bytes to client socket %d (tried %d)\n", this->socket, send_size);
		return EVPROC_ERROR;
	} else if (sent == -1) {
		if (errno == EAGAIN || errno == EWOULDBLOCK)
			/* Linux is guilty of this: sending a POLLOUT-like event without buffers ready */
			return EVPROC_OK; /* Wait another round. */
		else {
			mc_log(LOG_ERR, "Error %d (%s) while sending to client socket %d -- disconnecting the client", errno, strerror(errno), this->socket);
			disconnect_client();
			return EVPROC_ERROR;
		}
	}
	this->outbuf_pos += sent;
	if (this->outbuf_pos == this->outbuf_len) {
		this->outbuf_pos = this->outbuf_len = 0;
		disable_write_event();
		free(this->outbuf);
		this->outbuf = NULL;
#ifndef NDEBUG
		if (internal_debug)
			_mc_debug_printf("process_write_event: done with buffer on fd %d\n", this->socket);
#endif
	}
	return EVPROC_OK;
}


/**
 * Enable the receiving of read events for this client-server (stream) socket.
 */
int
ServerClientConnection::enable_read_event()
{
	_mc_debug_printf(":: enabling READ on %d\n", socket);
	assert(myevq != NULL);
	assert(socket != 0);
	ev_list new_ev = ev_allocate_list(1);
	ev_set_list_element(new_ev, 0, socket, EV_READ, this);
	int rv = ev_enqueue_list(*myevq, new_ev, 1);
	ev_free_list(new_ev);
	return rv;
}


/**
 * Disable the receiving of read events for this client-server (stream) socket.
 */
int
ServerClientConnection::disable_read_event()
{
	_mc_debug_printf(":: disabling READ on %d\n", socket);
	assert(myevq != NULL);
	assert(socket != 0);
	ev_list new_ev = ev_allocate_list(1);
	ev_set_list_element_remove(new_ev, 0, socket, EV_READ, this);
	int rv = ev_enqueue_list(*myevq, new_ev, 1);
	ev_free_list(new_ev);
	return rv;
}


/**
 * Enable the receiving of write events for this client-server (stream) socket.
 */
int
ServerClientConnection::enable_write_event()
{
	_mc_debug_printf(":: enabling WRITE on %d\n", socket);
	assert(myevq != NULL);
	assert(socket != 0);
	ev_list new_ev = ev_allocate_list(1);
	ev_set_list_element(new_ev, 0, socket, EV_WRITE, this);
	int rv = ev_enqueue_list(*myevq, new_ev, 1);
	ev_free_list(new_ev);
	return rv;
}


/**
 * Disable the receiving of write events for this client-server (stream) socket.
 */
int
ServerClientConnection::disable_write_event()
{
	_mc_debug_printf(":: disabling WRITE on %d\n", socket);
	assert(myevq != NULL);
	assert(socket != 0);
	ev_list new_ev = ev_allocate_list(1);
	ev_set_list_element_remove(new_ev, 0, socket, EV_WRITE, this);
	int rv = ev_enqueue_list(*myevq, new_ev, 1);
	ev_free_list(new_ev);
	return rv;
}


/* ************************************************************************** */

/**
 * The big command switchboard: processes a command packet - dispatches
 * the packets to the command handlers.
 *
 * Called in several contexts: from process_read_event() in AMPED, SYMPED, SPED,
 * or directly from ServerWorkerThread in the case of the SEDA model.
 */
void
ServerClientConnection::process_command(ServerModelMetadata *mm)
{
	int mcmd = inbuf->mcmd;
	assert(socket > 0);
	mm->recordBufSize(((struct mc_header*)inbuf)->total_size);
	ops_count++;
	DEBUG_STATUS_POINT(st, 'C');
	_mc_debug_printf(">> process_command on %d: %d\n", socket, mcmd);
#ifdef LOCK_DEBUG
	if (!cache->allLocksClear())
		mc_log(LOG_CRIT, "entering process_command() locks held (cmd %d)", mcmd);
#endif
#ifdef LOCK_TRACE_ENABLE
	mc_log(LOG_DEBUG, "process_command: %d", mcmd);
#endif

	/* There's a race condition between outputting any response to the
	 * client and the client sending a new command into the same inbuf.
	 * Do not touch inbuf after outputting! */
	switch (mcmd) {
	case MCMD_HANDSHAKE:
		cmd_handshake();
		break;
	case MCMD_PUT_DATA:
		cmd_put_data();
		break;
	case MCMD_DEL_DATA_NAME:
		cmd_del_data_name();
		break;
	case MCMD_DEL_DATA_BY_TAG_VALUES:
		cmd_del_data_by_tag_values();
		break;
	case MCMD_GET_DATA_NAME:
		cmd_get_data_name();
		break;
	case MCMD_GET_DATA_BY_TAG_VALUES:
		cmd_get_data_by_tag_values();
		break;
	case MCMD_GET_ALL_KEYS:
		cmd_get_all_keys();
		break;
	case MCMD_ATOMIC_CMPSET:
	case MCMD_ATOMIC_ADD:
	case MCMD_ATOMIC_FETCHADD:
	case MCMD_ATOMIC_READANDCLEAR:
		cmd_atomic();
		break;
	case MCMD_MPUT:
		cmd_mput();
		break;
	case MCMD_MGET:
		cmd_mget();
		break;
	case MCMD_MDEL:
		cmd_mdel();
		break;
	case MCMD_TSTACK_PUSH:
		cmd_tstack_push();
		break;
	case MCMD_TSTACK_POP:
		cmd_tstack_pop();
		break;
	case MCMD_SET_TAGS:
		cmd_set_tags();
		break;
	default:
		cmd_error_inv_command();
		break;
	}
#ifdef LOCK_DEBUG
	if (!cache->allLocksClear())
		mc_log(LOG_CRIT, "process_command() about to return with locks held (cmd %d)", mcmd);
#endif
}


/**
 * Processes the handshake command.
 */
int
ServerClientConnection::cmd_handshake()
{
	struct mc_handshake *inpkt = (struct mc_handshake*)(inbuf);
	struct mc_resp_status outpkt;

	mc_init_packet(&outpkt, sizeof outpkt, inpkt->header.seq, MCRESP_STATUS);

	if (inpkt->endian_signature != MC_ENDIAN_SIGNATURE) {
		outpkt.status = MCSTATUS_ERROR;
		outpkt.data = MCREASON_ENDIANESS;
	} else if (inpkt->proto_ver_major != MC_VER_MAJOR) {
		outpkt.status = MCSTATUS_ERROR;
		outpkt.data = MCREASON_VERSION;
	} else {
		outpkt.status = MCSTATUS_OK;
		outpkt.data = 0;
	}
	output(&outpkt, sizeof outpkt, TRUE);
	return 0;
}


/**
 * Add a data key-value(+tags) tuple to the cache.
 */
int
ServerClientConnection::cmd_put_data()
{
	struct mc_add_data *inpkt = (struct mc_add_data*)(inbuf);
	struct dh_data *record;

	assert(inpkt != NULL);
	record = (struct dh_data*) malloc(sizeof(*record));
	if (record == NULL) {
	    mc_log(LOG_ERR, "malloc() failure");
	    return -1;
	}
	DataStore::fill_dh_data(record, inpkt);
	new_inbuf(); /* the current inbuf will be conserved in tuto in the dh tree */

	if (cache->put(record))
		send_status(inpkt->header.seq, MCSTATUS_OK, MCREASON_OK);
	else {
		mc_log(LOG_ERR, "Entry cannot be put in cache (various possible reasons)");
		send_status(inpkt->header.seq, MCSTATUS_ERROR, MCREASON_CACHE);
	}
	return 0;
}


/**
 * Get data by item name.
 */
int
ServerClientConnection::cmd_get_data_name()
{
	struct mc_get_data_name *inpkt = (struct mc_get_data_name*)(inbuf);
	struct mc_data_entry *outpkt;
	unsigned int lock_cookie;

	outpkt = cache->get_data_and_lock(inpkt->data, inpkt->name_size, &lock_cookie);
	if (outpkt != NULL) {
		mc_init_packet(outpkt, outpkt->header.total_size, inpkt->header.seq, MCRESP_DATA);
		output(outpkt, outpkt->header.total_size, TRUE);
		/* There is an awesome race condition between output() and any other
		 * operation that subsequently touches the inpkt, as it may be overwritten
		 * by the *next* incoming packet. So don't touch inpkt after output()ing. */
		cache->got_data_unlock(lock_cookie);
//		_mc_debug_printf("--lock_cookie: %d\n", lock_cookie);
	} else {
		send_status(inpkt->header.seq, MCSTATUS_ERROR, MCREASON_NOTFOUND);
	}
	return 0;
}


/**
 * Delete data by item name.
 */
int
ServerClientConnection::cmd_del_data_name()
{
	struct mc_del_data_name *inpkt = (struct mc_del_data_name*)(inbuf);

	if (cache->del_data_name(inpkt->data, inpkt->name_size))
		send_status(inpkt->header.seq, MCSTATUS_OK, MCREASON_OK);
	else
		send_status(inpkt->header.seq, MCSTATUS_ERROR, MCREASON_NOTFOUND);
	return 0;
}


/**
 * Get data by tags.
 */
int
ServerClientConnection::cmd_get_data_by_tag_values()
{
	struct mc_get_data_by_tag_values *inpkt = (struct mc_get_data_by_tag_values*)(inbuf);
	struct mc_resp_multidata *outpkt;

	outpkt = cache->get_multidata_by_tag_values(inpkt->tag_key, (tag_value_t*)inpkt->data, inpkt->n_values);
	if (outpkt != NULL) {
		mc_init_packet(outpkt, outpkt->header.total_size, inpkt->header.seq, MCRESP_MULTIDATA);
		output(outpkt, outpkt->header.total_size, TRUE);
		free(outpkt);
	} else
		send_status(inpkt->header.seq, MCSTATUS_ERROR, MCREASON_NOTFOUND);
	return 0;
}


/**
 * Handle MCMD_DEL_DATA_TAG_VALUES
 */
int
ServerClientConnection::cmd_del_data_by_tag_values()
{
	struct mc_del_data_by_tag_values *inpkt = (struct mc_del_data_by_tag_values*)(inbuf);
	int n_records;

	if ((n_records = cache->del_data_tag_values(inpkt->tag_key, inpkt->n_values, (tag_value_t*)inpkt->data)) != 0)
		send_status(inpkt->header.seq, MCSTATUS_OK, n_records);
	else
		send_status(inpkt->header.seq, MCSTATUS_ERROR, MCREASON_NOTFOUND);
	return 0;
}


/**
 Handle MCMD_GET_ALL_KEYS
*/
int
ServerClientConnection::cmd_get_all_keys()
{
	struct mc_simple *inpkt = (struct mc_simple*)(inbuf);
	struct mc_resp_multistring *outpkt;

	outpkt = cache->get_all_keys(inpkt->flags);
	if (outpkt != NULL) {
		mc_init_packet(outpkt, outpkt->header.total_size, inpkt->header.seq, MCRESP_MULTISTRING);
		output(outpkt, outpkt->header.total_size, TRUE);
		free(outpkt);
	} else
		send_status(inpkt->header.seq, MCSTATUS_ERROR, MCREASON_NOTFOUND);
	return 0;
}


/**
 * A common handler for atomic commands - because they have similar structures.
 */
int
ServerClientConnection::cmd_atomic()
{
	struct mc_atomic_op *inpkt = (struct mc_atomic_op*)(inbuf);
	struct mc_resp_atomic *outpkt;

	outpkt = cache->atomic_op(inpkt->header.mcmd, inpkt->name, inpkt->name_size, inpkt->arg1, inpkt->arg2);
	if (outpkt != NULL) {
		mc_init_packet(outpkt, outpkt->header.total_size, inpkt->header.seq, MCRESP_ATOMIC);
		output(outpkt, outpkt->header.total_size, TRUE);
		free(outpkt);
	} else
		send_status(inpkt->header.seq, MCSTATUS_ERROR, MCREASON_INVALID);
	return 0;
}


/**
 * Multi-put command handler - put multiple records into the datastore.
 *
 * @Input	struct mc_mput
 * @Output	status
 */
int
ServerClientConnection::cmd_mput()
{
	struct mc_mput *inpkt = (struct mc_mput*)(inbuf);
	struct dh_data *entries[inpkt->n_records];
	char *data = inpkt->data;

	/* Parse the incoming packet for individual mc_data_entries */
	for (unsigned i = 0; i < inpkt->n_records; i++) {
		struct mc_data_entry *mce0 = (struct mc_data_entry*)data;
		struct mc_data_entry *mce1 = (struct mc_data_entry*) malloc(mce0->header.total_size);

		memcpy(mce1, mce0, mce0->header.total_size);
		entries[i] = (struct dh_data*) malloc(sizeof(struct dh_data));
		DataStore::fill_dh_data(entries[i], mce1);

		data += mce0->header.total_size;
	}
	if (cache->mput(entries, inpkt->n_records, inpkt->flags))
		send_status(inpkt->header.seq, MCSTATUS_OK, MCREASON_OK);
	else
		send_status(inpkt->header.seq, MCSTATUS_ERROR, MCREASON_CACHE);
	return 0;
}


/**
 * Multi-get command handler - get multiple frecords from the datastore by
 * record keys.
 *
 * Input: struct mc_mget.
 * Output: struct mc_resp_multidata;
 */
int
ServerClientConnection::cmd_mget()
{
	struct mc_mget *inpkt = (struct mc_mget*)(inbuf);
	struct mc_key_name *keys[inpkt->n_names];
	struct mc_resp_multidata *outpkt;
	char *data = inpkt->data;

	for (unsigned i = 0; i < inpkt->n_names; i++) {
		keys[i] = (struct mc_key_name*)data;
		data += keys[i]->name_size + sizeof(keys[i]->name_size);
	}

	outpkt = cache->mget(keys, inpkt->n_names, inpkt->flags);
	if (outpkt != NULL) {
		mc_init_packet(outpkt, outpkt->header.total_size, inpkt->header.seq, MCRESP_MULTIDATA);
		output(outpkt, outpkt->header.total_size, TRUE);
		free(outpkt);
	} else
		send_status(inpkt->header.seq, MCSTATUS_ERROR, MCREASON_CACHE);
	return 0;
}


/**
 * Multi-del command handler - delete multiple frecords from the datastore by
 * record keys.
 *
 * Input: struct mc_mdel.
 * Output: struct mc_resp_status.
 */
int
ServerClientConnection::cmd_mdel()
{
	struct mc_mdel *inpkt = (struct mc_mdel*)(inbuf);
	struct mc_key_name *keys[inpkt->n_names];
	char *data = inpkt->data;
	int n_records;

	for (unsigned i = 0; i < inpkt->n_names; i++) {
		keys[i] = (struct mc_key_name*)data;
		data += keys[i]->name_size + sizeof(keys[i]->name_size);
	}

	if ((n_records = cache->mdel(keys, inpkt->n_names, inpkt->flags)) != 0)
		send_status(inpkt->header.seq, MCSTATUS_OK, n_records);
	else
		send_status(inpkt->header.seq, MCSTATUS_ERROR, MCREASON_CACHE);
	return 0;
}


/**
 * TStack push handler - creates a new cache record entry with a unique name and
 * the given data and tags. Returns the unique name created for the record.
 *
 * Input: struct mc_tstack_push.
 * Output: struct mc_resp_tstack_push.
 */
int
ServerClientConnection::cmd_tstack_push()
{
	struct mc_tstack_push *inpkt = (struct mc_tstack_push*)(inbuf);
	struct mc_resp_tstack_push *outpkt;
	uint64_t usrc;
	char name[32];
	struct dh_data *d;
	struct mc_data_entry *mce;
	struct mc_tag *tags;
	char *data;
	unsigned name_size;
	int err;
	
	/* This command requires extensive preparation but we will handle it as
	 * a normal mc_put on the backend. */
	usrc = 
		(uint64_t) socket << 54 |
		(uint64_t) (tsd_counter++) << 32 |
		(uint32_t)current_time;
	name_size = sprintf(name, "tstack__%llx", (long long unsigned int)usrc) + 1;
	
	tags = (struct mc_tag*) inpkt->data;
	data = inpkt->data + (inpkt->n_tags * sizeof(struct mc_tag));
	
	if ((err = mc_build_data_entry(&mce, name, name_size, data, inpkt->data_size, tags, inpkt->n_tags, 0, &(inpkt->header.seq))) != MCREASON_OK) {
		send_status(inpkt->header.seq, MCSTATUS_ERROR, err);
		return err;
	}
	d = (struct dh_data*)calloc(1, sizeof(*d));
	if (d == NULL) {
		send_status(inpkt->header.seq, MCSTATUS_ERROR, MCREASON_SERVER);
		return MCREASON_OK;
	}
	
	/* Both d and mce are recorded in the cache and must not be freed from here */
	DataStore::fill_dh_data(d, mce);
	cache->put(d);
	
	outpkt = (struct mc_resp_tstack_push*)malloc(sizeof(*outpkt) + name_size);
	outpkt->name_size = name_size;
	memcpy(outpkt->name, name, name_size);
	mc_init_packet(outpkt, sizeof(*outpkt) + name_size, inpkt->header.seq, MCRESP_TSTACK_PUSH);
	output(outpkt, outpkt->header.total_size, TRUE);
	free(outpkt);
	return 0;
}


/**
 * TStack pop operation. Pops a record (returns & removes it) from the virtual
 * stack identified by the given tag (key & value).
 *
 * Input: struct mc_tstack_pop.
 * Output: struct mc_data_entry.
 */
int
ServerClientConnection::cmd_tstack_pop()
{
	struct mc_tstack_pop *inpkt = (struct mc_tstack_pop*)(inbuf);
	struct mc_data_entry *outpkt;
	
	outpkt = cache->tstack_pop(inpkt->tag.key, inpkt->tag.value);
	if (outpkt == NULL) {
		send_status(inpkt->header.seq, MCSTATUS_ERROR, MCREASON_NOTFOUND);
		return 0;
	}
	outpkt->header.mcmd = MCRESP_DATA;
	output(outpkt, outpkt->header.total_size, TRUE);
	free(outpkt);
	return 0;
}


/**
 * (Re)set record tags.
 *
 * Input: struct mc_set_tags.
 * Output: struct mc_resp_status.
 */
int ServerClientConnection::cmd_set_tags()
{
	struct mc_set_tags *inpkt = (struct mc_set_tags*)(inbuf);
	struct mc_tag *tags;
	char *name;
	
	tags = (struct mc_tag*)inpkt->data;
	name = inpkt->data + inpkt->n_tags * sizeof(struct mc_tag);
	
	if (cache->set_tags(name, inpkt->name_size, tags, inpkt->n_tags))
		send_status(inpkt->header.seq, MCSTATUS_OK, 0);
	else
		send_status(inpkt->header.seq, MCSTATUS_ERROR, MCREASON_CACHE);
	return 0;
}


/**
 * Unknown command handler.
 */
int
ServerClientConnection::cmd_error_inv_command()
{
	struct mc_resp_status outpkt;

	mc_log(LOG_WARNING, "Client issued unknown command: %d", inbuf->mcmd);

	mc_init_packet(&outpkt, sizeof outpkt, inbuf->seq, MCRESP_STATUS);
	outpkt.status = MCSTATUS_ERROR;
	outpkt.data = MCREASON_INVALIDCMD;
	output(&outpkt, sizeof outpkt, TRUE);
	return 0;
}
