/*-
 * This file is a part of the MDCacheD server
 * (c) 2012. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: SizeHistogram.h 452 2012-03-20 16:03:25Z ivoras $
 */

#ifndef _SIZE_HISTOGRAM_H_
#define _SIZE_HISTOGRAM_H_

#include "misc.h"

/**
 * The SizeHistogram class implements a histogram of sizes binned as
 * powers-of-two, templatized by the N being the number of bins.
 * E.g. SizeHistogram<4> has bins 1,2,4,8.
 */
template <unsigned N>
class SizeHistogram {
private:
	unsigned	histogram[N];
public:
	SizeHistogram();
	unsigned 	getSize() { return N; }
	void 		addPoint(unsigned x);
	unsigned	getPoint(unsigned bin) { return histogram[bin]; }
	void		clear() { memset(&histogram, 0, sizeof(histogram)); };
	void		mergeHistogram(SizeHistogram<N> *h);
	unsigned	getPeak();
	int		getPeakBin();
	unsigned	getPeakSize();
	int		getMaxBin();
	void		getCopy(unsigned *h) { memcpy(h, histogram, sizeof(histogram)); }
};

template <unsigned N>
SizeHistogram<N>::SizeHistogram() {
	clear();
}

template <unsigned N>
void SizeHistogram<N>::addPoint(unsigned x) {
	unsigned p = po2(x);
	if (!ISPO2(x))
		p++;
	if (p >= N)
		p = N - 1;
	 histogram[p]++;
}

template <unsigned N>
void SizeHistogram<N>::mergeHistogram(SizeHistogram *h) {
	for (unsigned i = 0; i < N; i++)
		histogram[i] += h->histogram[i];
}

template <unsigned N>
unsigned SizeHistogram<N>::getPeak() {
	unsigned m = 0;
	for (unsigned i = 0; i < N; i++)
		if (histogram[i] >= m)
			m = histogram[i];
	return m;
}

template <unsigned N>
int SizeHistogram<N>::getPeakBin() {
	unsigned m = 0, mi = 0;
	for (unsigned i = 0; i < N; i++)
		if (histogram[i] >= m) {
			m = histogram[i];
			mi = i;
		}
	if (m != 0)
		return mi;
	else
		return -1;
}

template <unsigned N>
unsigned SizeHistogram<N>::getPeakSize() {
	return 1 << getPeakBin();
}

template <unsigned N>
int SizeHistogram<N>::getMaxBin() {
	int m = -1;
	for (unsigned i = 0; i < N; i++)
		if (histogram[i] != 0)
			m = i;
	return m;
}


#endif
