/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2011. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: Thread.h 452 2012-03-20 16:03:25Z ivoras $
 */

#ifndef _THREAD_H_
#define _THREAD_H_

#include <pthread.h>
#include <signal.h>
#include <exception>

class Thread {
public:
	Thread() {
		stop_requested = running = false;
		thread_name = NULL;
		sigemptyset(&sset);	/* Default to no signals to this thread */
		is_detached = false;
	}
	
	virtual ~Thread() {
		if (thread_name != NULL)
			free(thread_name);
	}
	
	void setDetached(bool detached) {
		is_detached = detached;
	}
	
	bool getDetached() {
		return is_detached;
	}
	
	void start() {
		assert(running == false);
		
		pthread_attr_t pa;
		pthread_attr_init(&pa);
		pthread_attr_setdetachstate(&pa, is_detached ? 1 : 0);
		pthread_create(&thr, &pa, &Thread::start_thread, this);
		pthread_attr_destroy(&pa);
#ifdef HAS_PTHREAD_SET_NAME_NP
		if (thread_name != NULL)
			pthread_set_name_np(thr, thread_name);
#endif
	}
	
	void request_stop() {
		stop_requested = true;
		//pthread_kill(thr, SIGHUP);
	}
	
	void join() {
		void *rval;
		if (pthread_join(thr, &rval) != 0)
			warnx("Thread::join() failed on %p", this);
	}
	
	void set_name(const char *name)
	{
		if (thread_name != NULL)
			free(thread_name);
		thread_name = strdup(name);
#ifdef HAS_PTHREAD_SET_NAME_NP
		if (running)
			pthread_set_name_np(thr, thread_name);
#endif
	}

	/** Adds a signal to the signal mask of signal this thread catches */
	void add_signal(int sig) {
		if (running)
			warnx("Signals must be set before the thread is running");
		sigaddset(&sset, sig);
	}
	
	/** Removes a signal from the thread's signal mask */
	void del_signal(int sig) {
		if (running)
			warnx("Signals must be set before the thread is running");
		sigdelset(&sset, sig);
	}
	
	bool is_running() {
		return running;
	}
	
	virtual void run() {}
protected:
	pthread_t thr;
	sigset_t sset;
	bool stop_requested;
	bool running;
	char *thread_name;
	bool is_detached;
private:	
	static void* start_thread(void *obj) {
		Thread *thr = reinterpret_cast<Thread*>(obj);

		thr->running = true;
		if (pthread_sigmask(SIG_SETMASK, &thr->sset, NULL) != 0)
			warnx("Cannot set signals mask for thread at %p", thr);

		thr->run();
		thr->running = false;
		return (NULL);
	}

};

#endif
