/*
 * MemcacheD multi-client benchmark
 * (c) 2008.-2012. Ivan Voras <ivoras@gmail.com>
 * $Id: bench.c 445 2012-03-11 22:52:11Z ivoras $
 */
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <time.h>

#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include <errno.h>
#include <string.h>
#include <paths.h>
#include <syslog.h>
#include <assert.h>
#include <err.h>

#include <memcache.h>
#include <credis.h>

#include "mc_protocol.h"
#include "mc_client.h"
#include "bench_utils.h"
#include "common_utils.h"

#ifndef TRUE
#define TRUE 1
#define FALSE (!TRUE)
#endif

enum BENCH_TYPE { BENCH_MDCACHED, BENCH_MEMCACHED, BENCH_REDIS };

#define DEFAULT_RAND_SEED 334455
#define DEFAULT_RECORD_DURATION 10*60

char *server_host = "localhost";
int server_port = MC_DEFAULT_INET_PORT;
int is_unix_socket = TRUE;
unsigned n_data = N_DATA;
unsigned n_requests = N_REQUESTS;
int n_clients = 2;	/* number of client threads */
int percent_read = 90; 	/* reader:writer ratio */
int verify_data = 0;
int verbose = FALSE;
float scale = 1;
unsigned record_duration = DEFAULT_RECORD_DURATION;
pthread_t *thrs;	/* all the client threads */
enum BENCH_TYPE btype = BENCH_MDCACHED;


struct ddata {
	char *key;
	unsigned int key_size;
	char *data;
	unsigned int data_size;
	int been_put;
};

struct ddata *gendata;


static inline char* bench_get_type(enum BENCH_TYPE btype)
{
	switch(btype) {
	case BENCH_MDCACHED:
		return "mdcached";
	case BENCH_MEMCACHED:
		return "memcached";
	case BENCH_REDIS:
		return "redis";
	default:
		errx(1, "Unknown benchmark type %d", btype);
		return NULL;
	}
}


void *cached_client_thread(void *arg)
{
	intptr_t thr_id;
	int rv;
	unsigned int i;
	char conn_string[80];
	struct memcache *memc = NULL;
	struct mc_connection *mdc = NULL;
	REDIS rh = NULL;
	char *err;

	thr_id = (intptr_t)pthread_self();

	if (btype == BENCH_MEMCACHED) {
		if (is_unix_socket)
			errx(1, "Memcached doesn't work over a unix socket yet.");
		sprintf(conn_string, "%s:%d", server_host, server_port);
		memc = mc_new();
		if (memc == NULL)
			errx(1, "Cannot create a mc object");
		mc_err_filter_del(MCM_ERR_LVL_INFO);
		mc_err_filter_del(MCM_ERR_LVL_NOTICE);
		mc_server_add4(memc, conn_string);
	} else if (btype == BENCH_MDCACHED) {
		if (is_unix_socket) {
			server_host = MC_DEFAULT_UNIX_SOCKET;
			mdc = mc_connect_local(server_host, TRUE, &err);
		} else
			mdc = mc_connect_tcp(server_host, server_port, TRUE, &err);
		if (mdc == NULL)
			errx(1, "Connection error: %s%s", err, is_unix_socket ? " (unix socket)" : "");

	} else if (btype == BENCH_REDIS) {
		rh = credis_connect(server_host, server_port, 1000);
		if (rh == NULL)
			errx(1, "Connection error to redis server on %s:%d", server_host, server_port);
		if (credis_ping(rh) == -1)
			errx(1, "Redis ping failure: %s", credis_errorreply(rh));
	} else
		errx(1, "Unknown benchmark type %d", btype);

	if (verbose)
		printf("Created %s connection from client %zx\n",
			bench_get_type(btype), thr_id);

	for (i = 0; i < n_requests; i++) {
		int r;
		int is_write;
		if (n_data > 1)
			r = rand() % n_data; /* choose an element */
		else
			r = 0;

		is_write = (rand() % 100) >= percent_read;
		if (is_write) {
			/* *** WRITE BENCHMARK *** */
			if (btype == BENCH_MEMCACHED) {
				if ((rv = mc_set(memc, gendata[r].key, gendata[r].key_size, gendata[r].data, gendata[r].data_size, record_duration, 0)) != MCM_ERR_NONE)
					errx(1, "Error in mc_set: %d", rv);
			} else if (btype == BENCH_MDCACHED) {
				if (mc_put_simple(mdc, gendata[r].key, gendata[r].key_size, gendata[r].data, gendata[r].data_size, record_duration, &err) != MC_RESULT_OK)
					errx(1, "Error in mc_put_simple: %s", err);
			} else if (btype == BENCH_REDIS) {
				if (credis_set(rh, gendata[r].key, gendata[r].data) == -1)
					errx(1, "Error in credis_set: %s", credis_errorreply(rh));
			} else
				errx(1, "Unknown benchmark type: %d", btype);
			gendata[r].been_put = TRUE;
		} else {
			/* *** READ BENCHMARK *** */
			char *data = NULL;
			unsigned int data_size;
			size_t memc_size;

			if (btype == BENCH_MEMCACHED) {
				data = mc_aget2(memc, gendata[r].key, gendata[r].key_size, &memc_size);
				data_size = memc_size;
				if (data == NULL && gendata[r].been_put)
					printf("Record has been put to cache but cannot be found: '%s' (ignore this)\n", gendata[r].key);
			} else if (btype == BENCH_MDCACHED) {
				rv = mc_get_simple(mdc, gendata[r].key, gendata[r].key_size, &data, &data_size, &err);
				if (rv != MC_RESULT_OK) {
					if (err != NULL)
						errx(1, "mc error: %s", err);
					else
						if (gendata[r].been_put)
							printf("Record has been put to cache but cannot be found: '%s' (ignore this)\n", gendata[r].key);
				}
			} else if (btype == BENCH_REDIS) {
				if (credis_get(rh, gendata[r].key, &data) == -1)
					if (gendata[r].been_put)
						printf("Record has been put to cache but cannot be found: '%s' (ignore this)\n", gendata[r].key);
				data = NULL; /* Apparently Redis manages the buffer */

			} else
				errx(1, "Unknown benchmark type %d", btype);
			if (data != NULL) {
				if (verify_data)
					if (memcmp(gendata[r].data, data, data_size) != 0)
						errx(1, "Data comparison fails for key %s", gendata[r].key);
				free(data);
			}
		}
	}
	return NULL;
}


int main(int argc, char **argv)
{
	int ch;
	double t1, t2, t;
	int i, count;
	unsigned int mem_key_size = 0, mem_data_size = 0;

	srand(DEFAULT_RAND_SEED); /* we want repeatable runs */

	while ((ch = getopt(argc, argv, "b:c:t:s:p:r:hVvi:n:T:")) != -1) {
		switch (ch) {
		case 'h':
			printf("usage: %s [-c server] [-p port] [-s #] [-i #] [-n #] [-h] [-v] [-r %%read] [-R #] [-V]\n", argv[0]);
			printf("	-c server	Server to connect to\n");
			printf("			(default = local unix for mdcached, localhost for others)\n");
			printf("	-s #		Scale of tests (larger # -> test more data)\n");
			printf("	-p port		TCP port (usable only with -c)\n");
			printf("	-r %%read	Percentage of read requests (1-99)\n");
			printf("	-t n_clients	Number of client threads to spawn\n");
			printf("	-b bench	Run some other benchmark instead of mdcached\n");
			printf("			(options are: mdcached, memcached, redis)\n");
			printf("	-i #		Number of iterations (default is %u)\n", n_requests);
			printf("	-n #		Number of records in the pool (default is %u)\n", n_data);
			printf("	-T #		Record expiry time (default is %u seconds)\n", record_duration);
			printf("	-R #		Seed the rand() number generator with #\n");
			printf("			(default seed is %d)\n", DEFAULT_RAND_SEED);
			printf("	-V		Verify fetched data\n");
			exit(0);
			break;
		case 'c':
			server_host = optarg;
			is_unix_socket = FALSE;
			break;
		case 'b':
			if (strcmp(optarg, "mdcached") == 0) {
				btype = BENCH_MDCACHED;
				is_unix_socket = TRUE;
				server_port = MC_DEFAULT_INET_PORT;
			} else if (strcmp(optarg, "memcached") == 0) {
				btype = BENCH_MEMCACHED;
				if (server_port == MC_DEFAULT_INET_PORT)
					server_port = 11211;
				is_unix_socket = FALSE;
			} else if (strcmp(optarg, "redis") == 0) {
				btype = BENCH_REDIS;
				is_unix_socket = FALSE;
				if (server_port == MC_DEFAULT_INET_PORT)
					server_port = 6379;
			} else
				errx(1, "Recognized bencchmark types are: mdcached, memcached, redis");
			break;
		case 'p':
			server_port = atoi(optarg);
			if (server_port < 1 || server_port > 65535)
				errx(1, "Invalid port %d", server_port);
			if (server_port > 0)
				is_unix_socket = FALSE;
			break;
		case 'r':
			percent_read = atoi(optarg);
			if (percent_read < 1 || percent_read > 99)
				errx(1, "Percentage invalid: %d", percent_read);
			break;
		case 't':
			n_clients = atoi(optarg);
			if (n_clients < 1)
				errx(1, "Invalid number of clients: %d", n_clients);
			break;
		case 'i':
			n_requests = atoi(optarg);
			break;
		case 'n':
			n_data = atoi(optarg);
			break;
		case 's':
			scale = atof(optarg);
			break;
		case 'R':
			srand(atoi(optarg));
			break;
		case 'T':
			record_duration = atoi(optarg);
			break;
		case 'V':
			verify_data = TRUE;
			break;
		case 'v':
			verbose++;
			break;
		default:
			fprintf(stderr, "Unknown argument: %c\n", ch);
			exit(1);
		}
	}

	if (RAND_MAX < n_data)
		errx(1, "RAND_MAX is too small (%d)", RAND_MAX);
	if (verbose)
		printf("Generating %u data records for the pool.\n", n_data);

	gendata = calloc(n_data, sizeof *gendata);
	for (i = 0; i < n_data; i++) {
		gen_data(i, &gendata[i].key, &gendata[i].key_size, &gendata[i].data, &gendata[i].data_size, scale);
		mem_key_size += gendata[i].key_size;
		mem_data_size += gendata[i].data_size;
	}

	printf("Generated %u records: %u KB (%u KB keys, %u KB data). Starting %d threads.\n", n_data,
		(mem_key_size + mem_data_size) / 1024, mem_key_size / 1024, mem_data_size / 1024, n_clients);
	printf("Average data size: %u bytes.\n", mem_data_size / n_data);
	printf("Benchmark type: %s\n", bench_get_type(btype));

	if (n_clients > 1) {
		thrs = malloc(n_clients * sizeof *thrs);
		for (i = 0; i < n_clients; i++)
			if (pthread_create(&thrs[i], NULL, cached_client_thread, &thrs[i]) != 0)
				err(1, "pthread_create() failed");

		t1 = now();
		printf("Waiting for threads (%u iterations).\n", n_requests);
		for (i = 0; i < n_clients; i++)
			pthread_join(thrs[i], NULL);
		t2 = now();
	} else {
		t1 = now();
		cached_client_thread(NULL);
		t2 = now();
	}
	t = t2 - t1;

	printf("%d clients * %u requests = %u requests: %0.2f seconds. That's %0.1lf requests/s\n",
		n_clients, n_requests, n_clients * n_requests, t, (n_clients * n_requests) / t);

	count = 0;
	for (i = 0; i < n_data; i++)
		if (gendata[i].been_put)
			count++;
	printf("%d records from the record pool have been put.\n", count);

	return 0;
}
