/*
 * MDCacheD multi-client benchmark
 * (c) 2008. Ivan Voras <ivoras@gmail.com>
 * $Id: bench_mdcached.c 380 2011-11-10 00:19:34Z ivoras $
 */
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <time.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include <errno.h>
#include <string.h>
#include <paths.h>
#include <syslog.h>
#include <assert.h>
#include <err.h>

#include "mc_protocol.h"
#include "mc_client.h"
#include "common_utils.h"
#include "bench_utils.h"

char *server_host = MC_DEFAULT_UNIX_SOCKET;
int server_port = MC_DEFAULT_INET_PORT;
int is_unix_socket = TRUE;
int n_clients = 2;	/* number of client threads */
int percent_read = 90; 	/* reader:writer ratio */
float scale = 1;
pthread_t *thrs;

//#define N_DATA 10000
//#define N_REQUESTS 100000

struct ddata {
	char *key;
	unsigned int key_size;
	char *data;
	unsigned int data_size;
	int been_put;
};

struct ddata *gendata;


void *client_thread(void *arg) 
{
	struct mc_connection *conn;
	int thr_id;
	unsigned int i;
	char *err = NULL;

	thr_id = (intptr_t)pthread_self();

	if (is_unix_socket)
		conn = mc_connect_local(server_host, TRUE, &err);
	else
		conn = mc_connect_tcp(server_host, server_port, TRUE, &err);
	if (conn == NULL)
		errx(1, "connection error: %s", err);

	printf("Created connection from client %x (including handshake)\n", thr_id);

	for (i = 0; i < N_REQUESTS; i++) {
		int r;
		int is_write;
#if N_DATA > 1
			r = rand() % N_DATA; /* choose an element */
#else
			r == 0;
#endif
		is_write = (rand() % 100) >= percent_read;	/* is it a read or a write request? */
		if (is_write) {
			if (mc_put_simple(conn, gendata[r].key, gendata[r].key_size, gendata[r].data, gendata[r].data_size, 3600, &err) != MC_RESULT_OK)
				errx(1, "Error in mc_put_simple: %s", err);
			gendata[r].been_put = TRUE;
		} else {
			char *data;
			unsigned int data_size;
			int res;

			res = mc_get_simple(conn, gendata[r].key, gendata[r].key_size, &data, &data_size, &err);
			if (res != MC_RESULT_OK) {
				if (err != NULL)
					errx(1, "Error in mc_get_simple: %d, %s (%s)\n", res, err, mc_result_to_string(res));
				else
					if (gendata[r].been_put)
						printf("Record has been put to cache but cannot be found: '%s' (ignore this)\n", gendata[r].key);
			} else {
				/*
				if (memcmp(gendata[r].data, data, data_size) != 0)
					errx(1, "Data comparison fails for key %s", gendata[r].key);
				*/
				free(data);
			}
		}
	}
	return NULL;
}



int main(int argc, char **argv)
{
	int ch;
	double t1, t2;
	int i, count;
	unsigned int mem_key_size = 0, mem_data_size = 0;

	if (RAND_MAX < N_DATA)
		errx(1, "RAND_MAX is too small.");
	srand(334455); /* we want repeatable runs */

	while ((ch = getopt(argc, argv, "c:t:s:p:r:h")) != -1) {
		switch (ch) {
		case 'h':
			printf("usage: %s [-c server] [-p port] [-s #] [-h] [-r %%read]\n", argv[0]);
			printf("	-c server	Server to connect to (default = local unix socket)\n");
			printf("	-s #		Scale of tests (larger # -> test larger data)\n");
			printf("	-p port		TCP port (usable only with -c)\n");
			printf("	-r %%read	Percentage of read requests (1-99)\n");
			printf("	-t n_clients	Number of client threads to spawn\n");
			exit(0);
			break;
		case 'c':
			server_host = optarg;
			is_unix_socket = FALSE;
			break;
		case 'p':
			server_port = atoi(optarg);
			if (server_port < 1 || server_port > 65535)
				errx(1, "Invalid port %d", server_port);
			break;
		case 'r':
			percent_read = atoi(optarg);
			if (percent_read < 1 || percent_read > 99)
				errx(1, "Percentage invalid: %d", percent_read);
			break;
		case 't':
			n_clients = atoi(optarg);
			if (n_clients < 1)
				errx(1, "Invalid number of clients: %d", n_clients);
			break;
		case 's':
			scale = atof(optarg);
			break;
		default:
			fprintf(stderr, "Unknown argument: %c\n", ch);
			exit(1);
		}
	}


	printf("Generating %u data records.\n", N_DATA);

	gendata = calloc(N_DATA, sizeof *gendata);
	for (i = 0; i < N_DATA; i++) {
		gen_data(i, &gendata[i].key, &gendata[i].key_size, &gendata[i].data, &gendata[i].data_size, scale);
		mem_key_size += gendata[i].key_size;
		mem_data_size += gendata[i].data_size;
	}
	
	printf("Generated %u KB (%u KB keys, %u KB data). Starting %d threads.\n", (mem_key_size + mem_data_size) / 1024, mem_key_size / 1024, mem_data_size / 1024, n_clients);
	printf("Average data size: %u bytes.\n", mem_data_size / N_DATA);

	if (n_clients > 1) {
		thrs = malloc(n_clients * sizeof *thrs);
		for (i = 0; i < n_clients; i++)
			pthread_create(&thrs[i], NULL, client_thread, &thrs[i]);

		t1 = now();
		printf("Waiting for threads (%u iterations).\n", N_REQUESTS);
		for (i = 0; i < n_clients; i++)
			pthread_join(thrs[i], NULL);
		t2 = now();
	} else {
		t1 = now();
		client_thread(NULL);
		t2 = now();
	}

	printf("%d clients * %u requests = %u requests. That's %0.1lf requests/s\n",
		n_clients, N_REQUESTS, n_clients * N_REQUESTS, (n_clients * N_REQUESTS) / (t2 - t1));

	count = 0;
	for (i = 0; i < N_DATA; i++)
		if (gendata[i].been_put)
			count++;
	printf("%d records have been put.\n", count);

	return 0;
}
