/*
 * MDCacheD multi-client benchmark
 * (c) 2008. Ivan Voras <ivoras@gmail.com>
 * $Id: bench_mdcached_async.c 387 2011-11-11 15:27:28Z ivoras $
 */
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <time.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include <errno.h>
#include <string.h>
#include <paths.h>
#include <syslog.h>
#include <assert.h>
#include <err.h>

#include "mc_protocol.h"
#include "mc_client.h"
#include "common_utils.h"
#include "bench_utils.h"
#include "minilibevent.h"

char *server_host = MC_DEFAULT_UNIX_SOCKET;
int server_port = MC_DEFAULT_INET_PORT;
int is_unix_socket = TRUE;
int percent_read = 90; 	/* reader:writer ratio */
float scale = 1;

struct ddata {
	char *key;
	unsigned int key_size;
	char *data;
	unsigned int data_size;
	int been_put;
};

struct ddata *gendata;

int n_clients = 10;		/* number of clients (connections) */
struct mc_connection **conns;	/* master connection list */

struct client {
	struct mc_connection *conn;
	unsigned int reqno;
	enum { REQ_PUT, REQ_GET } reqtype;
};

struct bench_thread {
	pthread_t thread;
	unsigned int n_clients;
	unsigned int clients_count;
	struct client *clients;
};

struct bench_thread *threads;
int n_threads = 2;

struct mc_connection *mkconn()
{
	char *err;
	struct mc_connection *conn;
	
	if (is_unix_socket)
		conn = mc_connect_local(server_host, TRUE, &err);
	else
		conn = mc_connect_tcp(server_host, server_port, TRUE, &err);
	if (conn == NULL)
		errx(1, "connection error: %s", err);
	mc_set_connection_async(conn);
	return conn;
}


/**
 * Initiate the next request on the client. It can be a PUT or a GET.
 */
static void next_request(struct client *cl)
{
	int is_write = (rand() % 100) >= percent_read;	/* is it a read or a write request? */
	int r = rand() % N_DATA;
	int res;
	char *err = NULL;
	
	if (is_write) {
		res = mc_put_simple_async(cl->conn, gendata[r].key, gendata[r].key_size, gendata[r].data, gendata[r].data_size, 3600, &err);
		if (res == MC_RESULT_PENDING)
			while ((res = mc_async_sendpoll(cl->conn)) == MC_RESULT_PENDING) {}
		if (res != MC_RESULT_OK)
			errx(1, "Error in mc_put_simple_async: %s", err);
		gendata[r].been_put = TRUE;
		cl->reqtype = REQ_PUT;
	} else {
		res = mc_get_simple_async(cl->conn, gendata[r].key, gendata[r].key_size, &err);
		if (res == MC_RESULT_PENDING)
			while ((res = mc_async_sendpoll(cl->conn)) == MC_RESULT_PENDING) {}
		if (res != MC_RESULT_OK)
			errx(1, "Error in mc_get_simple_async: %s (%s)", err, mc_result_to_string(res));
		cl->reqtype = REQ_GET;
	}
}

void *client_thread(void *arg) 
{
	struct bench_thread *thr = arg;
	unsigned int i, max_requests, n_requests, n_events, n_iter;
	char *err;
	ev_queue evq;
	ev_list evl;

	printf("Created thread %p (including handshake) with %d connections\n", thr, thr->n_clients);
	
	/* Create IO queue */
	evq = ev_create_queue();
	evl = ev_allocate_list(thr->n_clients);
	
	for (i = 0; i < thr->n_clients; i++) {
		assert(thr->clients);
		assert(thr->clients[i].conn);
		ev_set_list_element(evl, i, thr->clients[i].conn->sock, EV_READ, &thr->clients[i]);
	}
	ev_enqueue_list(evq, evl, thr->n_clients);
	
	/* Get a first round of requests going - this will start the async chain of events */
	for (i = 0; i < thr->n_clients; i++) {
		int r;
		if ((r = mc_put_simple_async(thr->clients[i].conn, gendata[i].key, gendata[i].key_size, gendata[i].data, gendata[i].data_size, 3600, &err)) != MC_RESULT_OK)
			errx(1, "Error sending initial data to client #%d on thread %p: %s, %d", i, thr, err, r);
		thr->clients[i].reqtype = REQ_PUT;
	}
	max_requests = thr->n_clients * N_REQUESTS;
	n_requests = n_events = n_iter = 0;

	while (1) {
		int r, err;
		struct client *cl;
		struct mc_header *pkt = NULL;
		
		r  = ev_get_new_events(evq, evl, thr->n_clients, NULL);
		n_events += r;
		
		for (i = 0; i < r; i++) {
			cl = evl[i].udata;
			if (evl[i].filter == EV_READ) {
				if ((err = mc_async_recvpoll(cl->conn, &pkt)) == MC_RESULT_OK) {
					if (cl->reqtype == REQ_PUT) {
						struct mc_resp_status *presp = (struct mc_resp_status*)pkt;
						if (presp->status != MCSTATUS_OK)
							errx(1, "Error in received status");
						free(pkt);
						/* We have just completed the first step of PUT(command) -> GET(data) chain of events. */
						/* We need to initiate a new request now. The new request will be randomly chosen. */
						next_request(cl);
						n_requests++;
					} else if (cl->reqtype == REQ_GET) {
						struct mc_resp_data *pget = (struct mc_resp_data*)pkt;
						if (pget->header.mcmd == MCRESP_DATA) {
							/* all is well. XXX: check if received data is valid */
						} else if (pget->header.mcmd == MCRESP_STATUS) {
							struct mc_resp_status *status = (struct mc_resp_status*)pkt;
							if (status->status == MCSTATUS_ERROR && status->data == MCREASON_NOTFOUND)
							{
								/* all is well. XXX: check if the data should have been found. */
							} else
								errx(1, "Unknown status: %d,%d", status->status, status->data);
						}
						free(pkt);
						/* We have just completed the second step of PUT(command) -> GET(data) chain of events. */
						/* We need to initiate a new request now. The new request will be randomly chosen. */
						next_request(cl);
						n_requests++;
					}
				} else if (err != MC_RESULT_PENDING) {
					fprintf(stderr, "async_recvpoll returned %d (%s)\n", err, mc_result_to_string(err));
					printf("conn type: %d, state: %d\n", cl->conn->type, cl->conn->async_state);
					abort();
				}
			}
		}
		if (n_requests > max_requests)
			break;
		n_iter++;
	}
	printf("Thread %p ends: %u events in %u iterations: %u events/iteration\n", thr, n_events, n_iter, n_events/n_iter);
	return NULL;
}



int main(int argc, char **argv)
{
	int ch;
	double t1, t2;
	int i, count;
	unsigned int mem_key_size = 0, mem_data_size = 0;
	unsigned int min_data_size = 10000000, max_data_size = 0;

	if (RAND_MAX < N_DATA)
		errx(1, "RAND_MAX is too small.");
	srand(334455); /* we want repeatable runs */

	while ((ch = getopt(argc, argv, "c:t:n:s:p:r:h")) != -1) {
		switch (ch) {
		case 'h':
			printf("usage: %s [-c server] [-p port] [-s #] [-h] [-r %%read]\n", argv[0]);
			printf("	-c server	Server to connect to (default = local unix socket)\n");
			printf("	-s #		Scale of tests (larger # -> test larger data)\n");
			printf("	-p port		TCP port (usable only with -c)\n");
			printf("	-r %%read	Percentage of read requests (1-99)\n");
			printf("	-n n_clients	Number of clients to connect\n");
			printf("	-t n_threads	Number of threads to spawn (clients will be distributed across threads)\n");
			exit(0);
			break;
		case 'c':
			server_host = optarg;
			is_unix_socket = FALSE;
			break;
		case 'p':
			server_port = atoi(optarg);
			if (server_port < 1 || server_port > 65535)
				errx(1, "Invalid port %d", server_port);
			break;
		case 'r':
			percent_read = atoi(optarg);
			if (percent_read < 1 || percent_read > 99)
				errx(1, "Percentage invalid: %d", percent_read);
			break;
		case 't':
			n_threads = atoi(optarg);
			if (n_threads < 1)
				errx(1, "Invalid number of threads: %d", n_threads);
			break;
		case 'n':
			n_clients = atoi(optarg);
			if (n_clients < 1)
				errx(1, "Invalid number of clients: %d", n_clients);
			break;
		case 's':
			scale = atof(optarg);
			break;
		default:
			fprintf(stderr, "Unknown argument: %c\n", ch);
			exit(1);
		}
	}

	printf("Generating %u data records.\n", N_DATA);

	gendata = calloc(N_DATA, sizeof *gendata);
	for (i = 0; i < N_DATA; i++) {
		gen_data(i, &gendata[i].key, &gendata[i].key_size, &gendata[i].data, &gendata[i].data_size, scale);
		mem_key_size += gendata[i].key_size;
		mem_data_size += gendata[i].data_size;
		if (gendata[i].data_size < min_data_size)
			min_data_size = gendata[i].data_size;
		if (gendata[i].data_size > max_data_size)
			max_data_size = gendata[i].data_size;
	}
	
	printf("Generated %u kB (%u kB keys, %u kB data). Starting %d connections in %d threads.\n", (mem_key_size + mem_data_size) / 1024, mem_key_size / 1024, mem_data_size / 1024, n_clients, n_threads);
	printf("Minimum data size: %d, Maximum data size: %d\n", min_data_size, max_data_size);
	printf("Average key size: %u bytes.\n", mem_key_size / N_DATA);
	printf("Average data size: %u bytes.\n", mem_data_size / N_DATA);

	threads = calloc(n_threads, sizeof(*threads));
	conns = calloc(n_clients, sizeof(*conns));
	for (i = 0; i < n_clients; i++) {
		struct bench_thread *thr;
		int cl;
		
		conns[i] = mkconn();
		assert(conns[i]);
		thr = &threads[i % n_threads];
		
		cl = thr->clients_count++;
		thr->clients = realloc(thr->clients, thr->clients_count * sizeof(*thr->clients));
		assert(thr->clients);
		
		thr->clients[cl].conn = conns[i];
		thr->clients[cl].reqno = 0;
		thr->n_clients = thr->clients_count;
		assert(thr->clients[cl].conn);
	}

	for (i = 0; i < n_threads; i++)
		pthread_create(&threads[i].thread, NULL, client_thread, &threads[i]);
	
	t1 = now();
	for (i = 0; i < n_threads; i++)
		pthread_join(threads[i].thread, NULL);
	t2 = now();

	printf("%d clients * %u requests = %u requests. That's %0.1lf requests/s\n",
		n_clients, N_REQUESTS, n_clients * N_REQUESTS, (n_clients * N_REQUESTS) / (t2 - t1));

	count = 0;
	for (i = 0; i < N_DATA; i++)
		if (gendata[i].been_put)
			count++;
	printf("%d records have been put.\n", count);

	return 0;
}
