/*
 * MDCacheD client library for C
 * Implements interfaces defined in mc_client.h
 * (c) 2007.-2012 Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: client.c 473 2012-04-08 10:02:18Z ivoras $
 */

#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <poll.h>
#include <time.h>

#include "linux-shims.h"

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <paths.h>
#include <assert.h>
#include <syslog.h>

#include "mc_protocol.h"
#include "mc_client.h"
#include "common_utils.h"

unsigned short int mc_client_seq = 0;
#define seq mc_client_seq

#define STR_BUF_SIZE 128

/* SNDBUF_SIZE and RCVBUF_SIZE should be large enough to hold
 * almost all request and response sizes. Outgoing packets that
 * are larger than SNDBUF_SIZE will be sent in pieces by the
 * network thread instead of worker threads. */
#ifndef SNDBUF_SIZE
#define SNDBUF_SIZE 8192
#endif

#ifndef RCVBUF_SIZE
#define RCVBUF_SIZE 8192
#endif

#ifndef DEFAULT_ASYNC_RECV_SIZE
#define DEFAULT_ASYNC_RECV_SIZE 8192
#endif


#if 0
static int
debug_elog(char *fmt, ...)
{
	va_list ap;
	char *msg;
	int fd, ss;

	va_start(ap, fmt);
	ss = vasprintf(&msg, fmt, ap);
	va_end(ap);
	if (ss == -1)
		return -1;
	fd = open("/tmp/mdcached_client.log", O_CREAT | O_WRONLY | O_APPEND | O_EXLOCK, 0644);
	if (fd < 0) {
		syslog(LOG_ERR, "Cannot open log file to write %s (%d)", msg, errno);
		free(msg);
		return -1;
	}
	if (write(fd, msg, ss) != ss) {
		syslog(LOG_ERR, "Cannot write msg to log: %s [%d] (%d)", msg, ss, errno);
		free(msg);
		return -1;
	}
	write(fd, "\n", 1);
	close(fd);
	free(msg);
	return 0;
}
#else
#define debug_elog(...) 
#endif


char*
mc_result_to_string(int code)
{
	switch (code) {
	case MC_RESULT_OK:
		return "No error or unknown error";
	case MC_RESULT_STATUS:
		return "Server reported error";
	case MC_RESULT_NETWORK:
		return "Network error";
	case MC_RESULT_NOT_FOUND:
		return "Data not found";
	case MC_RESULT_UNEXPECTED:
		return "Server returned unexpected response";
	case MC_RESULT_PARAM:
		return "Invalid parameter passed to client API";
	case MC_RESULT_STATE:
		return "Invalid state for the function call";
	case MC_RESULT_PENDING:
		return "Data still pending";
	case MC_ENOMEM:
		return "Out of memory";
	default:
		if (code > 0)
			return strerror(errno);
		else
			return "Unknown error";
	}
}


struct mc_connection *
mc_connect_local(char *unix_socket, int do_handshake, char **error_msg)
{
	struct sockaddr_un sux;
	struct mc_connection *cn;
	int flags;

	memset(&sux, 0, sizeof sux);
	cn = calloc(1, sizeof *cn);
	sux.sun_family = AF_UNIX;
	strncpy(sux.sun_path, unix_socket, sizeof sux.sun_path);
	cn->sock = socket(AF_UNIX, SOCK_STREAM, 0);
	if (cn->sock < 0) {
		if (error_msg != NULL) {
			*error_msg = malloc(STR_BUF_SIZE);
			sprintf(*error_msg, "cannot create AF_UNIX socket: %s", strerror(errno));
		}
		free(cn);
		return NULL;
	}

	flags = SNDBUF_SIZE;
	setsockopt(cn->sock, SOL_SOCKET, SO_SNDBUF, &flags, sizeof flags);
	flags = RCVBUF_SIZE;
	setsockopt(cn->sock, SOL_SOCKET, SO_RCVBUF, &flags, sizeof flags);

	if (connect(cn->sock, (struct sockaddr*)&sux, sizeof sux) < 0) {
		if (error_msg != NULL) {
			*error_msg = malloc(STR_BUF_SIZE);
			sprintf(*error_msg, "connect() failed: %s (server not running?)", strerror(errno));
		}
		close(cn->sock);
		free(cn);
		return NULL;
	}

	cn->ops_count = 0;
	cn->type = MC_SYNC;
	//printf("connection ok.\n");

	if (do_handshake) {
		//printf("handshaking...\n");
		if (mc_perform_handshake(cn, error_msg) != MC_RESULT_OK) {
			close(cn->sock);
			free(cn);
			if (error_msg != NULL && *error_msg == NULL) {
				*error_msg = malloc(STR_BUF_SIZE);
				sprintf(*error_msg, "Handshake error");
			}
			return NULL;
		}
		//printf("handshake ok.\n");
	}

	return cn;
}


/* TCPv4 */
struct mc_connection *
mc_connect_tcp(char *host_name, int port, int do_handshake, char **error_msg)
{
	struct sockaddr_in sin;
	struct hostent *he;
	struct mc_connection *cn;
	int flags;

	he = gethostbyname(host_name);
	if (he == NULL) {
		if (error_msg != NULL) {
			*error_msg = malloc(STR_BUF_SIZE);
			sprintf(*error_msg, "error resolving host %s: %s", host_name, strerror(errno));
		}
		return NULL;
	}

	memset (&sin, 0, sizeof sin);
	cn = calloc(1, sizeof *cn);
	sin.sin_family = AF_INET;

	cn->sock = socket(AF_INET, SOCK_STREAM, 0);
	if (cn->sock < 0) {
		if (error_msg != NULL) {
			*error_msg = malloc(STR_BUF_SIZE);
			sprintf(*error_msg, "cannot create AF_INET socket: %s", strerror(errno));
		}
		free(cn);
		return NULL;
	}

	sin.sin_family = he->h_addrtype;
	memcpy(&sin.sin_addr.s_addr, he->h_addr_list[0], he->h_length);
	sin.sin_port = htons(port);

	if (connect(cn->sock, (struct sockaddr*)&sin, sizeof sin) < 0) {
		if (error_msg != NULL) {
			*error_msg = malloc(STR_BUF_SIZE);
			sprintf(*error_msg, "error connecting to %s:%d: %s", host_name, port, strerror(errno));
		}
		free(cn);
		return NULL;
	}
	cn->ops_count = 0;
	cn->type = MC_SYNC;

	flags = 1;
	if (setsockopt(cn->sock, IPPROTO_TCP, TCP_NODELAY, (char*)&flags, sizeof(flags)) != 0)
		return NULL;
	flags = SNDBUF_SIZE;
	setsockopt(cn->sock, SOL_SOCKET, SO_SNDBUF, &flags, sizeof flags);
	flags = RCVBUF_SIZE;
	setsockopt(cn->sock, SOL_SOCKET, SO_RCVBUF, &flags, sizeof flags);

	if (do_handshake)
		if (mc_perform_handshake(cn, error_msg) != MC_RESULT_OK) {
			close(cn->sock);
			free(cn);
			return NULL;
		}

	return cn;
}


/**
 * Makes the specified connection async; from now on, only mc_async_* functions
 * may be used on the connections.
 */
int
mc_set_connection_async(struct mc_connection *cn)
{
	fcntl(cn->sock, F_SETFL, O_NONBLOCK);
	cn->type = MC_ASYNC;
	return MC_RESULT_OK;
}


/**
 * Performs a simple handshake operation.
 * Requires sync connection.
 */
int
mc_perform_handshake(struct mc_connection *cn, char **error_msg)
{
	struct mc_handshake hspacket;
	struct mc_resp_status status_packet;
	int r;

	cn->ops_count++;
	mc_init_packet(&hspacket, sizeof hspacket, seq++, MCMD_HANDSHAKE);
	hspacket.endian_signature = MC_ENDIAN_SIGNATURE;
	hspacket.proto_ver_major = MC_VER_MAJOR;
	hspacket.proto_ver_minor = MC_VER_MINOR;
	if (send(cn->sock, &hspacket, sizeof hspacket, 0) != sizeof hspacket) {
		if (error_msg != NULL) {
			*error_msg = malloc(STR_BUF_SIZE);
			sprintf(*error_msg, "send() failed on handshake packet: %s", strerror(errno));
		}
		return MC_RESULT_NETWORK;
	}
	memset(&status_packet, 0, sizeof status_packet);
	errno = 0;

	if ((r = recv(cn->sock, &status_packet, sizeof status_packet, MSG_WAITALL)) != sizeof status_packet) {
		if (error_msg != NULL) {
			*error_msg = malloc(STR_BUF_SIZE);
			if (errno == 0 && r == 0)
				sprintf(*error_msg, "recv() failed, possibly server disconnect");
			else
				sprintf(*error_msg, "recv() failed on handshake status packet: %s (recv returned %d)", strerror(errno), r);
		}
		return MC_RESULT_NETWORK;
	}

	if (status_packet.status != MCSTATUS_OK) {
		if (error_msg != NULL) {
			*error_msg = malloc(STR_BUF_SIZE);
			sprintf(*error_msg, "ERROR in handshake status: %d,%d", status_packet.status, status_packet.data);
		}
		return MC_RESULT_STATUS;
	}
	return MC_RESULT_OK;
}


/**
 * Closes the connection and frees the resources.
 */
void
mc_close_connection(struct mc_connection *mc)
{
	close(mc->sock);
	mc->sock = -1;
	if (mc->async_buf != NULL)
		free(mc->async_buf);
	free(mc);
}


#ifndef NDEBUG
/*
static void
_mc_dump_header(void *p)
{
	struct mc_header *pkt = (struct mc_header*)p;

	printf("mcmd=%u, flags=%u, seq=%u, total_size=%u\n",
		pkt->mcmd, pkt->flags, pkt->seq, pkt->total_size);
}
*/

/*
static void 
_dump_packet(struct mc_header *inpkt) {
	_dump_data((char*)inpkt, inpkt->total_size);
}
*/
#endif

/**
 * Does a simple conversation with the server: sends the outpkt and
 * receives a status response. Fails if the return packet isn't a
 * status response.
 */
static int
mc_converse_simple(struct mc_connection *cn,
	struct mc_header *outpkt, struct mc_resp_status *inpkt, char **error_msg)
{
	int r;

	assert(cn != NULL);
	assert(outpkt != NULL);
	assert(inpkt != NULL);
	memset(inpkt, 0, sizeof *inpkt);

	debug_elog("%s: send(%d)", __func__, outpkt->total_size);
	if (send(cn->sock, outpkt, outpkt->total_size, 0) != outpkt->total_size) {
		debug_elog("%s: send(%d) failed", __func__, outpkt->total_size);
		if (error_msg != NULL) {
			*error_msg = malloc(STR_BUF_SIZE);
			sprintf(*error_msg, "send() failed on handshake packet: %s", strerror(errno));
		}
		return MC_RESULT_NETWORK;
	}

	debug_elog("%s: recv(%d)", __func__, sizeof *inpkt);
	r = recv(cn->sock, inpkt, sizeof *inpkt, MSG_WAITALL);
	if (r != sizeof *inpkt) {
		debug_elog("%s: recv(%d) failed with %d", __func__, sizeof *inpkt, r);
		if (error_msg != NULL) {
			*error_msg = malloc(STR_BUF_SIZE);
			if (r == 0)
				sprintf(*error_msg, "recv() failed on: %s (probably server disconnect)", strerror(errno));
			else
				sprintf(*error_msg, "recv() failed on: %s (recv returned %d)", strerror(errno), r);
		}
		return MC_RESULT_NETWORK;
	}
	debug_elog("%s: recv(%d) ok", __func__, sizeof *inpkt);

	if (inpkt->header.mcmd != MCRESP_STATUS) {
		/* Something went wrong. See if there's garbage data on the
		 * line to be collected. */
		int garbage_size;
		if (error_msg != NULL) {
			*error_msg = malloc(STR_BUF_SIZE);
			sprintf(*error_msg, "received packet is not MCRESP_STATUS but %d", inpkt->header.mcmd);
/*
			warn("%s\n", *error_msg);
			abort();
*/
		}
		garbage_size = inpkt->header.total_size - sizeof *inpkt;
		if (garbage_size > 0) {
			char *buf;
			buf = malloc(garbage_size);
			recv(cn->sock, buf, garbage_size, MSG_WAITALL);
			free(buf);
		}
		return MC_RESULT_STATUS;
	}

	return MC_RESULT_OK;
}


/**
 * A simple wrapper for recv() to receive entire expected messages
 */
static int
recv_wrapper(int sock, char *buf, int size)
{
	int r, rsize = 0;

	while (size != 0) {
		r = recv(sock, buf, size, 0);
		if (r < 0)
			return r;
		size -= r;
		buf += r;
		rsize += r;
	}

	return rsize;
}


/**
 * A generic conversation, in which the other side is sent a packet (outpkt)
 * and the received packet is passed to the caller (inpkt). The caller must
 * free *inpkt.
 * TODO: what about async?
 */
static int
mc_converse_complex(struct mc_connection *cn,
	struct mc_header *outpkt, struct mc_header **inpkt, char **error_msg)
{
	int r;
	struct mc_header hdr;
	char *buf;

	assert(cn != NULL);
	assert(outpkt != NULL);
	assert(inpkt != NULL);

	debug_elog("%s: send(%d)", __func__, outpkt->total_size);
	if (send(cn->sock, outpkt, outpkt->total_size, 0) != outpkt->total_size) {
		debug_elog("%s: send(%d) failed", __func__, outpkt->total_size);
		if (error_msg != NULL) {
			*error_msg = malloc(STR_BUF_SIZE);
			sprintf(*error_msg, "send() failed on handshake packet: %s", strerror(errno));
		}
		return MC_RESULT_NETWORK;
	}

	/* Get the header first, it contains total_size */
	debug_elog("%s: recv(%d) [hdr]", __func__, sizeof hdr);
	if ((r = recv(cn->sock, &hdr, sizeof hdr, MSG_WAITALL)) != sizeof hdr) {
		debug_elog("%s: recv(%d) [hdr] failed", __func__, sizeof hdr);
		if (error_msg != NULL) {
			*error_msg = malloc(STR_BUF_SIZE);
			if (r == 0)
				sprintf(*error_msg, "recv() failed on header receive: %s (probably server disconnect)", strerror(errno));
			else
				sprintf(*error_msg, "recv() failed on header receive: %s (recv returned %d)", strerror(errno), r);
		}
		return MC_RESULT_NETWORK;
	}
	*inpkt = malloc(hdr.total_size);
	memcpy(*inpkt, &hdr, sizeof hdr);
	buf = (char*)(*inpkt);

	/* Get the rest of the data */
	debug_elog("%s: recv(%d) [body]", __func__, hdr.total_size - sizeof hdr);
	if ((r = recv_wrapper(cn->sock, buf + sizeof hdr, hdr.total_size - sizeof hdr)) != hdr.total_size - sizeof hdr) {
		debug_elog("%s: recv(%d) [body] failed: %d", __func__, hdr.total_size - sizeof hdr, r);
		if (error_msg != NULL) {
			*error_msg = malloc(STR_BUF_SIZE);
			if (r == 0)
				sprintf(*error_msg, "recv() failed on data body receive: %s (probably server disconnect)", strerror(errno));
			else
				sprintf(*error_msg, "recv() failed on data body receive: %s (recv returned %d)", strerror(errno), r);
		}
		return MC_RESULT_NETWORK;
	}
	return MC_RESULT_OK;
}


/**
 * The simplest API to put data into the server
 */
int
mc_put_simple(struct mc_connection *cn,
	char *key, unsigned int key_len,
	char *data, unsigned int data_len,
	time_t exptime, char **error_msg)
{
	struct mc_add_data *outpkt;
	struct mc_resp_status status_packet;
	size_t pktsize;
	int result;
	char *buf;

	if (key == NULL || data == NULL || key_len == 0)
		return MC_RESULT_PARAM;
	cn->ops_count++;

	pktsize = sizeof(*outpkt) + key_len + data_len;
	outpkt = malloc(pktsize);
	if (outpkt == NULL)
		return MC_ENOMEM;
	mc_init_packet(outpkt, pktsize, seq++, MCMD_PUT_DATA);
	outpkt->name_size = key_len;
	outpkt->data_size = data_len;
	outpkt->n_tags = 0;
	outpkt->exptime = exptime;
	buf = (char*)(&outpkt->data);
	memcpy(buf, key, key_len);
	memcpy(buf + key_len, data, data_len);

	result = mc_converse_simple(cn, (struct mc_header*)outpkt, &status_packet, error_msg);

	if (result == MC_RESULT_OK && (status_packet.status != MCSTATUS_OK)) {
		if (error_msg != NULL) {
			*error_msg = malloc(STR_BUF_SIZE);
			sprintf(*error_msg, "MCMD_PUT_DATA failed: %d,%d", status_packet.status, status_packet.data);
		}
		result = MC_RESULT_STATUS;
	}

	free(outpkt);
	return result;
}


/**
 * Put a single entry into the cache, a key-value record with tags.
 */
int
mc_put_simple_tags(struct mc_connection *cn,
	char *key, unsigned int key_len,
	char *data, unsigned int data_len,
	struct mc_tag *tags, unsigned int n_tags,
	time_t exptime, char **error_msg)
{
	struct mc_add_data *outpkt;
	struct mc_resp_status status_packet;
	unsigned pktsize;
	int result;
	char *buf;

	if (key == NULL || data == NULL || key_len == 0)
		return MC_INVALID_PARAM;
	if (n_tags != 0 && tags == NULL)
		return MC_INVALID_PARAM;

	cn->ops_count++;

	pktsize = sizeof(*outpkt) + sizeof(struct mc_tag) * n_tags + key_len + data_len;
	outpkt = malloc(pktsize);
	if (outpkt == NULL)
		return MC_ENOMEM;
	mc_init_packet(outpkt, pktsize, seq++, MCMD_PUT_DATA);
	outpkt->name_size = key_len;
	outpkt->data_size = data_len;
	outpkt->n_tags = n_tags;
	outpkt->exptime = exptime;
	buf = outpkt->data;
	if (n_tags != 0) {
		memcpy(buf, tags, sizeof(struct mc_tag) * n_tags);
		buf += sizeof(struct mc_tag) * n_tags;
	}
	memcpy(buf, key, key_len);
	buf += key_len;
	memcpy(buf, data, data_len);

	result = mc_converse_simple(cn, (struct mc_header*)outpkt, &status_packet, error_msg);

	if (result == MC_RESULT_OK && (status_packet.status != MCSTATUS_OK)) {
		if (error_msg != NULL) {
			*error_msg = malloc(STR_BUF_SIZE);
			sprintf(*error_msg, "MCMD_PUT_DATA failed: %d,%d", status_packet.status, status_packet.data);
		}
		result = MC_RESULT_STATUS;
	}

	free(outpkt);
	return result;
}


/**
 * Offload a packet for async delivery. The function returns the number of bytes
 * sent by the system.
 */
static int
async_offload(struct mc_connection *cn, struct mc_header *outpkt)
{
	int s;
	void *voutpkt;

	s = send(cn->sock, outpkt, outpkt->total_size, 0);
	if (s <= 0 || s == outpkt->total_size)
		return s;

	assert(cn->async_buf == NULL && cn->async_buf_pos == 0 && cn->async_buf_size == 0);
	voutpkt = outpkt;
	cn->async_buf_size = outpkt->total_size - s;
	cn->async_buf = malloc(cn->async_buf_size);
	memcpy(cn->async_buf, voutpkt + s, cn->async_buf_size);

	return s;
}


/**
 * Puts a key/value pair to the server, but doesn't wait for the response. Can only
 * be used with async clients.
 */
int
mc_put_simple_async(struct mc_connection *cn, char *key, unsigned int key_len, char *data, unsigned int data_len, time_t exptime, char **error_msg)
{
	struct mc_add_data *outpkt;
	int result;
	char *buf;

	if (key == NULL || data == NULL || key_len == 0)
		return MC_RESULT_PARAM;

	if (cn->async_state != MC_ASYNC_IDLE)
		return MC_RESULT_STATE;

	cn->ops_count++;

	outpkt = malloc(sizeof(*outpkt) + key_len + data_len);
	if (outpkt == NULL)
		return MC_ENOMEM;

	mc_init_packet(outpkt, sizeof(*outpkt) + key_len + data_len, seq++, MCMD_PUT_DATA);
	outpkt->name_size = key_len;
	outpkt->data_size = data_len;
	outpkt->n_tags = 0;
	outpkt->exptime = exptime;
	buf = (char*)(&outpkt->data);
	memcpy(buf, key, key_len);
	memcpy(buf + key_len, data, data_len);

	result = async_offload(cn, (struct mc_header*)outpkt);

	if (result == outpkt->header.total_size) {
		cn->async_state = MC_ASYNC_WANT_REPLY;
		free(outpkt);
		return MC_RESULT_OK;
	} else if (result < 0) {
		free(outpkt);
		return MC_RESULT_NETWORK;
	} else {
		cn->async_state = MC_ASYNC_WANT_REPLY;
		free(outpkt);
		return MC_RESULT_PENDING;
	}
}


/**
 * Gets simple key/value pair from the server.
 */
int
mc_get_simple(struct mc_connection *cn, char *key, unsigned int key_len, char **data, unsigned int *data_len, char **error_msg)
{
	struct mc_get_data_name *outpkt;
	struct mc_data_entry *inpkt;
	struct mc_header *inpkt_p = NULL;
	int result;

	cn->ops_count++;
	outpkt = malloc(sizeof(*outpkt) + key_len);
	if (outpkt == NULL)
		return MC_ENOMEM;
	mc_init_packet(outpkt, sizeof(*outpkt) + key_len, seq++, MCMD_GET_DATA_NAME);
	outpkt->name_size = key_len;
	memcpy(&outpkt->data, key, key_len);

	result = mc_converse_complex(cn, (struct mc_header*)outpkt, &inpkt_p, error_msg);
	free(outpkt);
	inpkt = (struct mc_resp_data*)inpkt_p;

	if (result == MC_RESULT_OK && inpkt->header.mcmd == MCRESP_DATA) {
		char *buf;
		buf = inpkt->data;
		/* Skip tags and name */
		buf += inpkt->name_size + inpkt->n_tags * sizeof(struct mc_tag);
		*data_len = inpkt->data_size;
		*data = malloc(*data_len);
		memcpy(*data, buf, inpkt->data_size);
	} else if (result == MC_RESULT_OK && inpkt->header.mcmd == MCRESP_STATUS) {
		struct mc_resp_status *status = (struct mc_resp_status*)inpkt;
		if (status->status == MCSTATUS_ERROR && status->data == MCREASON_NOTFOUND) {
			/* It's so common we won't give an error message to it, just set it to NULL */
			free(inpkt);
			if (error_msg != NULL)
				*error_msg = NULL;
			return MC_RESULT_NOT_FOUND;
		} else {
			if (error_msg != NULL && *error_msg == NULL)
				if (asprintf(error_msg, "Error in response: %d,%d", status->status, status->data) == -1)
					*error_msg = NULL;
			free(inpkt);
			return MC_RESULT_STATUS;
		}
	} else {
		if (error_msg != NULL && *error_msg == NULL)
			if (asprintf(error_msg, "Invalid response mcmd: %d (result=%d)", inpkt->header.mcmd, result) == -1)
				*error_msg = NULL;
		free(inpkt);
		return MC_RESULT_UNEXPECTED;
	}

	free(inpkt);
	return MC_RESULT_OK;
}


/**
 * Gets simple key/value pair from the server.
 */
int
mc_get_simple_tags(struct mc_connection *cn, char *key, unsigned int key_len, char **data, unsigned int *data_len, struct mc_tag **tags, unsigned int *n_tags, char **error_msg)
{
	struct mc_get_data_name *outpkt;
	struct mc_data_entry *inpkt;
	struct mc_header *inpkt_p = NULL;
	int result;

	cn->ops_count++;
	outpkt = malloc(sizeof(*outpkt) + key_len);
	if (outpkt == NULL)
		return MC_ENOMEM;
	mc_init_packet(outpkt, sizeof(*outpkt) + key_len, seq++, MCMD_GET_DATA_NAME);
	outpkt->name_size = key_len;
	memcpy(&outpkt->data, key, key_len);

	result = mc_converse_complex(cn, (struct mc_header*)outpkt, &inpkt_p, error_msg);
	free(outpkt);
	inpkt = (struct mc_resp_data*)inpkt_p;

	if (result == MC_RESULT_OK && inpkt->header.mcmd == MCRESP_DATA) {
		char *buf;
		buf = inpkt->data;
		if (inpkt->n_tags > 0) {
			*tags = malloc(inpkt->n_tags * sizeof(struct mc_tag));
			*n_tags = inpkt->n_tags;
			memcpy(*tags, buf, inpkt->n_tags * sizeof(struct mc_tag));
		} else {
			*tags = NULL;
			*n_tags = 0;
		}
		buf += inpkt->n_tags * sizeof(struct mc_tag);
		buf += inpkt->name_size; /* Skip name */
		*data_len = inpkt->data_size;
		*data = malloc(inpkt->data_size);
		memcpy(*data, buf, inpkt->data_size);
	} else if (result == MC_RESULT_OK && inpkt->header.mcmd == MCRESP_STATUS) {
		struct mc_resp_status *status = (struct mc_resp_status*)inpkt;
		if (status->status == MCSTATUS_ERROR && status->data == MCREASON_NOTFOUND) {
			/* It's so common we won't give an error message to it, just set it to NULL */
			free(inpkt);
			if (error_msg != NULL)
				*error_msg = NULL;
			return MC_RESULT_NOT_FOUND;
		} else {
			if (error_msg != NULL && *error_msg == NULL)
				if (asprintf(error_msg, "Error in response: %d,%d", status->status, status->data) == -1)
					*error_msg = NULL;
			free(inpkt);
			return MC_RESULT_STATUS;
		}
	} else {
		if (error_msg != NULL && *error_msg == NULL)
			if (asprintf(error_msg, "Invalid response mcmd: %d (result=%d)", inpkt->header.mcmd, result) == -1)
				*error_msg = NULL;
		free(inpkt);
		return MC_RESULT_UNEXPECTED;
	}

	free(inpkt);
	return MC_RESULT_OK;
}


/**
 * Initiates an asynchronous operation to retrieve a record.
 * The caller needs to call mc_async_recvpoll() to poll for the async operation
 * result.
 * Can only be used with async clients.
 */
int
mc_get_simple_async(struct mc_connection *cn, char *key, unsigned int key_len, char **error_msg)
{
	struct mc_get_data_name *outpkt;
	int result;

	cn->ops_count++;
	outpkt = malloc(sizeof(*outpkt) + key_len);
	if (outpkt == NULL)
		return MC_ENOMEM;
	mc_init_packet(outpkt, sizeof(*outpkt) + key_len, seq++, MCMD_GET_DATA_NAME);
	outpkt->name_size = key_len;
	memcpy(&outpkt->data, key, key_len);

	result = async_offload(cn, (struct mc_header*)outpkt);
	free(outpkt);

	if (result == outpkt->header.total_size) {
		cn->async_state = MC_ASYNC_WANT_REPLY;
		return MC_RESULT_OK;
	} else if (result < 0)
		return MC_RESULT_NETWORK;
	else {
		cn->async_state = MC_ASYNC_WANT_REPLY;
		return MC_RESULT_PENDING;
	}
}

/**
 * Delete a record by its name.
 */
int
mc_del_simple(struct mc_connection *cn, char *key, unsigned int key_len, char **error_msg)
{
	struct mc_del_data_name *outpkt;
	struct mc_resp_status inpkt;
	int res;

	outpkt = malloc(sizeof(*outpkt) + key_len);
	if (outpkt == NULL)
		return MC_ENOMEM;
	mc_init_packet(outpkt, sizeof(*outpkt) + key_len, seq++, MCMD_DEL_DATA_NAME);
	outpkt->name_size = key_len;
	memcpy(outpkt->data, key, key_len);
	res = mc_converse_simple(cn, (struct mc_header*)outpkt, &inpkt, error_msg);
	free(outpkt);
	return res;
}


/**
 * Parse a struct mc_resp_multidata (the protocol structure) into struct mc_multidata_result
 * (the client API structure).
 */
static int
_mc_parse_multidata(struct mc_resp_multidata *inpkt, struct mc_multidata_result **mr)
{
	char *data = inpkt->data;
	unsigned i;

	*mr = malloc(sizeof(struct mc_multidata_result));
	if (*mr == NULL)
		return MC_ENOMEM;
	(*mr)->n_records = inpkt->n_records;
	(*mr)->pkt = inpkt;
	(*mr)->records = malloc(inpkt->n_records * sizeof((*mr)->records));
	/* Parse the multidata blob */
	for (i = 0; i < inpkt->n_records; i++) {
		(*mr)->records[i] = (struct mc_data_entry*)data;
		data += (*mr)->records[i]->header.total_size;
	}
	return MC_RESULT_OK;
}


/**
 * Gets a set of records which have the given key and any of the given values.
 * The result is returned as the struct mc_multidata_result, which is allocated
 * by the function and must be freed by mc_multidata_result_free().
 */
int
mc_get_by_tag_values(struct mc_connection *cn, tag_key_t key,
	tag_value_t *values, unsigned n_values,
	struct mc_multidata_result **mr, char **error_msg)
{
	struct mc_get_data_by_tag_values *outpkt;
	struct mc_resp_multidata *inpkt;
	struct mc_header *inpkt_p = NULL;
	int result;

	cn->ops_count++;
	outpkt = malloc(sizeof(*outpkt) + n_values * sizeof(*values));
	if (outpkt == NULL)
		return MC_ENOMEM;
	mc_init_packet(outpkt, sizeof(*outpkt) + n_values * sizeof(*values), seq++, MCMD_GET_DATA_BY_TAG_VALUES);
	outpkt->n_values = n_values;
	outpkt->tag_key = key;
	memcpy(outpkt->data, values, n_values * sizeof(*values));

	result = mc_converse_complex(cn, (struct mc_header*)outpkt, &inpkt_p, error_msg);
	free(outpkt);
	inpkt = (struct mc_resp_multidata*)inpkt_p;

	if (result == MC_RESULT_OK && inpkt->header.mcmd == MCRESP_MULTIDATA) {
		int res = _mc_parse_multidata(inpkt, mr);
		if (res != MC_RESULT_OK) {
			free(inpkt);
			return res;
		}
	} else if (result == MC_RESULT_OK && inpkt->header.mcmd == MCRESP_STATUS) {
		struct mc_resp_status *status = (struct mc_resp_status*)inpkt;
		if (status->status == MCSTATUS_ERROR && status->data == MCREASON_NOTFOUND) {
			/* It's so common we won't give an error message to it, just set it to NULL */
			if (error_msg != NULL)
				*error_msg = NULL;
			free(inpkt);
			return MC_RESULT_NOT_FOUND;
		} else {
			if (error_msg != NULL && *error_msg == NULL)
				if (asprintf(error_msg, "Error in response: %d,%d", status->status, status->data) == -1)
					*error_msg = NULL;
			free(inpkt);
			return MC_RESULT_STATUS;
		}
	} else {
		if (error_msg != NULL && *error_msg == NULL)
			if (asprintf(error_msg, "Invalid response mcmd: %d (result=%d)", inpkt->header.mcmd, result) == -1)
				*error_msg = NULL;
		free(inpkt);
		return MC_RESULT_UNEXPECTED;
	}

	/* inpkt is not freed; caller must use mc_multidata_result_free() */
	return MC_RESULT_OK;
}


/**
 * Gets a list of all keys (record names) stored in the cache. This is not a very
 * efficient operation.
 */
int
mc_get_all_keys(struct mc_connection *cn, unsigned flags, struct mc_multistring_result **mr, char **error_msg)
{
	struct mc_simple *outpkt;
	struct mc_resp_multistring *inpkt;
	struct mc_header *inpkt_p = NULL;
	int result;
	unsigned i;

	cn->ops_count++;
	outpkt = malloc(sizeof(*outpkt));
	if (outpkt == NULL)
		return MC_ENOMEM;
	mc_init_packet(outpkt, sizeof(*outpkt), seq++, MCMD_GET_ALL_KEYS);
	outpkt->flags = flags;

	result = mc_converse_complex(cn, (struct mc_header*)outpkt, &inpkt_p, error_msg);
	free(outpkt);
	inpkt = (struct mc_resp_multistring*)inpkt_p;

	if (result == MC_RESULT_OK && inpkt->header.mcmd == MCRESP_MULTISTRING) {
		char *data = inpkt->data;
		unsigned skip, total_skip = 0;
		
		*mr = malloc(sizeof(struct mc_multistring_result));
		if (*mr == NULL) {
			free(inpkt);
			return MC_ENOMEM;
		}
		(*mr)->n_records = inpkt->n_records;
		(*mr)->pkt = inpkt;
		(*mr)->records = malloc(inpkt->n_records * sizeof(struct mc_multistring_1string*));
		/* Parse the multistring blob */
		for (i = 0; i < inpkt->n_records; i++) {
			(*mr)->records[i] = (struct mc_multistring_1string*)data;
			skip = sizeof((*mr)->records[i]->str_len) + (*mr)->records[i]->str_len;
			total_skip += skip;
			data += skip;
		}
	} else if (result == MC_RESULT_OK && inpkt->header.mcmd == MCRESP_STATUS) {
		struct mc_resp_status *status = (struct mc_resp_status*)inpkt;
		if (status->status == MCSTATUS_ERROR && status->data == MCREASON_NOTFOUND) {
			/* It's so common we won't give an error message to it, just set it to NULL */
			if (error_msg != NULL)
				*error_msg = NULL;
			free(inpkt);
			return MC_RESULT_NOT_FOUND;
		} else {
			if (error_msg != NULL && *error_msg == NULL)
				if (asprintf(error_msg, "Error in response: %d,%d", status->status, status->data) == -1)
					*error_msg = NULL;
			free(inpkt);
			return MC_RESULT_STATUS;
		}
	} else {
		if (error_msg != NULL && *error_msg == NULL)
			if (asprintf(error_msg, "Invalid response mcmd: %d (result=%d)", inpkt->header.mcmd, result) == -1)
				*error_msg = NULL;
		free(inpkt);
		return MC_RESULT_UNEXPECTED;
	}

	/* inpkt is not freed; caller must use mc_multistring_result_free() */
	return MC_RESULT_OK;
}


/**
 * Frees the mc_multidata_result which was allocated & returned by one of the
 * multidata operations. Due to the internal structure of mc_multidata_result,
 * it should not be modified (or freed) by the client.
 */
void
mc_multidata_result_free(struct mc_multidata_result *mr)
{
	free(mr->pkt);
	free(mr);
}


/**
 * Frees a mc_multistring_result structure and its data.
 */
void
mc_multistring_result_free(struct mc_multistring_result *msr)
{
	free(msr->pkt);
	free(msr);
}


/**
 * Deletes a set of records which have the given tag key and any of the given tag values.
 */
int
mc_del_by_tag_values(struct mc_connection *cn, tag_key_t key, tag_value_t *tag_values, unsigned n_values, unsigned *n_records, char **err_msg)
{
	struct mc_del_data_by_tag_values *outpkt;
	struct mc_resp_status inpkt;
	int res;

	cn->ops_count++;

	outpkt = malloc(sizeof(*outpkt) + n_values * sizeof(*tag_values));
	if (outpkt == NULL)
		return MC_ENOMEM;
	mc_init_packet(outpkt, sizeof(*outpkt) + n_values * sizeof(*tag_values), seq++, MCMD_DEL_DATA_BY_TAG_VALUES);
	outpkt->n_values = n_values;
	outpkt->tag_key = key;
	memcpy(outpkt->data, tag_values, n_values * sizeof(*tag_values));
	res = mc_converse_simple(cn, (struct mc_header*)outpkt, &inpkt, err_msg);
	free(outpkt);
	if (n_records != NULL)
		*n_records = inpkt.data;
	return res;
}


/**
 * Performs an atomic operation on a data record; one of CMPSET, ADD, FETCHADD,
 * READANDCLEAR.
 */
int
mc_atomic_op(struct mc_connection *cn, int op, char *key, unsigned int key_len,
	int64_t arg1, int64_t arg2, int64_t *result, char **error_msg)
{
	struct mc_atomic_op *outpkt;
	struct mc_header *inpkt;
	struct mc_resp_atomic *mr;
	int res;

	switch (op) {
	case MCMD_ATOMIC_CMPSET:
	case MCMD_ATOMIC_ADD:
	case MCMD_ATOMIC_FETCHADD:
	case MCMD_ATOMIC_READANDCLEAR:
		break;
	default:
		return MC_RESULT_PARAM;
	}
	cn->ops_count++;

	outpkt = malloc(sizeof(*outpkt) + key_len);
	if (outpkt == NULL)
		return MC_ENOMEM;
	mc_init_packet(outpkt, sizeof(*outpkt) + key_len, seq++, op);
	outpkt->arg1 = arg1;
	outpkt->arg2 = arg2;
	outpkt->name_size = key_len;
	memcpy(outpkt->name, key, key_len);

	res = mc_converse_complex(cn, (struct mc_header*)outpkt, &inpkt, error_msg);
	free(outpkt);
	mr = (struct mc_resp_atomic*)inpkt;
	
	if (res != MC_RESULT_OK) {
		free(mr);
		return res;
	} else if (mr->header.mcmd != MCRESP_ATOMIC) {
		free(mr);
		return MC_RESULT_UNEXPECTED;
	}
	*result = mr->result;
	free(mr);
	return res;
}


int
mc_atomic_cmpset(struct mc_connection *cn, char *key, unsigned key_len, int64_t arg1, int64_t arg2, int64_t *res, char **error_msg)
{
	return mc_atomic_op(cn, MCMD_ATOMIC_CMPSET, key, key_len, arg1, arg2, res, error_msg);
}


int
mc_atomic_add(struct mc_connection *cn, char *key, unsigned key_len, int64_t arg1, int64_t *res, char **error_msg)
{
	return mc_atomic_op(cn, MCMD_ATOMIC_ADD, key, key_len, arg1, 0, res, error_msg);
}


int
mc_atomic_fetchadd(struct mc_connection *cn, char *key, unsigned key_len, int64_t arg1, int64_t *res, char **error_msg)
{
	return mc_atomic_op(cn, MCMD_ATOMIC_FETCHADD, key, key_len, arg1, 0, res, error_msg);
}


int
mc_atomic_readandclear(struct mc_connection *cn, char *key, unsigned key_len, int64_t *res, char **error_msg) 
{
	return mc_atomic_op(cn, MCMD_ATOMIC_READANDCLEAR, key, key_len, 0, 0, res, error_msg);
}


int
mc_atomic_put(struct mc_connection *cn, char *key, unsigned key_len, int64_t value, struct mc_tag *tags, unsigned n_tags, time_t exptime, char **error_msg)
{
	return mc_put_simple_tags(cn, key, key_len, (char*)&value, sizeof(value), tags, n_tags, exptime, error_msg);
}


int
mc_atomic_get(struct mc_connection *cn, char *key, unsigned key_len, int64_t *res, char **error_msg)
{
	char *data;
	unsigned data_len;
	int r;
	
	r = mc_get_simple(cn, key, key_len, &data, &data_len, error_msg);
	if (r == MC_RESULT_OK && data_len == sizeof(*res)) {
		memcpy(res, data, sizeof(*res));
		free(data);
		return MC_RESULT_OK;
	} else {
		*res = -1;
		return r;
	}
}


/**
 * Puts multiple records into the cache.
 */
int
mc_mput(struct mc_connection *cn, uint16_t n_records, struct mc_data_entry **records, unsigned flags, char **err_msg)
{
	struct mc_mput *outpkt;
	struct mc_resp_status inpkt;
	unsigned records_size, i;
	char *data;
	int res;

	for (i = 0, records_size = 0; i < n_records; i++)
		records_size += records[i]->header.total_size;
	outpkt = malloc(sizeof(*outpkt) + records_size);
	if (outpkt == NULL)
		return MC_ENOMEM;
	mc_init_packet(outpkt, sizeof(*outpkt) + records_size, seq++, MCMD_MPUT);
	outpkt->n_records = n_records;
	outpkt->flags = flags;
	data = outpkt->data;
	for (i = 0; i < n_records; i++) {
		memcpy(data, records[i], records[i]->header.total_size);
		data += records[i]->header.total_size;
	}
	res = mc_converse_simple(cn, (struct mc_header*)outpkt, &inpkt, err_msg);
	free(outpkt);
	return res;
}


/**
 * Gets a set of named records.
 */
int
mc_mget(struct mc_connection *cn, uint16_t n_records, uint16_t *key_lengths, char **keys, unsigned flags, struct mc_multidata_result **mr, char **error_msg)
{
	struct mc_mget *outpkt;
	struct mc_resp_multidata *inpkt;
	struct mc_header *inpkt_p;
	unsigned pkt_size, i;
	int result;
	char *d;

	pkt_size = sizeof(*outpkt);
	for (i = 0; i < n_records; i++)
		pkt_size += key_lengths[i] + sizeof(uint16_t);
	outpkt = malloc(pkt_size);
	if (outpkt == NULL)
		return MC_ENOMEM;
	mc_init_packet(outpkt, pkt_size, seq++, MCMD_MGET);
	outpkt->n_names = n_records;
	outpkt->flags = flags;
	d = outpkt->data;
	for (i = 0; i < n_records; i++) {
		memcpy(d, &key_lengths[i], sizeof(uint16_t));
		d += sizeof(uint16_t);
		memcpy(d, keys[i], key_lengths[i]);
		d += key_lengths[i];
	}

	result = mc_converse_complex(cn, (struct mc_header*)outpkt, &inpkt_p, error_msg);
	free(outpkt);
	inpkt = (struct mc_resp_multidata*)inpkt_p;
	if (result == MC_RESULT_OK && inpkt->header.mcmd == MCRESP_MULTIDATA) {
		int res = _mc_parse_multidata(inpkt, mr);
		if (res != 0) {
			free(inpkt);
			return res;
		}
	} else if (result == MC_RESULT_OK && inpkt->header.mcmd == MCRESP_STATUS) {
		struct mc_resp_status *status = (struct mc_resp_status*)inpkt;
		if (error_msg != NULL && *error_msg == NULL)
			if (asprintf(error_msg, "Error in response: %d,%d", status->status, status->data) == -1)
				*error_msg = NULL;
		free(inpkt);
		return MC_RESULT_STATUS;
	} else {
		if (error_msg != NULL && *error_msg == NULL)
			if (asprintf(error_msg, "Invalid response mcmd: %d (result=%d)", inpkt->header.mcmd, result) == -1)
				*error_msg = NULL;
		free(inpkt);
		return MC_RESULT_UNEXPECTED;
	}

	return MC_RESULT_OK;
}


/**
 * Deletes a set of named records.
 */
int
mc_mdel(struct mc_connection *cn, uint16_t n_records, uint16_t *key_lengths, char **keys, unsigned flags, unsigned *n_deleted, char **error_msg)
{
	struct mc_mdel *outpkt;
	struct mc_resp_status inpkt;
	unsigned pkt_size, i;
	int res;
	char *d;

	pkt_size = sizeof(*outpkt);
	for (i = 0; i < n_records; i++)
		pkt_size += key_lengths[i] + sizeof(uint16_t);
	outpkt = malloc(pkt_size);
	if (outpkt == NULL)
		return MC_ENOMEM;
	mc_init_packet(outpkt, pkt_size, seq++, MCMD_MDEL);
	outpkt->n_names = n_records;
	outpkt->flags = flags;
	d = outpkt->data;
	for (i = 0; i < n_records; i++) {
		memcpy(d, &key_lengths[i], sizeof(uint16_t));
		d += sizeof(uint16_t);
		memcpy(d, keys[i], key_lengths[i]);
		d += key_lengths[i];
	}

	res = mc_converse_simple(cn, (struct mc_header*)outpkt, &inpkt, error_msg);
	free(outpkt);
	if (n_deleted != NULL)
		*n_deleted = inpkt.data;
	return res;
}


/**
 * Puts a record to the server containing the specified data and tags but without
 * a key. The key is auto-generated on the server and returned back.
 */
int
mc_tstack_push(struct mc_connection *cn, struct mc_tag *tags, unsigned n_tags, char *data, unsigned data_size, char **new_key, unsigned *new_key_size, char **error_msg)
{
	struct mc_tstack_push *outpkt;
	struct mc_resp_tstack_push *inpkt;
	struct mc_header *inpkt_p;
	char *d;
	unsigned pkt_size;
	int result;
	
	pkt_size = sizeof(*outpkt) + n_tags * sizeof(*tags) + data_size;
	outpkt = malloc(pkt_size);
	if (outpkt == NULL)
		return MC_ENOMEM;
	mc_init_packet(outpkt, pkt_size, seq++, MCMD_TSTACK_PUSH);
	outpkt->flags = 0;
	outpkt->n_tags = n_tags;
	outpkt->data_size = data_size;
	d = outpkt->data;
	memcpy(d, tags, n_tags * sizeof(*tags));
	d += n_tags * sizeof(*tags);
	memcpy(d, data, data_size);
	
	result = mc_converse_complex(cn, (struct mc_header*)outpkt, &inpkt_p, error_msg);
	free(outpkt);
	inpkt = (struct mc_resp_tstack_push*)inpkt_p;
	
	if (result == MC_RESULT_OK && inpkt->header.mcmd == MCRESP_TSTACK_PUSH) {
		*new_key = malloc(inpkt->name_size);
		*new_key_size = inpkt->name_size;
		if (*new_key == NULL) {
			free(inpkt);
			return MC_ENOMEM;
		}
		memcpy(*new_key, inpkt->name, inpkt->name_size);
	} else if (result == MC_RESULT_OK && inpkt->header.mcmd == MCRESP_STATUS) {
		struct mc_resp_status *status = (struct mc_resp_status*)inpkt_p;
		if (error_msg != NULL && *error_msg == NULL)
			if (asprintf(error_msg, "Error in response: %d,%d", status->status, status->data) == -1)
				*error_msg = NULL;
		free(inpkt);
		*new_key = NULL;
		return MC_RESULT_STATUS;
	} else {
		if (error_msg != NULL && *error_msg == NULL)
			if (asprintf(error_msg, "Invalid response mcmd: %d (result=%d)", inpkt->header.mcmd, result) == -1)
				*error_msg = NULL;
		free(inpkt);
		*new_key = NULL;
		return MC_RESULT_UNEXPECTED;
	}
	free(inpkt);
	return MC_RESULT_OK;
}


/**
 * "Pops" a record off a virtual stack on the server identified by the tag (i.e.
 * its key and value).
 */ 
int
mc_tstack_pop(struct mc_connection *cn, tag_key_t tkey, tag_value_t tval, struct mc_data_entry **mce, char **error_msg)
{
	struct mc_tstack_pop *outpkt;
	struct mc_data_entry *inpkt;
	struct mc_header *inpkt_p;
	int result;
	
	outpkt = malloc(sizeof(*outpkt));
	if (outpkt == NULL)
		return MC_ENOMEM;
	mc_init_packet(outpkt, sizeof(*outpkt), seq++, MCMD_TSTACK_POP);
	outpkt->tag.key = tkey;
	outpkt->tag.value = tval;
	outpkt->flags = 0;
	
	result = mc_converse_complex(cn, (struct mc_header*)outpkt, &inpkt_p, error_msg);
	free(outpkt);
	inpkt = (struct mc_data_entry*)inpkt_p;
	
	if (result == MC_RESULT_OK && inpkt->header.mcmd == MCRESP_DATA) {
		*mce = inpkt;
	} else if (result == MC_RESULT_OK && inpkt->header.mcmd == MCRESP_STATUS) {
		struct mc_resp_status *status = (struct mc_resp_status*)inpkt_p;
		if (status->data == MCREASON_NOTFOUND) {
			*error_msg = NULL;
			free(inpkt);
			return MC_RESULT_NOT_FOUND;
		}
		if (error_msg != NULL && *error_msg == NULL)
			if (asprintf(error_msg, "Error in response: %d,%d", status->status, status->data) == -1)
				*error_msg = NULL;
		free(inpkt);
		return MC_RESULT_STATUS;
	} else {
		if (error_msg != NULL && *error_msg == NULL)
			if (asprintf(error_msg, "Invalid response mcmd: %d (result=%d)", inpkt->header.mcmd, result) == -1)
				*error_msg = NULL;
		free(inpkt);
		return MC_RESULT_UNEXPECTED;
	}
	return MC_RESULT_OK;
}


/**
 * (Re)sets record's tag to the ones given.
 */
int
mc_set_tags(struct mc_connection *cn, char *key, unsigned key_len, struct mc_tag *tags, unsigned n_tags, char **error_msg)
{
	struct mc_set_tags *outpkt;
	struct mc_resp_status inpkt;
	char *buf;
	int res;
	
	if (tags == NULL && n_tags != 0)
		return MC_INVALID_PARAM;
	
	outpkt = malloc(sizeof(*outpkt) + sizeof(struct mc_tag) * n_tags + key_len);
	if (outpkt == NULL)
		return MC_ENOMEM;
	mc_init_packet(outpkt, sizeof(*outpkt) + sizeof(struct mc_tag) * n_tags + key_len, seq++, MCMD_SET_TAGS);
	outpkt->n_tags = n_tags;
	outpkt->name_size = key_len;
	outpkt->flags = 0;
	outpkt->_unused = 0;
	
	buf = outpkt->data;
	memcpy(buf, tags, sizeof(struct mc_tag) * n_tags);
	buf += sizeof(struct mc_tag) * n_tags;
	memcpy(buf, key, key_len);
	
	res = mc_converse_simple(cn, (struct mc_header*)outpkt, &inpkt, error_msg);
	free(outpkt);
	return res;
}


/**
 * Poll the connection if async data needs to be sent.
 * Returns MC_RESULT_PENDING if the function needs to be called again to finish
 * sending, MC_RESULT_OK if all data was sent.
 * Can only be used with async clients.
 */
int
mc_async_sendpoll(struct mc_connection *cn)
{
	int s;
	if (cn->type != MC_ASYNC)
		return MC_RESULT_STATE;
	if (cn->async_buf == NULL || cn->async_buf_pos == cn->async_buf_size)
		return MC_RESULT_OK; /* nothing to send */
	s = send(cn->sock, cn->async_buf + cn->async_buf_pos, cn->async_buf_size - cn->async_buf_pos, 0);
	if (s < 0)
		return MC_RESULT_NETWORK;
	cn->async_buf_pos += s;
	if (cn->async_buf_pos == cn->async_buf_size) {
		cn->async_buf_pos = cn->async_buf_size = 0;
		free(cn->async_buf);
		cn->async_buf = NULL;
		cn->async_state = MC_ASYNC_IDLE;
		return MC_RESULT_OK;
	}
	return MC_RESULT_PENDING;
}


/**
 * Handle incoming data. If it's called when the internal async buffers are
 * empty, it will assume a new response is being
 * received and set up the state accordingly; otherwise, it will assume the
 * state has been set and will receive the data assuming it's a part of the
 * response. Erronously mixing mc_async_sendpoll() and mc_async_recvpoll() will
 * result in unpredictable (and always erroneous) behaviour, as they are separate
 * states, only one of which can be active at one time.
 * Can only be used with async clients.
 */
int
mc_async_recvpoll(struct mc_connection *cn, struct mc_header **outpkt)
{
	int r;
	struct mc_header *pkt;

	if (cn->type != MC_ASYNC || cn->async_state != MC_ASYNC_WANT_REPLY)
		return MC_RESULT_STATE;

	if (cn->async_buf == NULL) {
		/* Assume we're just initiating waiting for the response */
		cn->async_buf = malloc(DEFAULT_ASYNC_RECV_SIZE);
		if (cn->async_buf == NULL)
			return MC_ENOMEM;
		cn->async_buf_size = DEFAULT_ASYNC_RECV_SIZE;
		cn->async_buf_pos = 0;
	}
	r = recv(cn->sock, cn->async_buf, cn->async_buf_size, MSG_DONTWAIT);
	if (r < 0)
		return MC_RESULT_NETWORK;
	cn->async_buf_pos += r;
	pkt = cn->async_buf;
	if (pkt->total_size > cn->async_buf_size) {
		void *newbuf;
		newbuf = realloc(cn->async_buf, pkt->total_size);
		if (newbuf == NULL)
			return MC_ENOMEM;
		cn->async_buf = newbuf;
		cn->async_buf_size = pkt->total_size;
		return MC_RESULT_PENDING;
	}
	if (cn->async_buf_pos == pkt->total_size) {
		*outpkt = cn->async_buf;
		cn->async_buf_size = cn->async_buf_pos = 0;
		cn->async_buf = NULL;
		cn->async_state = MC_ASYNC_IDLE;
		return MC_RESULT_OK;
	} else
		return MC_RESULT_PENDING;
}
