/*
 * Circular queue implementation - used for queueing structs in ordered lists
 * without the need for excessive pointer-hopping and memory allocations that
 * are common with linked lists.
 * By Ivan Voras <ivoras@gmail.com> 2008.
 * This file is released under the 2-clause BSD license.
 * $Id: cqueue.h 268 2011-07-10 16:36:17Z ivoras $
 */


/*
 * Example usage:
 *
 * struct mine {
 * 	int my;
 *	int data;
 * };
 * CQUEUE_DECLARE(mine);
 * void main() {
 *	CQUEUE(mine) cq;
 *	struct mine *e;
 *	CQUEUE_CREATE(mine, &cq);
 *	e = CQUEUE_ADD(mine, &cq);
 *	if (e == NULL) errx(1, "ENOMEM");
 *	e->my = q->data = 42;
 *	e = CQUEUE_GET(mine, &cq);
 *	if (e == NULL) errx(1, "CQ empty");
 *	CQUEUE_DESTROY(&cq);
 * }
 *
 * Warning: queue order is not preserved if CQUEUE_ADD needs to allocate more
 * memory for the queue. Adjust CQUEUE_INITIAL_SIZE to avoid this.
 */

#ifndef CQUEUE_INITIAL_SIZE
#define CQUEUE_INITIAL_SIZE 10
#endif

#define CQUEUE_DECLARE(sname) 		\
struct cqueue_##sname { 		\
	unsigned int size;		\
	unsigned int first;		\
	unsigned int last;		\
	struct sname *items;		\
}

struct dummy_placeholder { int i; };
CQUEUE_DECLARE(dummy_placeholder);

#define CQUEUE(sname)	struct cqueue_##sname

#define CQUEUE_CREATE(sname, cq)		\
do {						\
	(cq)->items = malloc(CQUEUE_INITIAL_SIZE * sizeof(struct sname));	\
	(cq)->size = CQUEUE_INITIAL_SIZE;	\
	(cq)->first = 0;			\
	(cq)->last = 0;				\
} while (0)

#define CQUEUE_ADD(sname, cq)	_cqueue_add(cq, sizeof(struct sname))

static void *_cqueue_add(void *_cq, unsigned int item_size)
{
	struct cqueue_dummy_placeholder *cq = _cq;
	void *e;
	
	e = (void*)(cq->items) + (cq->last * item_size);
	cq->last = (cq->last + 1) % cq->size;
	/* INVARIANT: after ADD, cq->first is always != cq->last */
	if (cq->last == cq->first) {
		unsigned int newsize = cq->size * 2;
		void *newitems = realloc(cq->items, newsize * item_size);
		if (newitems == NULL)
			return NULL;
		cq->first = 0;
		cq->last = cq->size;
		cq->items = newitems;
		cq->size = newsize;
	}
	
	return e;
}

#define CQUEUE_GET(sname, cq)	_cqueue_get(cq, sizeof(struct sname))

static inline void *_cqueue_get(void *_cq, unsigned int item_size)
{
	struct cqueue_dummy_placeholder *cq = _cq;
	void *e;
	
	if (cq->first == cq->last)
		return NULL;
	
	e = (void*)(cq->items) + (cq->first * item_size);
	cq->first = (cq->first + 1) % cq->size;
	return e;
}

#define CQUEUE_PEEK(sname, cq) (cq)->first != (cq)->last ? (cq)->items[(cq)->first] : NULL

#define CQUEUE_DESTROY(cq) do { (cq)->size = 0; (cq)->first = 0; (cq)->last = 0; free((cq)->items); } while (0)

#define CQUEUE_EMPTY(cq) ( (cq)->first == (cq)->last )

