/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2011. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: logger.c 409 2012-01-20 15:01:41Z ivoras $
 */

#include <sys/time.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <syslog.h>
#include <pthread.h>
#include <errno.h>
#include <string.h>

#ifdef DMALLOC
#include <dmalloc.h>
#endif

#include "logger.h"

extern int log_level;
extern int debug_mode;
extern char **argv;

#define STR_BUF_SIZE 1024

/* Currently this log mutex isn't really needed, but it might be for
 * future logging mechanisms (e.g. shared memory logging, database
 * logging, etc) so put the infrastructure in place */
static pthread_mutex_t log_mtx;
static struct timeval start_time;


void
setup_logger()
{
	if (!debug_mode)
		openlog("mdcached", LOG_PID, LOG_DAEMON);
	pthread_mutex_init(&log_mtx, NULL);
	gettimeofday(&start_time, 0);
}


void
mc_log(int prio, const char *msg, ...)
{
	va_list ap;
	char *buf;
	int orig_errno = errno;

	if (prio > log_level)
		return;

	pthread_mutex_lock(&log_mtx);

	buf = (char*) malloc(STR_BUF_SIZE);
	va_start(ap, msg);
	vsnprintf(buf, STR_BUF_SIZE-1, msg, ap);
	va_end(ap);

	if (debug_mode) {
		if (prio != LOG_DEBUG && prio != LOG_CRIT)
			fprintf(stderr, "%s (%d): %s\n", argv[0], prio, buf);
		else {
			struct timeval tv;
			gettimeofday(&tv, NULL);
			fprintf(stderr, "%s (%d) @ %u.%u: %s\n", argv[0], prio,
				(unsigned int)(tv.tv_sec - start_time.tv_sec), (unsigned int)(tv.tv_usec), buf);
			if (prio == LOG_CRIT && orig_errno != 0)
				fprintf(stderr, "errno (possibly bogus) %d (%s)\n", orig_errno, strerror(orig_errno));
		}
	} else
		syslog(prio, "%s", buf);

	free(buf);
	pthread_mutex_unlock(&log_mtx);

	if (prio == LOG_CRIT)
		abort();
}


void
log_hexdump(const void *d, size_t len, unsigned width) {
	size_t c = 0;
	int llen = 0;

	if (width != 1 && width != 2 && width != 4 && width != 8) {
		mc_log(LOG_ERR, "log_hexdump given invalid width %u", width);
		return;
	}
	pthread_mutex_lock(&log_mtx);

	printf("** Dump of %zu bytes at %p (in units of %u):\n", len, d, width);
	while (c < len) {
		switch (width) {
		case 1:
			llen += printf("%02x ", *(uint8_t*)d);
			d++;
			c++;
			break;
		case 2:
			llen += printf("%04x ", *(uint16_t*)d);
			d += 2;
			c += 2;
			break;
		case 4:
			llen += printf("%08x ", *(uint32_t*)d);
			d += 4;
			c += 4;
			break;
		case 8:
			llen += printf("%016jx ", *(uintmax_t*)d);
			d += 8;
			c += 8;
			break;
		}
		if (llen + (2 * width + 1) > 79) {
			printf("\n");
			llen = 0;
		}
	}
	if (llen > 0)
		printf("\n");
	pthread_mutex_unlock(&log_mtx);
}


void shutdown_logger()
{
	pthread_mutex_destroy(&log_mtx);
}
