/*-
 * MiniLibEvent header library
 * (c) 2008.-2011. Ivan Voras <ivoras@gmail.com>
 * Released under the BSD License
 * $Id: minilibevent.h 474 2012-04-08 10:26:01Z ivoras $
 */

#ifndef _MINILIBEVENT_H_
#define _MINILIBEVENT_H_

#ifdef __cplusplus
extern "C" {
#endif

/* Step 1 : detect supported environments */
#include <stdarg.h>

#if __FreeBSD__ > 3 || defined(__NetBSD__)
/* XXX: exact version numbers */
#define HAS_KQUEUE
#endif

#ifdef HAS_KQUEUE
#include <sys/event.h>
#endif

/* Step 2 : Define common interfaces */
#if 0
static void
_ev_debug_printf(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);
}
#else
#define _ev_debug_printf(...)
#endif

/* Step 3 : Implemente kqueue if it's available */
#if defined(HAS_KQUEUE)

/* kqueue behaviour is canonical - other methods should behave the same and
 * have the same guarantees. This specially requires a ev_list structure like
 * kevent that has to have at least fields named "ident" (which holds socket 
 * or file descriptor for which the event was generated or fired), "udata"
 * (which holds user-specified data from ev_set_list_element()) and "flags",
 * which holds 0 or EV_ERROR */

#define EV_READ EVFILT_READ
#define EV_WRITE EVFILT_WRITE

typedef int ev_queue;
typedef int ev_result;
typedef struct kevent* ev_list;


/**
 * Allocates an event list.
 */
static inline ev_list ev_allocate_list(const unsigned int n_elements)
{
#ifdef NDEBUG
	return (ev_list) calloc(n_elements, sizeof(struct kevent));
#else
	/* valgrind complains about structures filled in by the kernel 
	 * being uninitialized, so fuzz it */
	void *lst = malloc(n_elements * sizeof(struct kevent));
	memset(lst, 0, n_elements * sizeof(struct kevent));
	return (ev_list) lst;
#endif
}


/**
 * Expands an event list.
 */
static inline ev_list ev_expand_list(const ev_list l, const unsigned int new_elements)
{
	/* XXX: valgrind will complain about this the same as for ev_allocate_list(). */
	return (ev_list) realloc(l, new_elements * sizeof(struct kevent));
}

/**
 * Frees an event list.
 */
static inline void ev_free_list(const ev_list l)
{
	free(l);
}


/**
 * Describes an event list element.
 */
static inline void ev_set_list_element(const ev_list l, const unsigned int n, const int socket, const int how, const void *data)
{
	_ev_debug_printf("scheduling fd %d for addition\n", socket);
	if (how == EV_READ || how == EV_WRITE)
		EV_SET(&l[n], socket, how, EV_ADD | EV_ENABLE, 0, 0, (void*)data);
	else
		abort();
}


/*
 * Describes a "negative" list element, i.e. schedules it to be removed from the queue.
 */
static inline void ev_set_list_element_remove(const ev_list l, const unsigned int n, const int socket, const int how, const void *data)
{
	_ev_debug_printf("scheduling fd %d for removal\n", socket);
	if (how == EV_READ || how == EV_WRITE)
		EV_SET(&l[n], socket, how, EV_DELETE, 0, 0, (void*)data);
	else
		abort();
}


/**
 * Creates an event queue
 */
static inline ev_queue ev_create_queue()
{
	return kqueue();
}


/**
 * Is the queue valid (i.e. has the creation failed)?
 */
static inline int ev_valid_queue(ev_queue q)
{
	return q >= 0;
}


/**
 * Destroy the queue when it's no longer needed.
 */
static inline void ev_destroy_queue(ev_queue q)
{
	close(q);
}


/**
 * Adds the event list to the queue. Existing events in the queue are not
 * affected.
 */
static inline ev_result ev_enqueue_list(const ev_queue q, const ev_list l, const unsigned int n_elements)
{
	return kevent(q, l, n_elements, NULL, 0, NULL);
}


/**
 * Returnes new events from the event queue.
 */
static inline ev_result ev_get_new_events(const ev_queue q, const ev_list l, const unsigned int max_elements, const struct timespec *wait)
{
	return kevent(q, NULL, 0, l, max_elements, wait);
}


/**
 * Removes a fd from the queue.
 */
static inline void ev_remove_fd(const ev_queue q, const int fd, const int mandatory)
{
	/* kqueue fd's are automatically garbage collected on close(2) - we 
	 * don't need to manually remove them if they are going to be close()d
	 * soon. */
	if (!mandatory)
		return; 
	struct kevent ev[2];
	EV_SET(&ev[0], fd, EVFILT_READ, EV_DELETE, 0, 0, NULL);
	EV_SET(&ev[1], fd, EVFILT_WRITE, EV_DELETE, 0, 0, NULL);
	kevent(q, ev, 2, NULL, 0, NULL);
}


#else
/* Finally, if no special method is available, use a generic implementation
 * based on poll(2). This is suboptimal in many ways, mostly relating to the
 * fact that almost all operations are O(N) complexity, but for simple 
 * tasks this implementation is around 3% slower compared to kqueues. */

#include <poll.h>
#include <pthread.h>

#define EV_READ 0x1
#define EV_WRITE 0x2
#define EV_ERROR 0x4
#define EV_X_FLAG 0x010
#define EV_REMOVE_FLAG 0x20

/* Queue item */
struct ev_item {
	 int fd;
	 void *udata;
	 int pollflags;
	 int filter;
};

struct ev_queue_ {
	unsigned int n_items;
	struct ev_item *items;
	pthread_mutex_t qmtx;
};

typedef struct ev_queue_* ev_queue;

typedef int ev_result;

/* List item */
struct ev_list_item {
	int ident;
	void *udata;
	int pollflags;
	int filter;
	int flags;
};

typedef struct ev_list_item *ev_list;

#ifndef mmin
#define mmin(a,b) (a) < (b) ? (a) : (b)
#endif

static inline ev_list ev_allocate_list(const unsigned int n_elements)
{
	return (ev_list) calloc(n_elements, sizeof(struct ev_list_item));
}


static inline ev_list ev_expand_list(const ev_list l, const unsigned int new_elements)
{
	return (ev_list) realloc(l, new_elements * sizeof(struct ev_list_item));
}


static inline void ev_free_list(const ev_list l)
{
	free(l);
}


static inline void ev_set_list_element(const ev_list l, const unsigned int n, const int socket, const unsigned int filter, const void *data)
{
	l[n].ident = socket;
	l[n].udata = (void*)data;
	l[n].filter = filter;
	l[n].flags = 0;
	if (filter == EV_READ)
		l[n].pollflags = POLLIN;
	else if (filter == EV_WRITE)
		l[n].pollflags = POLLOUT;
	else
		abort();
}


static inline void ev_set_list_element_remove(const ev_list l, const unsigned int n, const int socket, const unsigned int filter, const void *data)
{
	l[n].ident = socket;
	l[n].udata = (void*)data;
	l[n].filter = filter;
	l[n].flags = EV_REMOVE_FLAG;
	if (filter == EV_READ)
		l[n].pollflags = POLLIN;
	else if (filter == EV_WRITE)
		l[n].pollflags = POLLOUT;
	else
		abort();
}


static inline ev_queue ev_create_queue()
{
	ev_queue q;
	q = (ev_queue) malloc(sizeof(struct ev_queue_));
	if (q == NULL)
		return NULL;
	q->n_items = 0;
	q->items = NULL;
	pthread_mutex_init(&q->qmtx, NULL);
	return q;
}


static inline int ev_valid_queue(ev_queue q)
{
	return q != NULL;
}


static inline void ev_destroy_queue(ev_queue q)
{
	pthread_mutex_destroy(&q->qmtx);
	free(q->items);
	free(q);
}

static ev_result ev_enqueue_list(const ev_queue q, const ev_list l, const unsigned int n_elements)
{
	int n_new_items = 0;
	unsigned int i, j;
	int found;
	struct ev_item *nq;
	
	if (n_elements == 0)
		return 0;

	assert(pthread_mutex_lock(&q->qmtx) == 0);
	for (i = 0; i < n_elements; i++) {
		_ev_debug_printf("%p enqueue: fd %d, flags %d\n", q, l[i].ident, l[i].flags);
		if (l[i].flags & EV_REMOVE_FLAG) {
			for (j = 0; j < q->n_items; j++)
				if (q->items[j].fd == l[i].ident && q->items[j].filter == l[i].filter) {
					/* Mark it removed */
					q->items[j].fd = -1;
				}
			continue;
		}
		/* try and fit the new event from l into an empty slot in q */
		found = 0;
		for (j = 0; j < q->n_items; j++) {
			if (q->items[j].fd == l[i].ident && q->items[j].filter == l[i].filter) {
				found = 1;
				break;
			} else if (q->items[j].fd == -1) {
				/* it'll fit in an empty slot */
				q->items[j].fd = l[i].ident;
				q->items[j].pollflags = l[i].pollflags;
				q->items[j].udata = l[i].udata;
				q->items[j].filter = l[i].filter;
				found = 1;
				break;
			}
		}
		if (!found) {
			n_new_items++;
			l[i].pollflags |= EV_X_FLAG;
		}
	}
	
	if (n_new_items == 0) {
		assert(pthread_mutex_unlock(&q->qmtx) == 0);
		return 0;
	}
	
	nq = (struct ev_item*) realloc(q->items, (q->n_items + n_new_items)*sizeof(struct ev_item));
	if (nq == NULL) {
		assert(pthread_mutex_unlock(&q->qmtx) == 0);
		return ENOMEM;
	}
	q->items = nq;
	
	j = q->n_items;
	for (i = 0; i < n_elements; i++) {
		if ( (l[i].pollflags & EV_X_FLAG) != 0) {
			q->items[j].fd = l[i].ident;
			q->items[j].udata = l[i].udata;
			q->items[j].pollflags = l[i].pollflags;
			q->items[j].filter = l[i].filter;
			l[i].pollflags &= ~EV_X_FLAG;
			j++;
		}
	}
	q->n_items = j;
	assert(pthread_mutex_unlock(&q->qmtx) == 0);
	return 0;
}


static struct ev_item *_ev_find_qitem(int fd, const ev_queue q)
{
	unsigned int j;

	for (j = 0; j < q->n_items; j++)
		if (fd == q->items[j].fd)
			return &(q->items[j]);
	return NULL;
}


static ev_result ev_get_new_events(const ev_queue q, const ev_list l, const unsigned int max_elements, const struct timespec *wait)
{
	struct pollfd *fds;
	int timeout;
	unsigned int nfds;
	int r;
	unsigned int i, j, k;
	
	if (max_elements == 0)
		return -1;

	assert(pthread_mutex_lock(&q->qmtx) == 0);
	nfds = q->n_items;
	if (wait != NULL)
		timeout = (wait->tv_sec * 1000) + (wait->tv_nsec / 1000000);
	else
		timeout = 0;
	fds = (struct pollfd*) malloc(nfds * sizeof(*fds));
	if (fds == NULL)
		abort();
	k = 0;
	for (i = 0; i < q->n_items; i++) {
		if (q->items[i].fd == -1)
			continue;
		r = -1;
		for (j = 0; j < k; j++)
			if (fds[j].fd == q->items[i].fd) {
				r = j;
				break;
			}
		if (r == -1) {
			fds[k].fd = q->items[i].fd;
			fds[k].events = q->items[i].pollflags;
			fds[k].revents = 0;
			k++;
		} else {
			fds[r].events |= q->items[i].pollflags;
		}
	}
	nfds = k;
	_ev_debug_printf("%p polling fds (%d): ", q, nfds);
	for (i = 0; i < nfds; i++) {
		_ev_debug_printf("%d", fds[i].fd);
		assert((fds[i].events & POLLIN) || (fds[i].events & POLLOUT));
		if (fds[i].events & POLLIN)
			_ev_debug_printf("r");
		if (fds[i].events & POLLOUT)
			_ev_debug_printf("w");
		_ev_debug_printf(" ");
	}
	_ev_debug_printf(" -- ");

	pthread_mutex_unlock(&q->qmtx);		/* Allow connection acceptor to queue more sockets */
	r = poll(fds, nfds, timeout);
	if (r <= 0) {
		/* _ev_debug_printf("%p: poll(%d) returned %d, errno %d\n", q, nfds, r, errno); */
		free(fds);
		return r;
	}
	pthread_mutex_lock(&q->qmtx);
	
	k = 0; /* number of elements in the returned list l */
	for (i = 0; i < nfds; i++) {
		struct ev_item *evi;
		if (fds[i].revents == 0)
			continue;
		evi = _ev_find_qitem(fds[i].fd, q);
		_ev_debug_printf("%d", fds[i].fd);
		if (evi == NULL) {
			continue;
		}
		if ((fds[i].revents & POLLERR) || (fds[i].revents & POLLHUP) || (fds[i].revents & POLLNVAL)) {
			l[k].ident = evi->fd;
			l[k].udata = evi->udata;
			l[k].pollflags = evi->pollflags;
			l[k].filter = EV_ERROR;
			k++;
			_ev_debug_printf("E");
			if (k >= max_elements)
				goto end;
			continue;
		}
		if (fds[i].revents & POLLIN) {
			l[k].ident = evi->fd;
			l[k].udata = evi->udata;
			l[k].pollflags = evi->pollflags;
			l[k].filter = EV_READ;
			k++;
			_ev_debug_printf("R");
		}
		if (k >= max_elements)
			goto end;
/*		if (fds[i].events & POLLOUT)
			fds[i].revents |= POLLOUT;*/
		if (fds[i].revents & POLLOUT) {
			l[k].ident = evi->fd;
			l[k].udata = evi->udata;
			l[k].pollflags = evi->pollflags;
			l[k].filter = EV_WRITE;
			k++;
			_ev_debug_printf("W");
		}
		if (k >= max_elements)
			goto end;
		_ev_debug_printf(" ");
	}
end:
	free(fds);
	_ev_debug_printf("\n");
	assert(pthread_mutex_unlock(&q->qmtx) == 0);
	return k;
}


static inline void ev_remove_fd(const ev_queue q, int fd, const int mandatory)
{
	unsigned int i;
	/* no distinction between mandatory and non-mandatory remove here */
	pthread_mutex_lock(&q->qmtx);
	for (i = 0; i < q->n_items; i++)
		if (q->items[i].fd == fd) {
			q->items[i].fd = -1;
		}
	pthread_mutex_unlock(&q->qmtx);
}

#endif

#ifdef __cplusplus
}
#endif

#endif /* _MINILIBEVENT_H_ */
