/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2012. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: server.cpp 480 2012-04-10 22:37:02Z ivoras $
 */

#include <sys/param.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <pthread.h>
#include <poll.h>
#include <time.h>
#include <signal.h>

#include "linux-shims.h"

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <paths.h>
#include <syslog.h>
#include <assert.h>
#include <err.h>

#include <new>
#include <list>

#ifdef DMALLOC
#include <dmalloc.h>
#endif

#if __FreeBSD__ >= 7
#include <pthread_np.h>
#define HAS_PTHREAD_SET_NAME_NP
#endif

#include "minilibevent.h"

#include "queue.h"
#include "tree.h"
#include <list>
#include <iterator>

#include "debug.h"
#include "misc.h"
#include "logger.h"
#include "server.h"
#include "mc_protocol.h"
#include "DataStore.h"
#include "BackendMesh.h"
#include "Thread.h"
#include "Mutex.h"
#include "ServerWorkerThread.h"
#include "ServerNetworkThread.h"
#include "ServerReaperThread.h"
#include "ServerMonitorThread.h"
#include "ServerModelMetadata.h"
#include "ServerClientConnection.h"
#include "ServerModelMetadata.h"

#define _MDCACHED_ID_ "$Id: server.cpp 480 2012-04-10 22:37:02Z ivoras $"

int argc;
char **argv;

static unsigned int inet_port = MC_DEFAULT_INET_PORT;
static const char *unix_soname = MC_DEFAULT_UNIX_SOCKET;

static std::list<ServerClientConnection*> connections;
static Mutex connections_mutex;
static int sock_tcp, sock_udp, sock_unix;	/* listener sockets */

static pthread_attr_t pthread_detached;

unsigned int n_worker_threads = DEFAULT_N_WORKER_THREADS;
ServerWorkerThread *worker_threads;		/* worker thread pool */

unsigned int n_network_threads = DEFAULT_N_NETWORK_THREADS;
ServerNetworkThread *network_threads;		/* network thread pool */
static unsigned int rr_network_thread = 0;	/* round-robin network thread ID */

DataStore *cache = NULL;
ServerReaperThread *reaper = NULL;
ServerMonitorThread *monitor = NULL;

unsigned int inbuf_alloc_size = DEFAULT_INBUF_ALLOC_SIZE;
unsigned int outbuf_alloc_size = DEFAULT_OUTBUF_ALLOC_SIZE;
bool adaptive_reconfiguration = true;

ServerModelMetadata *model_metadata = NULL;
unsigned int n_model_metadata = 0;

BackendMesh *bmesh = NULL;
char **backend_peers = 0;
unsigned int n_backend_peers = 0;

static char *cache_dump_file = NULL;
static bool cache_read_dump_flag = false;
static bool cache_dump_flag = false;
static bool cache_dump_running = false;
static int cache_dump_interval = 0;
static pthread_t cache_dump_thread;

/** SIGINT flag */
static bool interrupted_flag = false;

/** Strongly bind network threads to user threads */
bool strongly_bind_flag = false;

int log_level = LOG_ERR;
int debug_mode = 0;
#ifndef NDEBUG
bool internal_debug = true;
#endif

/** Memory limit, in bytes */
uint64_t server_memory_limit = 0;

/** Avoid time() syscalls by setting this global variable from the ticker thread. */
time_t current_time;

/** Ticker thread handle */
static pthread_t ticker_thr;

#ifndef NDEBUG
struct debug_status_record global_st;
#ifdef DMALLOC
int sig_dmalloc = 0;
#endif
#endif

/* A helper buffer for UDP trickery */
#define HELPERBUF_ALLOC_SIZE 65536

extern "C" {
static void atexit_handler();
static void signal_exit_handler(int sig);
static void signal_dump_data(int sig);
}

static void
atexit_handler()
{
	if (unix_soname != NULL)
		unlink(unix_soname);
}


/**
 * Handler of various exit signals (SIGINT, SIGTERM, etc.)
 */
static void
signal_exit_handler(int sig)
{
	if (unix_soname != NULL)
		unlink(unix_soname);
	if (sig == SIGINT) {
		printf("Interrupted.\n");
		if (interrupted_flag)
			_exit(0);
		else
			interrupted_flag = TRUE;
		return;
	}
	_exit(0);
}


#ifdef SIGINFO
/**
 * SIGINFO handler.
 */
extern "C" {
static void signal_info(int sig);
}

static void
signal_info(int sig)
{
#ifndef NDEBUG
	printf("global_status:\t\t'%c' at %s:%d (int:%d)\n", global_st.status, global_st.file, global_st.line, interrupted_flag);
	printf("mesh_status:\t\t'%c' at %s:%d\n", bmesh->mesh_st.status, bmesh->mesh_st.file, bmesh->mesh_st.line);
	for (unsigned int i = 0; i < n_worker_threads; i++)
		printf("worker[%u]_status:\t'%c' at %s:%d\n", i, worker_threads[i].st.status, worker_threads[i].st.file, worker_threads[i].st.line);
	for (unsigned int i = 0; i < n_network_threads; i++)
		printf("network[%u]_status:\t'%c' at %s:%d\n", i, network_threads[i].st.status, network_threads[i].st.file, network_threads[i].st.line);
	connections_mutex.Lock();
	for (std::list<ServerClientConnection*>::iterator it = connections.begin(); it != connections.end(); it++) {
		ServerClientConnection *cn = *it;
		printf("client[%p]_status:\t'%c' at %s:%d\n", cn, cn->st.status, cn->st.file, cn->st.line);
	}
	connections_mutex.Unlock();
#ifdef DMALLOC
	sig_dmalloc = 1;
#endif
#else
	printf("Program compiled with NDEBUG: debugging points not present.\n");
#endif
	SizeHistogram<MODEL_SIZE_HISTOGRAM_BINS> totals;
	for (unsigned i = 0; i < n_model_metadata; i++)
		totals.mergeHistogram(model_metadata[i].getHistogram());
	printf("Buffer sizes histogram (peak = %u):\n", 1 << totals.getPeakBin());
	for (unsigned i = 0; i < MODEL_SIZE_HISTOGRAM_BINS; i++)
		if (totals.getPoint(i) != 0)
			printf("\t%5u : %u\n", 1 << i, totals.getPoint(i));
}
#endif


/**
 * Set cache_dump_flag, causing the cache to be dumped real soon now.
 */
static void
signal_dump_data(int sig)
{
	cache_dump_flag = true;
}


/**
 * Setup signal handling
 */
static void
setup_signals()
{
	signal(SIGPIPE, SIG_IGN);
	signal(SIGINT, signal_exit_handler);
	signal(SIGHUP, signal_exit_handler);
	signal(SIGTERM, signal_exit_handler);
	signal(SIGUSR1, signal_dump_data);
#ifdef SIGINFO
	signal(SIGINFO, signal_info);
#endif
}


/**
 * Ticker thread
 * Basically an attempt to avoid time() syscalls by updating the global variable
 * current_time. Since it is unsynchronized, CPUs will pick up the actual time
 * "eventually" which is hopefully good enough.
 * The thread also servers as an internal "cron job driver".
 */
static void *
ticker_thread(void *arg)
{
	while (true) {
		current_time = time(NULL);
		if (cache_dump_interval > 0 && current_time % cache_dump_interval == 0)
			cache_dump_flag = TRUE;
		sleep(1);
	}
	return NULL;
}


/**
 * Fires up the ticket thread.
 */
static void
setup_ticker()
{
	pthread_create(&ticker_thr, NULL, ticker_thread, NULL);
}


/**
 * Separates the process from its controlling terminal. Taken from FreeBSD's
 * libc.
 */
static void
daemonize()
{
	int fd;
	switch(fork()) {
	case -1:
		fprintf(stderr, "Error forking to background: %s", strerror(errno));
		exit(1);
		break;
	case 0:
		break;
	default:
		exit(0);
	}
	if ((fd = open(_PATH_DEVNULL, O_RDWR, 0)) != -1) {
		dup2(fd, STDIN_FILENO);
		dup2(fd, STDOUT_FILENO);
		dup2(fd, STDERR_FILENO);
		if (fd > 2)
			close(fd);
	}
}


/**
 * OOM handler.
 */
void out_of_memory() {
	fprintf(stderr, "Memory exhausted - \"new\" failed\n");
	exit(1);
}


/**
 * Dump the cache to the cache_dump_file.
 */
static void *
cache_dump_f(void *a)
{
	char *tempfn;

	if (cache_dump_file == NULL) {
		mc_log(LOG_WARNING, "Attempted to dump cache prewarm file but no filename defined");
		cache_dump_running = false;
		return (NULL);
	}
	if (asprintf(&tempfn, "%s.%d", cache_dump_file, getpid()) == -1) {
		tempfn = NULL;
		mc_log(LOG_WARNING, "asprintf() failed in cache_dump_f()");
		return (NULL);
	}
	if (cache->prewarm_cache_dump(tempfn, false)) {
		if (rename(tempfn, cache_dump_file) == 0)
			mc_log(LOG_INFO, "Dumped cache data to %s", cache_dump_file);
		else
			mc_log(LOG_ERR, "Error replacing cache dump file '%s' with '%s' - temporary file not deleted",
				cache_dump_file, tempfn);
	} else {
		mc_log(LOG_ERR, "Error dumping cache prewarm file to '%s'", tempfn);
		unlink(tempfn);
	}
	free(tempfn);
	cache_dump_running = false;
	return (NULL);
}


/**
 * Read cache dump from cache_dump_file.
 */
static void
read_cache_dump()
{
	if (cache->prewarm_cache_read(cache_dump_file, false))
		mc_log(LOG_INFO, "Read prewarm cache dump from '%s'", cache_dump_file);
	else
		mc_log(LOG_ERR, "Error reading prewarm cache dump from '%s'", cache_dump_file);
}


/**
 * Shuts down a connection, frees it, and removes it from the list of connections.
 */
void
shutdown_connection(ServerClientConnection *conn)
{
	assert(conn->socket >= 0);
	conn->disconnect_client();
#ifndef NDEBUG
	conn->socket = -1;
#endif
	connections_mutex.Lock();
	connections.remove(conn);
	connections_mutex.Unlock();
	delete conn;
}


/**
 * For TCP and Unix server sockets: Loops to accept new connections and process
 * old connections. Thread running this network loop is the connection accepting
 * thread.
 * For UDP server sockets: TODO: Accepts packets, dispatches them to worker threads.
 */
static void
network_loop()
{
	class ServerClientConnection *conn = NULL, *newconn;
	struct timespec ts_wait;
	int i, ne, max_evs;
	char *helperbuf;
	ev_queue evq;
	ev_result evr;
	ev_list evs;

	helperbuf = (char*) malloc(HELPERBUF_ALLOC_SIZE);
	CRIT_ASSERT(helperbuf != NULL);

	/*max_evs = FIRST_GENERIC_CONN;*/
	max_evs = 10;

	evs = ev_allocate_list(max_evs);
	if (evs == NULL)
		mc_log(LOG_CRIT, "Error allocating evs: %s", strerror(errno));

	CRIT_ASSERT((evq = ev_create_queue()) >= 0);
	/* Add server sockets to event queue */
	ne = 0;
	for (std::list<ServerClientConnection*>::iterator it = connections.begin(); it != connections.end(); it++) {
		ServerClientConnection *cn = *it;
		if (cn->type == CONN_TYPE_SERVER_UNIX || cn->type == CONN_TYPE_SERVER_TCP) {
			ev_set_list_element(evs, ne++, cn->socket, EV_READ, cn);
		} else if (cn->type == CONN_TYPE_SERVER_UDP) {
			/* XXX: UDP comm is not yet implemented! */
			mc_log(LOG_WARNING, "UDP is not implemented");
		}

	}
	if (ne == 0)
		CRIT_ERROR("No server sockets?");
	CRIT_ASSERT(ev_enqueue_list(evq, evs, ne) == 0);

	ts_wait.tv_sec = 0;
	ts_wait.tv_nsec = _s_to_ns(0.1);

	while (1) {
		DEBUG_STATUS_POINT(global_st, 'l');
		if (cache_dump_flag) {
			/* Delegate cache dumping to a separate thread */
			if (!cache_dump_running) {
				cache_dump_running = true;
				pthread_create(&cache_dump_thread, &pthread_detached, cache_dump_f, NULL);
			}
			cache_dump_flag = false;
		}
		if (interrupted_flag) {
			mc_log(LOG_ERR, "Interrupted by SIGINT");
			break;
		}
#ifdef DMALLOC
#ifndef NDEBUG
		if (sig_dmalloc) {
			dmalloc_log_unfreed();
			sig_dmalloc = 0;
		}
#endif
#endif
		DEBUG_STATUS_POINT(global_st, 'e');
		evr = ev_get_new_events(evq, evs, max_evs, &ts_wait);
		DEBUG_STATUS_POINT(global_st, 'E');
		if (evr < 0) {
			if (errno != EINTR)
				mc_log(LOG_CRIT, "Critical event queue failure: %s", strerror(errno));
			else
				mc_log(LOG_ERR, "Event queue interrupted: %s", strerror(errno));
		}

		for (i = 0; i < evr; i++) {
			conn = (ServerClientConnection*) evs[i].udata;	/* The connection which generated the event */
			CRIT_ASSERT(conn != NULL);
			if (conn->type == CONN_TYPE_NO)
				continue;
			if ((evs[i].flags & EV_ERROR) != 0) {
				printf("Trying to remove fd socket %d\n", conn->socket);
				ev_remove_fd(evq, conn->socket, FALSE);
				shutdown_connection(conn);
			} else if (evs[i].filter == EV_READ && (conn->type == CONN_TYPE_SERVER_UNIX || conn->type == CONN_TYPE_SERVER_TCP)) {
				/* Accept new connections */
				unsigned int j;
				bool can_connect;
				socklen_t ss;
				int rv;

				can_connect = true;
				DEBUG_STATUS_POINT(global_st, 'C');
				newconn = new ServerClientConnection();
				if (conn->type == CONN_TYPE_SERVER_UNIX) {
					newconn->type = CONN_TYPE_CLIENT_UNIX;
					ss = sizeof(newconn->sun);
					newconn->socket = accept(conn->socket, (struct sockaddr*)&newconn->sun, &ss);
					if (newconn->socket < 0) {
						mc_log(LOG_ERR, "Socket accept() error: %s", strerror(errno));
						delete newconn;
						continue;
					}
					connections_mutex.Lock();
					connections.push_back(newconn);
					connections_mutex.Unlock();
				} else if (conn->type == CONN_TYPE_SERVER_TCP) {
					newconn->type = CONN_TYPE_CLIENT_TCP;
					ss = sizeof(newconn->sin);
					newconn->socket = accept(conn->socket, (struct sockaddr*)&newconn->sin, &ss);
					if (newconn->socket < 0) {
						mc_log(LOG_ERR, "Socket accept() error: %s", strerror(errno));
						delete newconn;
						continue;
					}
					j = 1; /* TCP_NODELAY enable */
					setsockopt(newconn->socket, IPPROTO_TCP, TCP_NODELAY, (char*)&j, sizeof(j));
					connections_mutex.Lock();
					connections.push_back(newconn);
					connections_mutex.Unlock();
				} else
					mc_log(LOG_CRIT, "Server connection type unknown: %d", conn->type);
				newconn->flags = ZERO;
				/* Set nonblocking socket */
				int sflag;
				if ((sflag = fcntl(newconn->socket, F_GETFL, 0)) == -1)
					sflag = 0;
				fcntl(newconn->socket, F_SETFL, sflag | O_NONBLOCK);
				/* Set SO_SNDBUF and SO_RCVBUF */
				sflag = SNDBUF_SIZE;
				setsockopt(newconn->socket, SOL_SOCKET, SO_SNDBUF, &sflag, sizeof(sflag));
				sflag = RCVBUF_SIZE;
				setsockopt(newconn->socket, SOL_SOCKET, SO_RCVBUF, &sflag, sizeof(sflag));
				/* Initialize our own buffers */
				newconn->assure_bufs(false);
				if (!newconn->inbufValid()) {
					mc_log(LOG_ERR, "Cannot allocate %d bytes for inbuf", inbuf_alloc_size);
					close(newconn->socket);
					newconn->type = CONN_TYPE_NO;
					continue;
				}
				newconn->inbufReset();

				if (n_network_threads > 0) {
					/* Add the new connection to the event queue of a selected network thread */
					newconn->setEvq(network_threads[rr_network_thread].getEvq());
				} else {
					/* Add the new connection to the main loop event queue */
					newconn->setEvq(&evq);
				}
				rv = newconn->enable_read_event();
				if (rv != 0) {
					mc_log(LOG_ERR, "Cannot add a new connection (fd %d) to the event queue: %s (%d)",
							newconn->socket, strerror(errno), errno);
					if (close(newconn->socket) < 0)
						mc_log(LOG_ERR, "Cannot close newly accepted socket %d: %s", newconn->socket, strerror(errno));
					newconn->type = CONN_TYPE_NO;
				} else
					mc_log(LOG_DEBUG, "Accepted a %s (%p, fd %d)", conn_type_name(newconn->type), newconn, newconn->socket);

				if (n_network_threads > 0) {
					rr_network_thread++;
					if (rr_network_thread >= n_network_threads)
						rr_network_thread = 0;
				}
				/* That's it. We've accepted a new client */
				DEBUG_STATUS_POINT(global_st, 'c');
			} else {
				/* Handle in-line network IO in SEDA (no additional network threads). */
				assert(n_network_threads == 0);
				assert(conn->type == CONN_TYPE_CLIENT_UNIX || conn->type == CONN_TYPE_CLIENT_TCP || conn->type == CONN_TYPE_CLIENT_UDP);
				if (evs[i].filter == EV_READ) {
					/* Handle incoming data */
					if (conn->process_read_event(NULL, &model_metadata[0]) == EVPROC_ERROR) {
						ev_remove_fd(evq, conn->socket, FALSE);
						shutdown_connection(conn);
						continue;
					}
					if (n_network_threads == 0 && n_worker_threads > 0) {
						worker_threads[0].queueUnlock();
						worker_threads[0].queueSignal();
					}
				} else if (evs[i].filter == EV_WRITE && conn->outputDataReady()) {
					/* Handle outgoing data */
					if (conn->process_write_event() == EVPROC_ERROR) {
						ev_remove_fd(evq, conn->socket, FALSE);
						shutdown_connection(conn);
						continue;
					}
				}
			}
		}
		time(&current_time);
	}
	ev_free_list(evs);
	ev_destroy_queue(evq);
	free(helperbuf);
	DEBUG_STATUS_POINT(global_st, 'X');
}


/**
 * Sets up the server sockets, the connections array and the thread pools.
 */
static void
setup_server()
{
	int flags;
#ifdef HAS_PTHREAD_SET_NAME_NP
	char thr_name[50];
#endif

	CRIT_ASSERT(pthread_attr_init(&pthread_detached) == 0);
	CRIT_ASSERT(pthread_attr_setdetachstate(&pthread_detached, PTHREAD_CREATE_DETACHED) == 0);

	if (inet_port != 0) {
		/* Set up TCP server socket */
		struct sockaddr_in sin;

		memset(&sin, 0, sizeof sin);
		sock_tcp = socket(AF_INET, SOCK_STREAM, 0);
		if (sock_tcp < 0)
			mc_log(LOG_CRIT, "Cannot create TCP socket: %s", strerror(errno));
		sin.sin_family = AF_INET;
		sin.sin_addr.s_addr = htonl(INADDR_ANY);
		sin.sin_port = htons(inet_port);
		if (bind(sock_tcp, (struct sockaddr*)&sin, sizeof(sin)) < 0)
			mc_log(LOG_CRIT, "Cannot bind to TCP port %d: %s", inet_port, strerror(errno));
		if (listen(sock_tcp, 100) < 0)
			mc_log(LOG_CRIT, "Cannot listen on TCP socket: %s", strerror(errno));
		if ((flags = fcntl(sock_tcp, F_GETFL, 0)) == -1)
			flags = 0;
		fcntl(sock_tcp, F_SETFL, flags | O_NONBLOCK);
		flags = 1;
		setsockopt(sock_tcp, IPPROTO_TCP, TCP_NODELAY, (char*)&flags, sizeof(flags));
		connections.push_back(new ServerClientConnection(sock_tcp, CONN_TYPE_SERVER_TCP));

		/* Set up UDP server socket */
		memset(&sin, 0, sizeof sin);
		sock_udp = socket(AF_INET, SOCK_DGRAM, 0);
		if (sock_udp < 0)
			mc_log(LOG_CRIT, "Cannot create UDP socket: %s", strerror(errno));
		sin.sin_family = AF_INET;
		sin.sin_addr.s_addr = htonl(INADDR_ANY);
		sin.sin_port = htons(inet_port);
		if (bind(sock_udp, (struct sockaddr*)&sin, sizeof(sin)) < 0)
			mc_log(LOG_CRIT, "Cannot bind to UDP port %d: %s", inet_port, strerror(errno));
/*		if (listen(sock_udp, n_conns*2) < 0)
			mc_log(LOG_CRIT, "Cannot listen on UDP socket: %s", strerror(errno));*/
		if ((flags = fcntl(sock_udp, F_GETFL, 0)) == -1)
			flags = 0;
		fcntl(sock_udp, F_SETFL, flags | O_NONBLOCK);
		connections.push_back(new ServerClientConnection(sock_udp, CONN_TYPE_SERVER_UDP));
	}
	if (unix_soname != NULL) {
		/* Set up Unix socket */
		struct sockaddr_un sun;

		memset(&sun, 0, sizeof sun);
		sock_unix = socket(AF_LOCAL, SOCK_STREAM, 0);
		if (sock_unix < 0)
			mc_log(LOG_CRIT, "Cannot create Unix socket: %s", strerror(errno));
		strncpy(sun.sun_path, unix_soname, sizeof(sun.sun_path));
		sun.sun_family = AF_LOCAL;
#ifdef __BSD_VISIBLE
		sun.sun_len = strlen(unix_soname);
#endif
		/* Delete the old file/socket if it exists */
		if (access(unix_soname, F_OK) == 0) {
			mc_log(LOG_WARNING, "Warning: Unix socket file (%s) exists - overwriting!", unix_soname);
			if (unlink(unix_soname) != 0)
				mc_log(LOG_ERR, "Error: Cannot unlink %s: %s", unix_soname, strerror(errno));
		}
		if (bind(sock_unix, (struct sockaddr*)&sun, SUN_LEN(&sun)) < 0)
			mc_log(LOG_CRIT, "Cannot bind Unix socket to %s: %s", unix_soname, strerror(errno));
		if (listen(sock_unix, 100) < 0)
			mc_log(LOG_CRIT, "Cannot listen on Unix socket: %s", strerror(errno));
		if ((flags = fcntl(sock_unix, F_GETFL, 0)) == -1)
			flags = 0;
		fcntl(sock_unix, F_SETFL, flags | O_NONBLOCK);
		connections.push_back(new ServerClientConnection(sock_unix, CONN_TYPE_SERVER_UNIX));
	}
	cache = new DataStore();
	bmesh = new BackendMesh(cache);
	bmesh->start();
	reaper = new ServerReaperThread(cache);
	reaper->start();
	monitor = new ServerMonitorThread();
	monitor->start();
	/* Initialize the worker thread pool */
	CRIT_ASSERT(cache != NULL);
	worker_threads = new ServerWorkerThread[n_worker_threads];
	for (unsigned i = 0; i < n_worker_threads; i++) {
		worker_threads[i].index = i;
		worker_threads[i].start();
#ifdef HAS_PTHREAD_SET_NAME_NP
		sprintf(thr_name, "Worker#%d", i);
		worker_threads[i].set_name(thr_name);
#endif
	}
	/* Initialize the network thread pool */
	network_threads = new ServerNetworkThread[n_network_threads];
	for (unsigned i = 0; i < n_network_threads; i++) {
		network_threads[i].index = i;
		if (n_worker_threads != 0)
			network_threads[i].setWorkerIndex(i % n_worker_threads);
		network_threads[i].start();
#ifdef HAS_PTHREAD_SET_NAME_NP
		sprintf(thr_name, "Network#%d", i);
		network_threads[i].set_name(thr_name);
#endif
	}
	mc_log(LOG_INFO, "%d worker threads, %d network threads", n_worker_threads, n_network_threads);
	if (cache_dump_file != NULL)
		mc_log(LOG_INFO, "Cache dump file: '%s'", cache_dump_file);
	/* Model metadata is intended to be updated by the threads doing the
	 * actual work, i.e. those accesing the DataStore, to avoid locking
	 * and contention. */
	if (n_worker_threads > 0)
		/* SEDA, AMPED */
		n_model_metadata = n_worker_threads;
	else if (n_network_threads > 0)
		/* SYMPED */
		n_model_metadata = n_network_threads;
	else if (n_network_threads == 0 && n_worker_threads == 0)
		n_model_metadata = 1;
	assert(n_model_metadata > 0);
	model_metadata = new ServerModelMetadata[n_model_metadata];
}


/**
 * Free memory, free worker threads, everything.
 * TODO: wait until the threads stop.
 */
static void
stop_everything()
{
	mc_log(LOG_INFO, "Shutting down");
	bmesh->request_stop();
	reaper->request_stop();
	monitor->request_stop();
	for (unsigned int i = 0; i < n_worker_threads; i++)
		worker_threads[i].request_stop();
	for (unsigned int i = 0; i < n_network_threads; i++)
		network_threads[i].request_stop();

	DEBUG_STATUS_POINT(global_st, 'b');
	bmesh->join();
	reaper->join();
	monitor->join();
	
	DEBUG_STATUS_POINT(global_st, 'j');
	for (unsigned int i = 0; i < n_worker_threads; i++)
		worker_threads[i].join();
	DEBUG_STATUS_POINT(global_st, 'J');
	for (unsigned int i = 0; i < n_network_threads; i++)
		network_threads[i].join();

	delete cache;

	while (!connections.empty()) {
		delete connections.front();
		connections.pop_front();
	}
	delete[] worker_threads;
	delete[] network_threads;

	for (unsigned i = 0; i < n_backend_peers; i++)
		free(backend_peers[i]);
	free(backend_peers);

	shutdown_logger();
}


static void
parse_backends(char *str)
{
	const char *sep = " ,";
	char *word, *ctx;

	for (word = strtok_r(str, sep, &ctx); word; word = strtok_r(NULL, sep, &ctx)) {

		/* TODO: inet_aton(), gethostbyname() */

		n_backend_peers++;
		backend_peers = (char**)realloc(backend_peers, n_backend_peers * sizeof(*backend_peers));
		backend_peers[n_backend_peers-1] = strdup(word);
	}
}


/**
 * Output a help message describing command line arguments.
 */
static void
usage()
{
	printf("usage: %s [-h] [-a] [-d] [-f file] [-i] [-k] [-m #] [-n #] [-r] [-s #] [-t #] [-v #]\n", argv[0]);
	printf("	-a	Disable runtime adaptive reconfiguration\n");
/*	printf("	-b	Comma-separated list of peers to connect to\n");*/
	printf("	-d	Debug mode (doesn't fork to background)\n");
	printf("	-f file	Set dump file name (for use with SIGUSR1)\n");
	printf("	-h	Show this help message\n");
	printf("	-i #	Automatically create a cache dump every # seconds\n");
	printf("	-k	Bind network threads to worker threads\n");
	printf("	-m #	Limit memory use to # MiB (default=0, unlimited)\n");
	printf("	-n #	Set number of network threads to # (default=%d)\n", DEFAULT_N_NETWORK_THREADS);
	printf("	-r	Read cache data from the dump file (prewarm the cache)\n");
	printf("	-s #	Set default (median) record size, in bytes (>= 32, default=%u)\n", inbuf_alloc_size);
	printf("	-t #	Set number of worker threads to # (default=%d)\n", DEFAULT_N_WORKER_THREADS);
	printf("	-v #	Set minimum log (verbosity) level to # [%d-%d]\n", LOG_EMERG, LOG_DEBUG);
	printf("\n%s\n", _MDCACHED_ID_);
}


/**
 * Parse command line arguments.
 */
static void
parse_args()
{
	int ch;
	while ((ch = getopt(argc, argv, "ab:df:hi:km:n:rs:t:v:")) != -1) {
		switch(ch) {
		case 'h':
		case '?':
			printf("Bullet Cache server version: %d, protocol version: %d\n",
				SERVER_VERSION, MC_PROTOCOL_VERSION);
			printf("DataStore is created from %d item buckets and %d tag buckets\n",
				N_ITEM_BUCKETS, N_TAG_BUCKETS);
			printf("Default Unix socket: %s\n", unix_soname);
			usage();
			exit(1);
			break;
		case 'a':
			adaptive_reconfiguration = false;
			break;
		case 'b':
			parse_backends(optarg);
			break;
		case 'd':
			if (!debug_mode) {
				debug_mode = true;
				log_level = LOG_DEBUG;
			}
			break;
		case 'f':
			cache_dump_file = (char*) malloc(PATH_MAX);
			if (realpath(optarg, cache_dump_file) == NULL && errno != ENOENT)
				errx(1, "Cannot resolve dump path: %s (problem at '%s'?): %s", optarg, cache_dump_file, strerror(errno));
			break;
		case 'i':
			cache_dump_interval = atoi(optarg);
			if (cache_dump_interval == 0)
				warn("Invalid cache dump interval (-i) -- disabled");
			break;
		case 'k':
			strongly_bind_flag = TRUE;
			break;
		case 'm':
			server_memory_limit = atoi(optarg) * (1024L * 1024L);
			break;
		case 'n':
			n_network_threads = atoi(optarg);
			break;
		case 'r':
			cache_read_dump_flag = TRUE;
			break;
		case 's':
			inbuf_alloc_size = outbuf_alloc_size = atoi(optarg);
			if (inbuf_alloc_size < 32)
				errx(1, "Invalid default buffer size: %d", inbuf_alloc_size); 
			break;
		case 't':
			n_worker_threads = atoi(optarg);
#ifdef DMALLOC
			if (n_worker_threads != 0)
				errx(1, "DMALLOC sucks with threading");
#endif
			break;
		case 'v':
			log_level = atoi(optarg);
			break;
		default:
			fprintf(stderr, "Invalid argument: %c\n", ch);
			usage();
			exit(1);
			break;
		}
	}
	if (strongly_bind_flag && n_worker_threads != n_network_threads && n_worker_threads != 0)
		mc_log(LOG_WARNING, "Binding network threads to worker threads makes no sense if thread counts don't match.");
	else if (strongly_bind_flag)
		mc_log(LOG_DEBUG, "Binding network threads to worker threads");
}


/**
 * Preliminary environment sanity checks.
 */
static void
internal_checks() {
	if (!ISPO2(N_ITEM_BUCKETS) || N_ITEM_BUCKETS <= 0)
		errx(1, "N_ITEM_BUCKETS must be defined as a power-of-2 integer");
	if (!ISPO2(N_TAG_BUCKETS) || N_TAG_BUCKETS <= 0)
		errx(1, "N_TAG_BUCKETS must be defined as a power-of-2 integer");
	if (DEFAULT_N_DATA_NODES <= 0)
		errx(1, "DEFAULT_N_DATA_NODES must be a positive nonzero integer");
}


/**
 * Rock and roll!
 */
int
main(int _argc, char **_argv)
{
	internal_checks();
#ifndef NDEBUG
	log_level = LOG_DEBUG;	/* increase verbosity */
#endif

	argc = _argc;
	argv = _argv;
	parse_args();

	std::set_new_handler(out_of_memory);
	DEBUG_STATUS_POINT(global_st, '.');
	time(&current_time);

	setup_logger();
	umask(0111);
	if (!debug_mode)
		daemonize();
	setup_signals();
	atexit(atexit_handler);
	setup_ticker();
	setup_server();
	if (cache_read_dump_flag)
		read_cache_dump();
	network_loop();
	mc_log(LOG_WARNING, "Cleaning up before exiting...");
	stop_everything();
	return 0;
}
